
/*******************************************************************************
*
* File check3.c
*
* Copyright (C) 2016 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Importing gauge potentials previously exported by check2.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "pmdb.h"
#include "random.h"
#include "utils.h"
#include "lattice.h"
#include "linalg.h"
#include "uflds.h"
#include "archive.h"
#include "global.h"

#define NMAX 4


static void set_act(int n,sl3_alg_dble **ad,complex_dble *act)
{
   su3_dble **wud,*ud;

   wud=reserve_wud(n+1);
   ud=wud[0];
   a2uflds(n,ad,ud);
   Splaq(n,ud,act);
   release_wud();
}


static void random_afld(sl3_alg_dble *ad)
{
   if (cpr[0]==(NPROC0-1))
      gauss_dble((double*)(ad),16*(4*VOLUME+7*(BNDRY/4)+3));
   else
      gauss_dble((double*)(ad),16*(4*VOLUME+7*(BNDRY/4)));

   set_adbc(ad);
}


static double dist(complex_dble w,complex_dble z)
{
   double d,m;

   w.re-=z.re;
   w.im-=z.im;

   d=w.re*w.re+w.im*w.im;
   m=z.re*z.re+z.im*z.im;

   return sqrt(d/m);
}


int main(int argc,char *argv[])
{
   int my_rank,bc,ir,ie;
   int n,i,nsize;
   stdint_t istd[5];
   complex_dble *act0,*act1,*act2;
   sl3_alg_dble **wad;
   char afld_dir[NAME_SIZE],name[NAME_SIZE];
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   set_error_file("check3.log",0);

   if (my_rank==0)
   {
      flog=freopen("check3.log","w",stdout);
      fin=freopen("check2.in","r",stdin);

      printf("\n");
      printf("Importing gauge potentials exported by check2\n");
      printf("---------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("afld_dir","%s\n",afld_dir);
      fclose(fin);

      bc=find_opt(argc,argv,"-bc");
      error_root((bc==0)||(sscanf(argv[bc+1],"%d",&bc)!=1),1,
                 "main [check3.c]","Syntax: check3 -bc <type>");
   }

   MPI_Bcast(afld_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,9876);
   geometry();

   check_dir_root(afld_dir);
   nsize=name_size("%s/testafldn%d",afld_dir,NMAX);
   error_root(nsize>=NAME_SIZE,1,"main [check3.c]","afld_dir name is too long");

   alloc_wad(NMAX);
   alloc_wud(NMAX);
   act0=malloc(3*NMAX*sizeof(*act0));
   act1=act0+NMAX;
   act2=act1+NMAX;

   for (n=0;n<NMAX;n++)
   {
      wad=reserve_wad(n+1);

      for (i=0;i<=n;i++)
         random_afld(wad[i]);

      set_act(n,wad,act0);

      if (my_rank==0)
      {
         printf("Random gauge potential of order %d\n",n);
         printf("Action:\n");

         for (i=0;i<=n;i++)
            printf("act[%d] = (%+.15e,%+.15e)\n",i,act0[i].re,act0[i].im);

         printf("\n");
         sprintf(name,"%s/testafldn%d",afld_dir,n);
         printf("Now read gauge potential from file %s:\n",name);
         fin=fopen(name,"rb");
         error_root(fin==NULL,1,"main [check3.c]","Unable to open input file");

         ir=fread(istd,sizeof(stdint_t),5,fin);
         ir+=fread(act1,sizeof(complex_dble),n+1,fin);
         error_root(ir!=6+n,1,"main [check3.c]","Incorrect read count");
         fclose(fin);

         if (endianness()==BIG_ENDIAN)
         {
            bswap_int(5,istd);
            bswap_double(2*(n+1),act1);
         }

         printf("%dx%dx%dx%d lattice, number of fields = %d\n",
                (int)(istd[0]),(int)(istd[1]),(int)(istd[2]),
                (int)(istd[3]),(int)(istd[4]));
         printf("Action:\n");

         for (i=0;i<=n;i++)
            printf("act[%d] = (%+.15e,%+.15e)\n",i,act1[i].re,act1[i].im);
      }

      import_afld(name,n,wad);
      ie=1;

      for (i=0;i<=n;i++)
         ie&=check_adbc(wad[i]);

      error(ie!=1,1,"main [check3.c]","Boundary conditions are not preserved");
      set_act(n,wad,act2);

      if (my_rank==0)
      {
         printf("Should be:\n");

         for (i=0;i<=n;i++)
            printf("act[%d] = (%+.15e,%+.15e), relative deviation = %.1e\n",
                   i,act2[i].re,act2[i].im,dist(act1[i],act2[i]));

         printf("\n\n");
         remove(name);
      }

      release_wad();
   }

   if (my_rank==0)
      fclose(flog);

   MPI_Finalize();
   exit(0);
}
