
/*******************************************************************************
*
* File cnrm2.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Exact leading-order contribution to the time-slice norm of the ghost
* field.
*
* The externally accessible function is
*
*   void cnrm2(double *cnrm)
*     Computes the leading-order contribution to the time-slice norm of
*     the ghost field as a function of the time x0 (see the notes).
*
* Notes:
*
* For both SF and open-SF boundary conditions, the values assigned to
* cnrm[x0] are given by
*
*   <c^a(x)*cbar^a(x)> for x0=0,..,T-1,
*
* where T=NPROC0*L0 and the correlation function is the exact leading-order
* expression for the ghost-field two-point function.
*
* The program cnrm2() acts globally and must be called on all MPI processes
* simultaneously.
*
*******************************************************************************/

#define CNRM2_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "utils.h"
#include "pmdb.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)

static int init=0;
static double *phat[4],co[L0][N0],si[L0][N0],loc_cnrm[N0];


static void set_phat(void)
{
   int l[4],mu,k,k0,x0,bc;
   double r[4],pi,p0;

   phat[0]=malloc((L0+L1+L2+L3)*sizeof(double));
   error(phat[0]==NULL,1,"set_phat [cnrm2.c]",
         "Unable to allocate auxiliarly array");
   phat[1]=phat[0]+L0;
   phat[2]=phat[1]+L1;
   phat[3]=phat[2]+L2;

   pi=4.0*atan(1.0);
   r[0]=pi/(double)(2*N0);
   r[1]=pi/(double)(N1);
   r[2]=pi/(double)(N2);
   r[3]=pi/(double)(N3);

   l[0]=L0;
   l[1]=L1;
   l[2]=L2;
   l[3]=L3;

   bc=bc_type();

   for (mu=0;mu<4;mu++)
   {
      k0=cpr[mu]*l[mu];

      if ((bc==1)||(mu>0))
      {
         for (k=0;k<l[mu];k++)
            phat[mu][k]=2.0*sin((double)(k+k0)*r[mu]);
      }
      else
      {
         for (k=0;k<l[mu];k++)
            phat[mu][k]=2.0*sin(((double)(k+k0)+0.5)*r[mu]);
      }
   }

   k0=cpr[0]*L0;

   for (k=0;k<L0;k++)
   {
      if (bc==1)
      {
         p0=(double)(k+k0)*pi/(double)(N0);

         for (x0=0;x0<N0;x0++)
         {
            co[k][x0]=cos(p0*((double)(x0)+0.5));
            si[k][x0]=sin(p0*(double)(x0));
         }
      }
      else
      {
         p0=((double)(k+k0)+0.5)*pi/(double)(N0);

         for (x0=0;x0<N0;x0++)
         {
            co[k][x0]=cos(p0*(double)(x0));
            si[k][x0]=sin(p0*((double)(x0)+0.5));
         }
      }
   }

   init=1;
}


void cnrm2(double *cnrm)
{
   int bc,x0;
   int k0,k1,k2,k3;
   double r0,r1;

   if (init==0)
      set_phat();

   bc=bc_type();

   for (x0=0;x0<N0;x0++)
      loc_cnrm[x0]=0.0;

   for (k0=0;k0<L0;k0++)
   {
      for (k1=0;k1<L1;k1++)
      {
         for (k2=0;k2<L2;k2++)
         {
            for (k3=0;k3<L3;k3++)
            {
               if ((bc!=1)||(k0>0)||(cpr[0]>0))
               {
                  r0=phat[0][k0]*phat[0][k0]+
                     phat[1][k1]*phat[1][k1]+
                     phat[2][k2]*phat[2][k2]+
                     phat[3][k3]*phat[3][k3];

                  for (x0=0;x0<N0;x0++)
                  {
                     if (bc==1)
                        r1=si[k0][x0];
                     else
                        r1=co[k0][x0];

                     loc_cnrm[x0]+=(r1*r1)/r0;
                  }
               }
            }
         }
      }
   }

   r0=16.0/((double)(N0)*(double)(N1*N2*N3));

   for (x0=0;x0<N0;x0++)
      loc_cnrm[x0]*=r0;

   if (bc==1)
      loc_cnrm[N0]*=8.0*(double)(N0*N1)*(double)(N2*N3);
   else
      loc_cnrm[N0]*=16.0*(double)(N1*N2*N3)/(double)(N0);

   MPI_Reduce(loc_cnrm,cnrm,N0,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
   MPI_Bcast(cnrm,N0,MPI_DOUBLE,0,MPI_COMM_WORLD);

   if (bc==1)
   {
      r0=8.0*(double)(N0)/(double)(N1*N2*N3);

      for (x0=0;x0<N0;x0++)
      {
         r1=1.0-(double)(x0)/(double)(N0);
         cnrm[x0]+=r0*r1*r1;
      }
   }
}
