
/*******************************************************************************
*
* File ispt1.c
*
* Copyright (C) 2016 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* ISPT simulation and computation of the spatial and time-like parts of
* t^2*E, both using the symmetric and the plaquette expression for E.
*
* Syntax: ispt1 -i <filename> [-noloc] [-noexp] [-noms] [-a [-norng]]
*
* For usage instructions see the file README.ispt1.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "mpi.h"
#include "pmdb.h"
#include "utils.h"
#include "random.h"
#include "lattice.h"
#include "uflds.h"
#include "tupel.h"
#include "linalg.h"
#include "su3fcts.h"
#include "ftensor.h"
#include "ispt.h"
#include "wflow.h"
#include "archive.h"
#include "version.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)

static struct
{
   int ic;
   double (*mA)[2];
   complex_dble *E_sp,*E_tm;
} log_data;

static struct
{
   int ntm,order;
   double *tm;
} file_head;

static struct
{
   int ic;
   complex_dble **Eclov_sp,**Eclov_tm;
   complex_dble **Eplaq_sp,**Eplaq_tm;
} data;

static int my_rank,noloc,noexp,noms;
static int append,norng,endian;
static int level,seed,ncfg,idmp,rule;
static int opv[(int)(PM_TYPES)],ipgrd[2];
static double eps0,tol,*pv[(int)(PM_TYPES)],*step;
static su3_dble *Xw,*Yw;
static sl3_alg_dble *Aw;
static theory_t th;

static char theory[NAME_SIZE],line[NAME_SIZE];
static char nbase[NAME_SIZE];
static char log_dir[NAME_SIZE],dat_dir[NAME_SIZE];
static char loc_dir[NAME_SIZE],cnfg_dir[NAME_SIZE];
static char log_file[NAME_SIZE],log_save[NAME_SIZE];
static char dat_file[NAME_SIZE],dat_save[NAME_SIZE];
static char par_file[NAME_SIZE],par_save[NAME_SIZE];
static char msdat_file[NAME_SIZE],msdat_save[NAME_SIZE];
static char rng_file[NAME_SIZE],rng_save[NAME_SIZE];
static char cnfg_file[NAME_SIZE],end_file[NAME_SIZE];
static FILE *fin=NULL,*flog=NULL,*fdat=NULL,*fend=NULL;


static void alloc_log_data(void)
{
   int m;
   double (*mA)[2];
   complex_dble *E_sp;
   ispt_parms_t ispm;

   ispm=ispt_parms();
   m=ispm.order+1;

   mA=malloc(m*sizeof(*mA));
   E_sp=malloc(2*m*sizeof(*E_sp));

   error((mA==NULL)||(E_sp==NULL),1,"alloc_log_data [ispt1.c]",
         "Unable to allocate data arrays");

   log_data.ic=0;
   log_data.mA=mA;
   log_data.E_sp=E_sp;
   log_data.E_tm=E_sp+m;
}


static int write_log_data(void)
{
   int m,iw,io;
   stdint_t istd[1];
   double (*mA)[2],dstd[2];
   complex_dble *E_sp;
   ispt_parms_t ispm;

   ispm=ispt_parms();
   m=ispm.order+1;
   mA=log_data.mA;
   E_sp=log_data.E_sp;

   istd[0]=(stdint_t)(log_data.ic);

   if (endian==BIG_ENDIAN)
      bswap_int(1,istd);

   iw=fwrite(istd,sizeof(stdint_t),1,fdat);

   for (io=0;io<m;io++)
   {
      dstd[0]=mA[io][0];
      dstd[1]=mA[io][1];

      if (endian==BIG_ENDIAN)
         bswap_double(2,dstd);

      iw+=fwrite(dstd,sizeof(double),2,fdat);
   }

   for (io=0;io<(2*m);io++)
   {
      dstd[0]=E_sp[io].re;
      dstd[1]=E_sp[io].im;

      if (endian==BIG_ENDIAN)
         bswap_double(2,dstd);

      iw+=fwrite(dstd,sizeof(double),2,fdat);
   }

   if (iw!=(1+6*m))
      return 0;
   else
      return 1;
}


static int read_log_data(void)
{
   int m,ir,io;
   stdint_t istd[1];
   double (*mA)[2],dstd[2];
   complex_dble *E_sp;
   ispt_parms_t ispm;

   ispm=ispt_parms();
   m=ispm.order+1;
   mA=log_data.mA;
   E_sp=log_data.E_sp;

   ir=fread(istd,sizeof(stdint_t),1,fdat);

   if (ir!=1)
      return 0;

   if (endian==BIG_ENDIAN)
      bswap_int(1,istd);

   log_data.ic=(int)(istd[0]);

   for (io=0;io<m;io++)
   {
      ir+=fread(dstd,sizeof(double),2,fdat);

      if (endian==BIG_ENDIAN)
         bswap_double(2,dstd);

      mA[io][0]=dstd[0];
      mA[io][1]=dstd[1];
   }

   for (io=0;io<(2*m);io++)
   {
      ir+=fread(dstd,sizeof(double),2,fdat);

      if (endian==BIG_ENDIAN)
         bswap_double(2,dstd);

      E_sp[io].re=dstd[0];
      E_sp[io].im=dstd[1];
   }

   if (ir!=(1+6*m))
      return 0;
   else
      return 1;
}


static void alloc_data(void)
{
   int ntm,m,it;
   complex_dble **pp,*p;

   ntm=file_head.ntm;
   m=file_head.order+1;
   data.ic=0;

   if (ntm>0)
   {
      pp=malloc(4*ntm*sizeof(*pp));
      p=malloc(4*ntm*m*sizeof(*p));

      error((pp==NULL)||(p==NULL),1,"alloc_data [ispt1.c]",
            "Unable to allocate data arrays");

      data.Eclov_sp=pp;
      data.Eclov_tm=pp+ntm;
      data.Eplaq_sp=pp+2*ntm;
      data.Eplaq_tm=pp+3*ntm;

      for (it=0;it<(4*ntm);it++)
      {
         *pp=p;
         pp+=1;
         p+=m;
      }
   }
   else
   {
      data.Eclov_sp=NULL;
      data.Eclov_tm=NULL;
      data.Eplaq_sp=NULL;
      data.Eplaq_tm=NULL;
   }

   Aw=amalloc(m*sizeof(*Aw),4);
   Xw=amalloc(2*m*sizeof(*Xw),4);
   Yw=Xw+m;
   error((Aw==NULL)||(Xw==NULL),1,"alloc_data [ispt1.c]",
         "Unable to allocate auxiliary arrays");
}


static void write_file_head(void)
{
   int ntm,it,iw;
   stdint_t istd[2];
   double *tm,dstd[1];

   istd[0]=(stdint_t)(file_head.ntm);
   istd[1]=(stdint_t)(file_head.order);

   if (endian==BIG_ENDIAN)
      bswap_int(2,istd);

   iw=fwrite(istd,sizeof(stdint_t),2,fdat);

   ntm=file_head.ntm;
   tm=file_head.tm;

   for (it=0;it<ntm;it++)
   {
      dstd[0]=tm[it];

      if (endian==BIG_ENDIAN)
         bswap_double(1,dstd);

      iw+=fwrite(dstd,sizeof(double),1,fdat);
   }

   error_root(iw!=(2+ntm),1,"write_file_head [ispt1.c]",
              "Incorrect write count");
}


static void check_file_head(void)
{
   int ntm,it,ir,ie;
   stdint_t istd[2];
   double *tm,dstd[1];

   ir=fread(istd,sizeof(stdint_t),2,fdat);

   error_root(ir!=2,1,"check_file_head [ispt1.c]",
              "Incorrect read count");

   if (endian==BIG_ENDIAN)
      bswap_int(2,istd);

   error_root(((int)(istd[0])!=file_head.ntm)||
              ((int)(istd[1])!=file_head.order),1,"check_file_head [ispt1.c]",
              "Unexpected value of nt or order");

   ie=0;
   ntm=file_head.ntm;
   tm=file_head.tm;

   for (it=0;it<ntm;it++)
   {
      ir+=fread(dstd,sizeof(double),1,fdat);

      if (endian==BIG_ENDIAN)
         bswap_double(1,dstd);

      ie|=(dstd[0]!=tm[it]);
   }

   error_root(ir!=(2+ntm),1,"check_file_head [ispt1.c]",
              "Incorrect read count");

   error_root(ie!=0,1,"check_file_head [ispt1.c]",
              "Unexpected flow times");
}


static void write_data(void)
{
   int ntm,m,it,io,iw,i;
   stdint_t istd[1];
   double dstd[2];
   complex_dble *dat[4];

   ntm=file_head.ntm;
   m=file_head.order+1;
   istd[0]=data.ic;

   if (endian==BIG_ENDIAN)
      bswap_int(1,istd);

   iw=fwrite(istd,sizeof(stdint_t),1,fdat);

   for (it=0;it<ntm;it++)
   {
      dat[0]=data.Eclov_sp[it];
      dat[1]=data.Eclov_tm[it];
      dat[2]=data.Eplaq_sp[it];
      dat[3]=data.Eplaq_tm[it];

      for (i=0;i<4;i++)
      {
         for (io=0;io<m;io++)
         {
            dstd[0]=dat[i][io].re;
            dstd[1]=dat[i][io].im;

            if (endian==BIG_ENDIAN)
               bswap_double(2,dstd);

            iw+=fwrite(dstd,sizeof(double),2,fdat);
         }
      }
   }

   error_root(iw!=(1+8*ntm*m),1,"write_data [ispt1.c]",
              "Incorrect write count");
}


static int read_data(void)
{
   int ntm,m,it,io,ir,i;
   stdint_t istd[1];
   double dstd[2];
   complex_dble *dat[4];

   ir=fread(istd,sizeof(stdint_t),1,fdat);

   if (ir!=1)
      return 0;

   if (endian==BIG_ENDIAN)
      bswap_int(1,istd);

   data.ic=(int)(istd[0]);
   ntm=file_head.ntm;
   m=file_head.order+1;

   for (it=0;it<ntm;it++)
   {
      dat[0]=data.Eclov_sp[it];
      dat[1]=data.Eclov_tm[it];
      dat[2]=data.Eplaq_sp[it];
      dat[3]=data.Eplaq_tm[it];

      for (i=0;i<4;i++)
      {
         for (io=0;io<m;io++)
         {
            ir+=fread(dstd,sizeof(double),2,fdat);

            if (endian==BIG_ENDIAN)
               bswap_double(2,dstd);

            dat[i][io].re=dstd[0];
            dat[i][io].im=dstd[1];
         }
      }
   }

   error_root(ir!=(1+8*ntm*m),1,"read_data [ispt1.c]",
              "Read error or incomplete data record");
   return 1;
}


static void read_dirs(void)
{
   if (my_rank==0)
   {
      find_section("Run name");
      read_line("name","%s",nbase);

      find_section("Directories");
      read_line("log_dir","%s",log_dir);
      read_line("dat_dir","%s",dat_dir);

      if (noloc)
         loc_dir[0]='\0';
      else
         read_line("loc_dir","%s",loc_dir);

      if (noexp)
         cnfg_dir[0]='\0';
      else
         read_line("cnfg_dir","%s",cnfg_dir);
   }

   MPI_Bcast(nbase,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(log_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(dat_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(loc_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(cnfg_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
}


static void setup_files(void)
{
   if (noloc==0)
      check_dir(loc_dir);

   if (noexp==0)
      check_dir_root(cnfg_dir);

   check_dir_root(log_dir);
   check_dir_root(dat_dir);
   error_root(name_size("%s/%s.log~",log_dir,nbase)>=NAME_SIZE,1,
              "setup_files [ispt1.c]","log file name is too long");

   if (noms)
   {
      error_root(name_size("%s/%s.dat~",dat_dir,nbase)>=NAME_SIZE,1,
                 "setup_files [ispt1.c]","dat file name is too long");
      msdat_file[0]='\0';
      msdat_save[0]='\0';
   }
   else
   {
      error_root(name_size("%s/%s.ms.dat~",dat_dir,nbase)>=NAME_SIZE,1,
                 "setup_files [ispt1.c]","ms.dat file name is too long");
      sprintf(msdat_file,"%s/%s.ms.dat",dat_dir,nbase);
      sprintf(msdat_save,"%s~",msdat_file);
   }

   sprintf(log_file,"%s/%s.log",log_dir,nbase);
   sprintf(par_file,"%s/%s.par",dat_dir,nbase);
   sprintf(dat_file,"%s/%s.dat",dat_dir,nbase);
   sprintf(rng_file,"%s/%s.rng",dat_dir,nbase);
   sprintf(end_file,"%s/%s.end",log_dir,nbase);

   sprintf(log_save,"%s~",log_file);
   sprintf(par_save,"%s~",par_file);
   sprintf(dat_save,"%s~",dat_file);
   sprintf(rng_save,"%s~",rng_file);
}


static void set_th(void)
{
   if (strcmp("PHI4",theory)==0)
      th=PHI4;
   else if (strcmp("YM",theory)==0)
      th=YM;
   else if (strcmp("QCD",theory)==0)
      th=QCD;
   else if (strcmp("LYM",theory)==0)
      th=LYM;
   else if (strcmp("LQCD",theory)==0)
      th=LQCD;
   else
      error(1,1,"set_th [ispt1.c]","Unknown theory");
}


static void default_pv(int order)
{
   int pmax,m,i;
   double *p;

   pmax=(int)(PM_TYPES);
   m=order+1;
   p=malloc(pmax*m*sizeof(*p));
   error(p==NULL,1,"default_pv [ispt1.c]",
         "Unable to allocate auxiliary array");

   for (i=0;i<(pmax*m);i++)
      p[i]=0.0;

   for (i=0;i<pmax;i++)
   {
      opv[i]=order;
      pv[i]=p;
      p+=m;
   }

   pv[(int)(C0)][0]=1.0;
   pv[(int)(LAMBDA)][0]=1.0;
   pv[(int)(MASS)][0]=0.0;
   pv[(int)(CG)][0]=1.0;
   pv[(int)(CGP)][0]=1.0;
   pv[(int)(CSW)][0]=1.0;
   pv[(int)(CF)][0]=1.0;
   pv[(int)(CFP)][0]=1.0;
}


static void read_theory_parms(void)
{
   int bc,order,iroot;
   double lambda;

   if (my_rank==0)
   {
      find_section("Theory");
      read_line("theory","%s",theory);
      read_line("bc","%d",&bc);

      find_section("ISPT");
      read_line("order","%d",&order);
      read_line("iroot","%d",&iroot);
      read_line("lambda","%lf",&lambda);
      read_line("ncfg","%d",&ncfg);
   }

   MPI_Bcast(theory,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&order,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&iroot,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&lambda,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncfg,1,MPI_INT,0,MPI_COMM_WORLD);

   set_th();
   error_root((th!=YM)&&(th!=LYM),1,"read_theory_parms [ispt1.c]",
              "Currently only the YM and LYM theories are supported");
   error_root(ncfg<=0,1,"read_theory_parms [ispt1.c]",
              "Number of configurations ncfg must be postive");

   set_th_parms(th,bc,0);
   default_pv(order);
   pv[(int)(LAMBDA)][0]=lambda;
   set_ispt_parms(order,iroot,opv,pv);
   file_head.order=order;

   if (append)
   {
      check_th_parms(fdat);
      check_ispt_parms(fdat);
   }
   else
   {
      write_th_parms(fdat);
      write_ispt_parms(fdat);
   }
}


static void read_flow_parms(void)
{
   int ntm,it,ir,iw,ie;
   stdint_t istd[3];
   double dstd[2],*tm;

   if (noms)
   {
      rule=0;
      idmp=0;
      eps0=0.0;
      tol=0.0;
      ntm=0;
      tm=NULL;
   }
   else
   {
      if (my_rank==0)
      {
         find_section("Gradient flow");
         read_line("integrator","%s",line);
         read_line("eps0","%lf",&eps0);
         read_line("tol","%lf",&tol);
         read_line("damp","%d",&idmp);
         ntm=count_tokens("times");

         if (strcmp(line,"EULER")==0)
            rule=1;
         else if (strcmp(line,"RK2")==0)
            rule=2;
         else if (strcmp(line,"RK3")==0)
            rule=3;
         else
            error_root(1,1,"read_flow_parms [ispt1.c]","Unknown integrator");

         error_root((eps0<=0.0)||(tol<=0.0),1,"read_flow_parms [ispt1.c]",
                    "Parameters eps0 and tol must be positive");
         error_root((idmp<0)||(idmp>1),1,"read_flow_parms [ispt1.c]",
                    "Parameter damp must be 0 or 1");
         error_root(ntm==0,1,"read_flow_parms [ispt1.c]",
                    "No flow times specified");
      }

      MPI_Bcast(&rule,1,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(&idmp,1,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(&ntm,1,MPI_INT,0,MPI_COMM_WORLD);

      MPI_Bcast(&eps0,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&tol,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

      tm=malloc(2*ntm*sizeof(*tm));
      step=tm+ntm;
      error(tm==NULL,1,"read_flow_parms [ispt1.c]",
            "Unable to allocate times array");

      if (my_rank==0)
      {
         read_dprms("times",ntm,tm);
         ie=(tm[0]<0.0);

         for (it=1;it<ntm;it++)
            ie|=(tm[it]<tm[it-1]);

         error_root(ie!=0,1,"read_flow_parms [ispt1.c]",
                    "Negative or not properly ordered flow times");
      }

      MPI_Bcast(tm,ntm,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }

   file_head.ntm=ntm;
   file_head.tm=tm;

   if (my_rank==0)
   {
      if (append)
      {
         ir=fread(istd,sizeof(stdint_t),3,fdat);
         ir+=fread(dstd,sizeof(double),2,fdat);
         error_root(ir!=5,1,"read_flow_parms [ispt1.c]",
                    "Incorrect read count (parameter file)");

         if (endian==BIG_ENDIAN)
         {
            bswap_int(3,istd);
            bswap_double(2,dstd);
         }

         ie=0;
         ie|=((int)(istd[0])!=ntm);
         ie|=((int)(istd[1])!=idmp);
         ie|=((int)(istd[2])!=rule);
         ie|=(dstd[0]!=eps0);
         ie|=(dstd[1]!=tol);

         error_root(ie!=0,1,"read_flow_parms [ispt1.c]",
                    "Parameters do not match previous run");
      }
      else
      {
         istd[0]=(stdint_t)(ntm);
         istd[1]=(stdint_t)(idmp);
         istd[2]=(stdint_t)(rule);
         dstd[0]=eps0;
         dstd[1]=tol;

         if (endian==BIG_ENDIAN)
         {
            bswap_int(3,istd);
            bswap_double(2,dstd);
         }

         iw=fwrite(istd,sizeof(stdint_t),3,fdat);
         iw+=fwrite(dstd,sizeof(double),2,fdat);
         error_root(iw!=5,1,"read_flow_parms [ispt1.c]",
                    "Incorrect write count (parameter file)");
      }
   }
}


static void read_infile(int argc,char *argv[])
{
   int ifile;

   if (my_rank==0)
   {
      flog=freopen("STARTUP_ERROR","w",stdout);

      ifile=find_opt(argc,argv,"-i");
      noloc=find_opt(argc,argv,"-noloc");
      noexp=find_opt(argc,argv,"-noexp");
      noms=find_opt(argc,argv,"-noms");
      append=find_opt(argc,argv,"-a");
      norng=find_opt(argc,argv,"-norng");
      endian=endianness();

      error_root((ifile==0)||(ifile==(argc-1)),1,"read_infile [ispt1.c]",
                 "Syntax: ispt1 -i <filename> [-noloc] [-noexp] [-noms] "
                 "[-a [-norng]]");

      error_root(endian==UNKNOWN_ENDIAN,1,"read_infile [ispt1.c]",
                 "Machine has unknown endianness");

      fin=freopen(argv[ifile+1],"r",stdin);
      error_root(fin==NULL,1,"read_infile [ispt1.c]",
                 "Unable to open input file");
   }

   MPI_Bcast(&noloc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&noexp,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&noms,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&append,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&norng,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&endian,1,MPI_INT,0,MPI_COMM_WORLD);

   if ((append==0)||(norng))
   {
      if (my_rank==0)
      {
         find_section("Random number generator");
         read_line("level","%d",&level);
         read_line("seed","%d",&seed);
      }

      MPI_Bcast(&level,1,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(&seed,1,MPI_INT,0,MPI_COMM_WORLD);
   }
   else
   {
      level=0;
      seed=0;
   }

   read_dirs();
   setup_files();

   if (my_rank==0)
   {
      if (append)
         fdat=fopen(par_file,"rb");
      else
         fdat=fopen(par_file,"wb");

      error_root(fdat==NULL,1,"read_infile [ispt1.c]",
                 "Unable to open parameter file");
   }

   read_theory_parms();
   read_flow_parms();

   if (my_rank==0)
   {
      fclose(fin);
      fclose(fdat);

      if (append==0)
         copy_file(par_file,par_save);
   }
}


static void check_old_log(int *last_ic)
{
   int ir,isv,ips;
   int pseed,np[4],bp[4];

   fend=fopen(log_file,"r");
   error_root(fend==NULL,1,"check_old_log [ispt1.c]",
              "Unable to open log file");
   (*last_ic)=0;
   ir=1;
   isv=0;
   ips=0;

   while (fgets(line,NAME_SIZE,fend)!=NULL)
   {
      ir=0;

      if (strstr(line,"process grid")!=NULL)
      {
         ir&=(sscanf(line,"%dx%dx%dx%d process grid, %dx%dx%dx%d",
                     np,np+1,np+2,np+3,bp,bp+1,bp+2,bp+3)==8);

         ipgrd[0]=((np[0]!=NPROC0)||(np[1]!=NPROC1)||
                   (np[2]!=NPROC2)||(np[3]!=NPROC3));
         ipgrd[1]=((bp[0]!=NPROC0_BLK)||(bp[1]!=NPROC1_BLK)||
                   (bp[2]!=NPROC2_BLK)||(bp[3]!=NPROC3_BLK));
      }
      else if (strstr(line,"seed =")!=NULL)
      {
         if (norng)
         {
            ir&=(sscanf(line,"level = %*d, seed = %d",&pseed)==1);
            ips|=(seed==pseed);
         }
      }
      else if (strstr(line,"max|A[")!=NULL)
         isv=0;
      else if (strstr(line,"fully processed")!=NULL)
      {
         ir&=(sscanf(line,"Configuration no %d",last_ic)==1);
         isv=1;
      }
   }

   fclose(fend);

   error_root(ir!=1,1,"check_old_log [ispt1.c]","Incorrect read count");

   error_root((ipgrd[0]==1)&&(norng==0),1,"check_old_log [ispt1.c]",
              "Continuation run:\n"
              "Process grid changed but -norng is not set");

   error_root(ips!=0,1,"check_old_log [ispt1.c]",
              "Continuation run:\n"
              "Attempt to reuse old seed for the random number generator");

   error_root((*last_ic)==0,1,"check_old_log [ispt1.c]",
              "Continuation run:\n"
              "No fully processed configuration in the preceding run");

   error_root(isv==0,1,"check_old_log [ispt1.c]",
              "Continuation run:\n"
              "Log entries after the last fully processed configuration");
}


static void check_old_dat(int *last_ic)
{
   int ic,ir,nic,pic;

   fdat=fopen(dat_file,"rb");
   error_root(fdat==NULL,1,"check_old_dat [ispt1.c]",
              "Unable to open data file");
   nic=0;
   pic=0;

   for (ic=0;;ic++)
   {
      ir=read_log_data();

      if (ir==0)
      {
         error_root(ic==0,1,"check_old_dat [ispt1.c]",
                    "No data records found");
         break;
      }

      nic=log_data.ic;

      if (ic==0)
         error_root(nic!=1,1,"check_old_dat [ispt1.c]",
                    "Configuration range does not start with 1");
      else
         error_root(nic!=(pic+1),1,"check_old_dat [ispt1.c]",
                    "Configuration sequence is not contiguous");

      pic=nic;
   }

   fclose(fdat);
   (*last_ic)=nic;
}


static void check_old_msdat(int *last_ic)
{
   int ic,ir,nic,pic;

   fdat=fopen(msdat_file,"rb");
   error_root(fdat==NULL,1,"check_old_msdat [ispt1.c]",
              "Unable to open data file");
   check_file_head();
   nic=0;
   pic=0;

   for (ic=0;;ic++)
   {
      ir=read_data();

      if (ir==0)
      {
         error_root(ic==0,1,"check_old_msdat [ispt1.c]",
                    "No data records found");
         break;
      }

      nic=data.ic;

      if (ic==0)
         error_root(nic!=1,1,"check_old_msdat [ispt1.c]",
                    "Configuration range does not start with 1");
      else
         error_root(nic!=(pic+1),1,"check_old_msdat [ispt1.c]",
                    "Configuration sequence is not contiguous");

      pic=nic;
   }

   fclose(fdat);
   (*last_ic)=nic;
}


static void check_files(int *last_ic)
{
   int ic1,ic2;

   ipgrd[0]=0;
   ipgrd[1]=0;

   if (my_rank==0)
   {
      if (append)
      {
         check_old_log(&ic1);
         check_old_dat(&ic2);
         error_root(ic1!=ic2,1,"check_files [ispt1.c]",
                    "Inconsistent configuration ranges");

         if (noms==0)
         {
            check_old_msdat(&ic2);
            error_root(ic1!=ic2,2,"check_files [ispt1.c]",
                       "Inconsistent configuration ranges");
         }

         (*last_ic)=ic1;
      }
      else
      {
         fin=fopen(log_file,"r");
         error_root(fin!=NULL,1,"check_files [ispt1.c]",
                    "Attempt to overwrite old *.log file");

         fin=fopen(dat_file,"r");
         error_root(fin!=NULL,1,"check_files [ispt1.c]",
                    "Attempt to overwrite old *.dat file");

         if (noms==0)
         {
            fin=fopen(msdat_file,"rb");
            error_root(fin!=NULL,1,"check_files [ispt1.c]",
                       "Attempt to overwrite old *.ms.dat file");

            fdat=fopen(msdat_file,"wb");
            error_root(fdat==NULL,1,"check_files [ispt1.c]",
                       "Unable to open measurement data file");
            write_file_head();
            fclose(fdat);
         }

         (*last_ic)=0;
      }
   }

   MPI_Bcast(ipgrd,2,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(last_ic,1,MPI_INT,0,MPI_COMM_WORLD);
}


static void init_rng(int last_ic)
{
   int ic;

   if (append)
   {
      if (norng)
         start_ranlux(level,seed);
      else
      {
         ic=import_ranlux(rng_file);
         error_root(ic!=last_ic,1,"init_rng [ispt1.c]",
                    "Configuration number mismatch");
      }
   }
   else
      start_ranlux(level,seed);
}


static void print_info(void)
{
   int n,i;
   long ip;

   if (my_rank==0)
   {
      ip=ftell(flog);
      fclose(flog);

      if (ip==0L)
         remove("STARTUP_ERROR");

      if (append)
         flog=freopen(log_file,"a",stdout);
      else
         flog=freopen(log_file,"w",stdout);

      error_root(flog==NULL,1,"print_info [ispt1.c]","Unable to open log file");

      if (append)
      {
         printf("Continuation run\n");
         printf("----------------\n\n");
      }
      else
      {
         printf("\n");
         printf("New ISPT run\n");
         printf("------------\n\n");
      }

      printf("Program version %s\n",ISPT_RELEASE);

      if (endian==LITTLE_ENDIAN)
         printf("The machine is little endian\n\n");
      else
         printf("The machine is big endian\n\n");

      if ((ipgrd[0]!=0)&&(ipgrd[1]!=0))
         printf("Process grid and process block size changed:\n");
      else if (ipgrd[0]!=0)
         printf("Process grid changed:\n");
      else if (ipgrd[1]!=0)
         printf("Process block size changed:\n");

      if ((append==0)||(ipgrd[0]!=0)||(ipgrd[1]!=0))
      {
         printf("%dx%dx%dx%d lattice, ",N0,N1,N2,N3);
         printf("%dx%dx%dx%d local lattice\n",L0,L1,L2,L3);
         printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
         printf("%dx%dx%dx%d process block size\n\n",
                NPROC0_BLK,NPROC1_BLK,NPROC2_BLK,NPROC3_BLK);
      }

      if (append==0)
      {
         print_th_parms();
         print_ispt_parms();
      }

      printf("Random number generator:\n");

      if (append)
      {
         if (norng)
            printf("level = %d, seed = %d\n\n",level,seed);
         else
         {
            printf("State of ranlxs and ranlxd reset to the\n");
            printf("last exported state\n\n");
         }
      }
      else
      {
         printf("level = %d, seed = %d\n\n",level,seed);

         if (noms==0)
         {
            printf("Gradient flow:\n");
            if (rule==1)
               printf("Euler integrator\n");
            else if (rule==2)
               printf("2nd order RK integrator\n");
            else
               printf("3rd order RK integrator\n");

            if (idmp)
               printf("Gauge-damping term switched on\n");
            else
               printf("Gauge-damping term switched off\n");

            n=fdigits(eps0);
            printf("eps0 = %.2e\n",eps0);
            printf("tol = %.2e\n",tol);
            printf("Flow times =");

            for (i=0;i<file_head.ntm;i++)
            {
               n=fdigits(file_head.tm[i]);
               printf(" %.*f",IMAX(n,1),file_head.tm[i]);
            }

            printf("\n\n");
         }
         else
            printf("No measurements are made\n\n");
      }

      printf("Number of configurations to be generated = %d\n",ncfg);

      if ((noloc)&&(noexp))
         printf("Configurations are not saved\n\n");
      else if (noexp)
         printf("Configurations are written to the local disks\n\n");
      else if (noloc)
         printf("Configurations are exported from process 0\n\n");
      else
         printf("Configurations are written to the local disks and\n"
                "exported from process 0\n\n");
      fflush(flog);
   }
}


static void save_flds(int icnfg,sl3_alg_dble **ad)
{
   int order;
   ispt_parms_t ispm;

   if ((noloc==0)||(noexp==0))
   {
      ispm=ispt_parms();
      order=ispm.order;

      if (noloc==0)
      {
         sprintf(cnfg_file,"%s/%sn%d_%d",loc_dir,nbase,icnfg,my_rank);
         write_afld(cnfg_file,order,ad);
      }

      if (noexp==0)
      {
         sprintf(cnfg_file,"%s/%sn%d",cnfg_dir,nbase,icnfg);
         export_afld(cnfg_file,order,ad);
      }

      if (my_rank==0)
         printf("Configuration no %d saved\n\n",icnfg);
   }
}


static void max_A(int i,sl3_alg_dble *A)
{
   double d,(*mA)[2];
   complex_dble z;

   mA=log_data.mA;

   z.re=(*A).c1.re+(*A).c2.re;
   z.im=(*A).c1.im+(*A).c2.im;
   d=z.re*z.re+z.im*z.im;

   d+=(*A).c1.re*(*A).c1.re+(*A).c1.im*(*A).c1.im;
   d+=(*A).c2.re*(*A).c2.re+(*A).c2.im*(*A).c2.im;
   d+=(*A).c3.re*(*A).c3.re+(*A).c3.im*(*A).c3.im;
   d+=(*A).c4.re*(*A).c4.re+(*A).c4.im*(*A).c4.im;
   d+=(*A).c5.re*(*A).c5.re+(*A).c5.im*(*A).c5.im;
   d+=(*A).c6.re*(*A).c6.re+(*A).c6.im*(*A).c6.im;
   d+=(*A).c7.re*(*A).c7.re+(*A).c7.im*(*A).c7.im;
   d+=(*A).c8.re*(*A).c8.re+(*A).c8.im*(*A).c8.im;

   d=sqrt(d);

   if (d>mA[i][0])
      mA[i][0]=d;

   mA[i][1]+=d;
}


static void set_mA(sl3_alg_dble **ad)
{
   int bc,ix,t,ifc;
   int order,io;
   double r,d[2],(*mA)[2];
   sl3_alg_dble *A;
   ispt_parms_t ispm;

   bc=bc_type();
   ispm=ispt_parms();
   order=ispm.order;
   mA=log_data.mA;

   if (bc==1)
      r=1.0/((double)(4*N0-3)*(double)(N1*N2*N3));
   else
      r=1.0/((double)(4*N0)*(double)(N1*N2*N3));

   for (io=0;io<=order;io++)
   {
      mA[io][0]=0.0;
      mA[io][1]=0.0;
      A=ad[io];

      for (ix=(VOLUME/2);ix<VOLUME;ix++)
      {
         t=global_time(ix);

         for (ifc=0;ifc<8;ifc++)
         {
            if ((t>0)||(ifc<2)||(bc!=1))
               max_A(io,A);

            A+=1;
         }
      }

      if (NPROC>1)
      {
         d[0]=mA[io][0];
         d[1]=mA[io][1];
         MPI_Reduce(d,mA[io],1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
         MPI_Reduce(d+1,mA[io]+1,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      }

      mA[io][1]*=r;
   }

   MPI_Bcast(mA[0],2*(order+1),MPI_DOUBLE,0,MPI_COMM_WORLD);
}


static void set_log_data(su3_dble *ud)
{
   int order;
   ispt_parms_t ispm;

   ispm=ispt_parms();
   order=ispm.order;
   Eplaq(order,ud,log_data.E_sp,log_data.E_tm);
}


static void save_log_data(void)
{
   int iw;

   if (my_rank==0)
   {
      fdat=fopen(dat_file,"ab");
      error_root(fdat==NULL,1,"save_log_data [ispt1.c]",
                 "Unable to open data file");

      iw=write_log_data();
      error_root(iw!=1,1,"save_log_data [ispt1.c]",
                 "Incorrect write count");
      fclose(fdat);
   }
}


static double dev_sl3(su3_dble *ud,su3_dble *vd)
{
   int m,i;
   double d,dmax;
   su3_dble *um;

   m=file_head.order+1;
   dmax=0.0;
   um=ud+4*VOLUME*m;

   for (;ud<um;ud+=m)
   {
      cm3x3_inv_poly(m,vd,Xw);
      cm3x3_varmul_poly(m,ud,Xw,Yw);
      cm3x3_log_poly(m,Yw,Xw);

      for (i=0;i<m;i++)
         su3_to_sl3(Xw+i,Aw+i);

      d=norm_square_calg(m,0,Aw);
      if (d>dmax)
         dmax=d;

      vd+=m;
   }

   if (NPROC>1)
   {
      d=dmax;
      MPI_Reduce(&d,&dmax,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
      MPI_Bcast(&dmax,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }

   return sqrt(dmax);
}


static double new_eps(double eps,su3_dble *ud,su3_dble *vd)
{
   int order;
   double del;

   order=file_head.order;
   cm3x3_assign(4*VOLUME*(order+1),ud,vd);

   if (rule==1)
   {
      fwd_euler(1,2.0*eps,idmp,order,vd);
      fwd_euler(2,eps,idmp,order,ud);
   }
   else if (rule==2)
   {
      fwd_rk2(1,2.0*eps,idmp,order,vd);
      fwd_rk2(2,eps,idmp,order,ud);
   }
   else
   {
      fwd_rk3(1,2.0*eps,idmp,order,vd);
      fwd_rk3(2,eps,idmp,order,ud);
   }

   del=dev_sl3(ud,vd);

#ifdef FLOW_DBG
   if (my_rank==0)
      printf("FLOW_DBG: eps=%.1e, del=%.1e, ",eps,del);
#endif

   if (rule==1)
   {
      del*=(10.0/2.0);
      eps*=pow(tol/del,1.0/2.0);
   }
   else if (rule==2)
   {
      del*=(10.0/6.0);
      eps*=pow(tol/del,1.0/3.0);
   }
   else
   {
      del*=(10.0/14.0);
      eps*=pow(tol/del,1.0/4.0);
   }

#ifdef FLOW_DBG
   if (my_rank==0)
   {
      printf("new eps=%.1e",eps);

      if (eps<0.1)
         printf("\n");
      else
         printf(" (reset to 0.1)\n");
   }
#endif

   if (eps<0.1)
      return eps;
   else
      return 0.1;
}


static double flow(double eps,double t0,double t1,su3_dble *ud,su3_dble *vd)
{
   int ns,order;
   double dt,del;

   order=file_head.order;
   dt=t1-t0;

   while (dt>0.0)
   {
      ns=(int)(dt/eps);

      if (ns>=14)
      {
         ns=10;
         del=eps;
         dt-=10.0*del;
      }
      else
      {
         ns+=1;
         if (ns<2)
            ns=2;
         del=dt/(double)(ns);
         dt=0.0;
      }

      if (ns>2)
      {
         if (rule==1)
            fwd_euler(ns-2,del,idmp,order,ud);
         else if (rule==2)
            fwd_rk2(ns-2,del,idmp,order,ud);
         else
            fwd_rk3(ns-2,del,idmp,order,ud);
      }

#ifdef FLOW_DBG
      if (my_rank==0)
      {
         printf("FLOW_DBG: No of steps = %d, step size = %.2e\n",ns,del);
         t0+=(double)(ns)*del;
      }
#endif

      eps=new_eps(del,ud,vd);
   }

#ifdef FLOW_DBG
   if (my_rank==0)
      printf("FLOW_DBG: Integration time check = %.1e\n",fabs(t1-t0));
#endif

   return eps;
}


static void set_data(su3_dble *ud)
{
   int ntm,order,itm,io;
   double *tm,eps,t0,t1,tsq;
   complex_dble **Eclov_sp,**Eclov_tm;
   complex_dble **Eplaq_sp,**Eplaq_tm;
   su3_dble **wud,*vd,*wd;

   ntm=file_head.ntm;
   order=file_head.order;
   tm=file_head.tm;
   Eclov_sp=data.Eclov_sp;
   Eclov_tm=data.Eclov_tm;
   Eplaq_sp=data.Eplaq_sp;
   Eplaq_tm=data.Eplaq_tm;

   wud=reserve_wud(2*(order+1));
   vd=wud[0];
   wd=wud[order+1];
   cm3x3_assign(4*VOLUME*(order+1),ud,vd);
   eps=new_eps(eps0,vd,wd);

   for (itm=0;itm<ntm;itm++)
   {
      if (itm==0)
      {
         t0=0.0;
         t1=tm[0];
      }
      else
      {
         t0=tm[itm-1];
         t1=tm[itm];
      }

      step[itm]=eps;
      eps=flow(eps,t0,t1,ud,vd);
      Eclov(order,ud,Eclov_sp[itm],Eclov_tm[itm]);
      Eplaq(order,ud,Eplaq_sp[itm],Eplaq_tm[itm]);
      tsq=tm[itm]*tm[itm];

      for (io=0;io<=order;io++)
      {
         Eclov_sp[itm][io].re*=tsq;
         Eclov_sp[itm][io].im*=tsq;
         Eclov_tm[itm][io].re*=tsq;
         Eclov_tm[itm][io].im*=tsq;

         Eplaq_sp[itm][io].re*=tsq;
         Eplaq_sp[itm][io].im*=tsq;
         Eplaq_tm[itm][io].re*=tsq;
         Eplaq_tm[itm][io].im*=tsq;
      }
   }

   release_wud();
}


static void save_msdat(void)
{
   if (my_rank==0)
   {
      fdat=fopen(msdat_file,"ab");
      error_root(fdat==NULL,1,"save_msdat [ispt1.c]",
                 "Unable to open data file");
      write_data();
      fclose(fdat);
   }
}


static void print_log(int icfg,double *wdt,double *wtav)
{
   int order,io;
   int ntm,itm,n;
   double *tm,(*mA)[2];
   complex_dble *E_sp,*E_tm;
   ispt_parms_t ispm;

   if (my_rank==0)
   {
      ispm=ispt_parms();
      order=ispm.order;
      mA=log_data.mA;
      E_sp=log_data.E_sp;
      E_tm=log_data.E_tm;

      for (io=0;io<=order;io++)
      {
         printf("max|A[%d]| = %.1e, ",io,mA[io][0]);
         printf("avg|A[%d]| = %.1e, ",io,mA[io][1]);
         printf("E_sp[%d] = (% .6e,% .6e), ",io,E_sp[io].re,E_sp[io].im);
         printf("E_tm[%d] = (% .6e,% .6e)\n",io,E_tm[io].re,E_tm[io].im);
      }

      if (noms==0)
      {
         printf("\n");
         printf("Gradient-flow integration step sizes (flow time):\n");
         ntm=file_head.ntm;
         tm=file_head.tm;

         for (itm=0;itm<ntm;itm++)
         {
            if (itm>0)
            {
               n=fdigits(tm[itm-1]);
               printf(" %.1e  (%.*f)\n",step[itm],IMAX(n,1),tm[itm-1]);
            }
            else
               printf(" %.1e  (0.0)\n",step[itm]);
         }
      }

      printf("Time (configuration) = %.1e sec (average %.1e sec)\n",
             wdt[0],wtav[0]/(double)(icfg));
      printf("Time (measurement)   = %.1e sec (average %.1e sec)\n\n",
             wdt[1],wtav[1]/(double)(icfg));
      printf("Configuration no %d fully processed\n\n",log_data.ic);
      fflush(flog);
   }
}


static void check_endflag(int *iend)
{
   if (my_rank==0)
   {
      fend=fopen(end_file,"r");

      if (fend!=NULL)
      {
         fclose(fend);
         remove(end_file);
         (*iend)=1;
         printf("End flag set, run stopped\n\n");
      }
      else
         (*iend)=0;
   }

   MPI_Bcast(iend,1,MPI_INT,0,MPI_COMM_WORLD);
}


int main(int argc,char *argv[])
{
   int last_ic,icfg,order,iend;
   int i,k,icl,ncl,nv,nw[4];
   double wt1,wt2,wdt[2],wtav[2];
   sl3_alg_dble **wad;
   su3_dble **wud;
   tupel_t *tp;
   rcls_t *rc;
   ispt_parms_t ispm;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   read_infile(argc,argv);
   alloc_log_data();
   alloc_data();
   check_files(&last_ic);
   print_info();
   init_rng(last_ic);
   geometry();

   for (i=0;i<4;i++)
      nw[i]=0;

   ispm=ispt_parms();
   order=ispm.order;
   nw[1]=3*(order+1);
   nw[2]=order+1;
   init_ispt(nw);

   if (noms)
      alloc_wud(1*(order+1));
   else
      alloc_wud(3*(order+1));

   rc=rtree_classes(th);
   ncl=rc[0].cls[0].n;
   wtav[0]=0.0;
   wtav[1]=0.0;
   iend=0;

   for (icfg=1;(iend==0)&&(icfg<=ncfg);icfg++)
   {
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();

      wad=reserve_wad(order+1);
      generate_rflds(wad[0]);

      for (icl=0;icl<ncl;icl++)
      {
         tp=rc[0].cls[0].tpa[icl];
         nv=0;

         for (i=1;i<(*tp).n;i++)
            nv|=((*tp).i[i]!=0);

         if (nv==0)
         {
            k=(*tp).i[0];
            ispt(icl,wad[k]);
         }
      }

      save_flds(last_ic+icfg,wad);
      set_mA(wad);
      wud=reserve_wud(order+1);
      a2uflds(order,wad,wud[0]);
      release_wad();

      MPI_Barrier(MPI_COMM_WORLD);
      wt2=MPI_Wtime();
      wdt[0]=wt2-wt1;
      wtav[0]+=wdt[0];
      wt1=MPI_Wtime();

      log_data.ic=last_ic+icfg;
      data.ic=log_data.ic;
      set_log_data(wud[0]);
      if (noms==0)
         set_data(wud[0]);
      release_wud();

      MPI_Barrier(MPI_COMM_WORLD);
      wt2=MPI_Wtime();
      wdt[1]=wt2-wt1;
      wtav[1]+=wdt[1];

      save_log_data();
      if (noms==0)
         save_msdat();
      print_log(icfg,wdt,wtav);
      export_ranlux(log_data.ic,rng_file);
      check_endflag(&iend);

      if (my_rank==0)
      {
         fflush(flog);
         copy_file(log_file,log_save);
         copy_file(dat_file,dat_save);
         if (noms==0)
            copy_file(msdat_file,msdat_save);
         copy_file(rng_file,rng_save);
      }
   }

   if (my_rank==0)
      fclose(flog);

   MPI_Finalize();
   exit(0);
}
