
/*******************************************************************************
*
* File adcom.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Communication of the gluon field variables at the exterior boundaries
* of the local lattices.
*
*   void copy_bnd_ad(int n,sl3_alg_dble *ad)
*     Copies the type (1) and type (2) field variables on the +0,..,+3 faces
*     from the neighbouring MPI processes (see README.aflds). The meaning of
*     the parameter n>=0 is explained in the notes.
*
*   void add_bnd_ad(int n,sl3_alg_dble *ad)
*     Adds the type (1) and type (2) field variables on the +0,..,+3 faces
*     to the field on the neighbouring MPI processes (see README.aflds).
*     The meaning of the parameter n>=0 is explained in the notes.
*
*   void gather_bnd_ad0(sl3_alg_dble *ad)
*     Copies the type (1) field variables on the +0,..,+3 faces from the
*     neighbouring MPI processes (see README.aflds).
*
*   void scatter_bnd_ad0(sl3_alg_dble *ad)
*     The operation performed by this program is the inverse of the one
*     performed by gather_bnd_ad0().
*
* In all these programs, the gauge-field array ad is assumed to contain n+1
* gauge potentials of the kind returned by reserve_wad(n+1), where n=0 in
* the case of the gather* and scatter* programs (see utils/wspace.c). More
* precisely, the fields are expected to be arranged in memory such that the
* field variables on the link number l are ad[l*(n+1)+k], where k=0,..,n is
* the field index.
*
* The field variables of type (2) are not copied across the boundary of
* the lattice in the time direction, nor are they added to the field on
* neighbouring MPI processes across that boundary.
*
* The programs in this module perform global communications and must be
* called simultaneously on all MPI processes.
*
*******************************************************************************/

#define ADCOM_C

#include <stdlib.h>
#include <stdio.h>
#include "mpi.h"
#include "sl3.h"
#include "linalg.h"
#include "utils.h"
#include "aflds.h"
#include "global.h"

static int ms=0,np,tag0[8],tag1[8],io[8];
static sl3_alg_dble *buf,*sbuf,*rbuf;
static aidx_t *idx;


static void alloc_buf(int n)
{
   int m,ifc,mu,nbf;

   idx=aidx();
   nbf=0;

   for (mu=0;mu<4;mu++)
   {
      if (nbf<idx[mu].nak)
         nbf=idx[mu].nak;
   }

   if (ms>0)
      afree(buf);

   m=n+1;
   buf=amalloc(m*nbf*sizeof(*buf),ALIGN);
   error(buf==NULL,1,"alloc_buf [adcom.c]",
         "Unable to allocate communication buffer");

   if (ms==0)
   {
      np=(cpr[0]+cpr[1]+cpr[2]+cpr[3])&0x1;

      for (ifc=0;ifc<8;ifc++)
      {
         tag0[ifc]=mpi_permanent_tag();
         tag1[ifc]=mpi_permanent_tag();
         io[ifc]=1;
      }

      io[0]=(cpr[0]!=0);
      io[1]=(cpr[0]!=(NPROC0-1));
   }

   ms=m;
}


static void pack_ad0(int n,int mu,sl3_alg_dble *ad)
{
   int m,*ia,*iam;
   sl3_alg_dble *X;

   m=n+1;
   ia=idx[mu].ia0;
   iam=ia+idx[mu].na0;
   X=buf;

   for (;ia<iam;ia++)
   {
      assign_calg2calg(m,ad+m*ia[0],X);
      X+=m;
   }
}


static void unpack_ad0(int n,int mu,sl3_alg_dble *ad)
{
   int m,*ia,*iam;
   sl3_alg_dble *X;

   m=n+1;
   ia=idx[mu].ia0;
   iam=ia+idx[mu].na0;
   X=buf;

   for (;ia<iam;ia++)
   {
      assign_calg2calg(m,X,ad+m*ia[0]);
      X+=m;
   }
}


static void add_ad0(int n,int mu,sl3_alg_dble *ad)
{
   int m,*ia,*iam;
   sl3_alg_dble *X;

   m=n+1;
   ia=idx[mu].ia0;
   iam=ia+idx[mu].na0;
   X=buf;

   for (;ia<iam;ia++)
   {
      add_calg(m,X,ad+m*ia[0]);
      X+=m;
   }
}


static void pack_adk(int n,int mu,sl3_alg_dble *ad)
{
   int m,*ia,*iam;
   sl3_alg_dble *X;

   if (io[2*mu])
   {
      m=n+1;
      ia=idx[mu].iak;
      iam=ia+idx[mu].nak;
      X=buf;

      for (;ia<iam;ia++)
      {
         assign_calg2calg(m,ad+m*ia[0],X);
         X+=m;
      }
   }
}


static void add_adk(int n,int mu,sl3_alg_dble *ad)
{
   int m,*ia,*iam;
   sl3_alg_dble *X;

   if (io[2*mu])
   {
      m=n+1;
      ia=idx[mu].iak;
      iam=ia+idx[mu].nak;
      X=buf;

      for (;ia<iam;ia++)
      {
         add_calg(m,X,ad+ia[0]*m);
         X+=m;
      }
   }
}


static void send_ad0(int n,int ifc)
{
   int tag,saddr,raddr,nbf;
   MPI_Status stat;

   tag=tag0[ifc];
   saddr=npr[ifc];
   raddr=npr[ifc^0x1];
   nbf=16*idx[ifc/2].na0*(n+1);

   if (np==0)
   {
      MPI_Send(sbuf,nbf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
      MPI_Recv(rbuf,nbf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
   }
   else
   {
      MPI_Recv(rbuf,nbf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
      MPI_Send(sbuf,nbf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
   }
}


static void send_adk(int n,int ifc)
{
   int tag,saddr,raddr,nbf;
   MPI_Status stat;

   tag=tag1[ifc];
   saddr=npr[ifc];
   raddr=npr[ifc^0x1];
   nbf=16*idx[ifc/2].nak*(n+1);

   if (np==0)
   {
      if (io[ifc])
         MPI_Send(sbuf,nbf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
      if (io[ifc^0x1])
         MPI_Recv(rbuf,nbf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
   }
   else
   {
      if (io[ifc^0x1])
         MPI_Recv(rbuf,nbf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
      if (io[ifc])
         MPI_Send(sbuf,nbf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
   }
}


void copy_bnd_ad(int n,sl3_alg_dble *ad)
{
   int m,mu,na0,nak;

   if (NPROC>1)
   {
      m=n+1;

      if (m>ms)
         alloc_buf(n);

      sbuf=buf;
      rbuf=ad+4*VOLUME*m;

      for (mu=0;mu<4;mu++)
      {
         na0=idx[mu].na0*m;

         if (na0>0)
         {
            pack_ad0(n,mu,ad);
            send_ad0(n,2*mu);
            rbuf+=na0;
         }
      }

      for (mu=0;mu<4;mu++)
      {
         nak=idx[mu].nak*m;

         if (nak>0)
         {
            pack_adk(n,mu,ad);
            send_adk(n,2*mu);
            rbuf+=nak;
         }
      }
   }
}


void add_bnd_ad(int n,sl3_alg_dble *ad)
{
   int m,mu,na0,nak;

   if (NPROC>1)
   {
      m=n+1;

      if (m>ms)
         alloc_buf(n);

      sbuf=ad+(4*VOLUME+(BNDRY/4))*m;
      rbuf=buf;

      for (mu=0;mu<4;mu++)
      {
         nak=idx[mu].nak*m;

         if (nak>0)
         {
            send_adk(n,2*mu+1);
            add_adk(n,mu,ad);
            sbuf+=nak;
         }
      }

      sbuf=ad+4*VOLUME*m;

      for (mu=0;mu<4;mu++)
      {
         na0=idx[mu].na0*m;

         if (na0>0)
         {
            send_ad0(n,2*mu+1);
            add_ad0(n,mu,ad);
            sbuf+=na0;
         }
      }
   }
}


void gather_bnd_ad0(sl3_alg_dble *ad)
{
   int mu,na0;

   if (NPROC>1)
   {
      if (ms==0)
         alloc_buf(0);

      sbuf=buf;
      rbuf=ad+4*VOLUME;

      for (mu=0;mu<4;mu++)
      {
         na0=idx[mu].na0;

         if (na0>0)
         {
            pack_ad0(0,mu,ad);
            send_ad0(0,2*mu);
            rbuf+=na0;
         }
      }
   }
}


void scatter_bnd_ad0(sl3_alg_dble *ad)
{
   int mu,na0;

   if (NPROC>1)
   {
      if (ms==0)
         alloc_buf(0);

      sbuf=ad+4*VOLUME;
      rbuf=buf;

      for (mu=0;mu<4;mu++)
      {
         na0=idx[mu].na0;

         if (na0>0)
         {
            send_ad0(0,2*mu+1);
            unpack_ad0(0,mu,ad);
            sbuf+=na0;
         }
      }
   }
}
