
/*******************************************************************************
*
* File aidx.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Labeling of the gauge field variables on the faces of the local lattice.
*
*   aidx_t *aidx(void)
*     Returns an array idx[4] of aidx_t structures containing the offsets
*     of the link variables at the faces of the local lattice.
*
*   void plaq_aidx(int n,int ix,int *ip)
*     Calculates the offsets ip[4] of the links in the (mu,nu)-plaquette at
*     the point on the local lattice with label ix. The indices (mu,nu) are
*     determined by the parameter n=0,..,5.
*
* The layout of the gauge field arrays and the contents of the structures
* returned by aidx() are described in the file README.aflds in this directory.
*
* In four dimensions there are six planes
*
*  (mu,nu)={(0,1),(0,2),(0,3),(2,3),(3,1),(1,2)}
*
* labeled by an integer n running from 0 to 5. The links in the (mu,nu)-
* plaquette at the point x are ordered such that
*
*   ip[0] -> A_mu(x)
*   ip[1] -> A_nu(x+mu)
*   ip[2] -> A_nu(x)
*   ip[3] -> A_mu(x+nu)
*
* An error occurs if the arguments passed to the program plaq_aidx() are not
* in the range 0<=n<6 and 0<=ix<VOLUME.
*
* The offsets ip[4] returned by plaq_aidx() at global time NPROC0*L0-1 take
* into account the fact that the true boundary values of the gauge potential
* at time NPROC0*L0 are at offset 4*VOLUME+7*(BNDRY/4)+k (k=0,1,2). Along
* the face in direction +0 at this time, the offsets thus point to these
* values rather than to the (unused) field variables of type (2).
*
* The programs in this module do not perform any communications and can be
* locally called.
*
*******************************************************************************/

#define AIDX_C

#include <stdlib.h>
#include <stdio.h>
#include "utils.h"
#include "lattice.h"
#include "aflds.h"
#include "global.h"

#define N0 (NPROC0*L0)

static const int plns[6][2]={{0,1},{0,2},{0,3},{2,3},{3,1},{1,2}};
static int nfc[4],ofs[8],ofs0[4],init=0;
static aidx_t idx[4];


static void alloc_idx(void)
{
   int mu,na0,nak;
   int *ia0,*iak;

   error_loc(iup[0][0]==0,1,"alloc_idx [aidx.c]",
         "Geometry arrays are not set");

   nfc[0]=FACE0/2;
   nfc[1]=FACE1/2;
   nfc[2]=FACE2/2;
   nfc[3]=FACE3/2;

   ofs[0]=VOLUME;
   ofs[1]=ofs[0]+(FACE0/2);
   ofs[2]=ofs[1]+(FACE0/2);
   ofs[3]=ofs[2]+(FACE1/2);
   ofs[4]=ofs[3]+(FACE1/2);
   ofs[5]=ofs[4]+(FACE2/2);
   ofs[6]=ofs[5]+(FACE2/2);
   ofs[7]=ofs[6]+(FACE3/2);

   ofs0[0]=4*VOLUME;
   ofs0[1]=ofs0[0]+(FACE0/2);
   ofs0[2]=ofs0[1]+(FACE1/2);
   ofs0[3]=ofs0[2]+(FACE2/2);

   if (BNDRY>0)
   {
      ia0=malloc(7*(BNDRY/4)*sizeof(*ia0));
      error_loc(ia0==NULL,1,"alloc_idx [aidx.c]",
            "Unable to allocate index array");
      iak=ia0+(BNDRY/4);
   }
   else
   {
      ia0=NULL;
      iak=NULL;
   }

   for (mu=0;mu<4;mu++)
   {
      na0=nfc[mu];
      nak=6*nfc[mu];

      idx[mu].na0=na0;
      idx[mu].nak=nak;

      if (na0>0)
      {
         idx[mu].ia0=ia0;
         idx[mu].iak=iak;
         ia0+=na0;
         iak+=nak;
      }
      else
      {
         idx[mu].ia0=NULL;
         idx[mu].iak=NULL;
      }
   }
}


static int offset(int ix,int mu)
{
   int iy,ib;

   if (ix<(VOLUME/2))
   {
      iy=iup[ix][mu];

      if (iy<VOLUME)
         return 8*(iy-(VOLUME/2))+2*mu+1;
      else
      {
         ib=iy-ofs[2*mu+1]-(BNDRY/2);

         return ofs0[mu]+ib;
      }
   }
   else
      return 8*(ix-(VOLUME/2))+2*mu;
}


static void set_idx(void)
{
   int mu,nu,k;
   int ib,iy,iz;
   int na0,*ia0,*iak;

   alloc_idx();

   for (mu=0;mu<4;mu++)
   {
      na0=idx[mu].na0;
      ia0=idx[mu].ia0;
      iak=idx[mu].iak;

      for (ib=0;ib<na0;ib++)
      {
         iy=ib+ofs[2*mu+1]+(BNDRY/2);
         iz=map[iy-VOLUME];
         ia0[ib]=8*(iz-(VOLUME/2))+2*mu+1;
      }

      for (ib=0;ib<na0;ib++)
      {
         iy=ib+ofs[2*mu+1];
         iz=map[iy-VOLUME];

         for (k=0;k<3;k++)
         {
            nu=k+(k>=mu);
            iak[3*ib+k]=offset(iz,nu);
         }
      }

      for (ib=0;ib<na0;ib++)
      {
         iy=ib+ofs[2*mu+1]+(BNDRY/2);
         iz=map[iy-VOLUME];

         for (k=0;k<3;k++)
         {
            nu=k+(k>=mu);
            iak[3*(ib+na0)+k]=offset(iz,nu);
         }
      }
   }

   init=1;
}


aidx_t *aidx(void)
{
   if (init==0)
      set_idx();

   return idx;
}


void plaq_aidx(int n,int ix,int *ip)
{
   int mu,nu;
   int iy,ic;

   if (init==0)
      set_idx();

   error_loc((n<0)||(n>=6)||(ix<0)||(ix>=VOLUME),1,"plaq_aidx [aidx.c]",
             "Arguments are out of range");

   mu=plns[n][0];
   nu=plns[n][1];

   ip[0]=offset(ix,mu);

   if ((mu==0)&&(global_time(ix)==(N0-1)))
   {
      ip[1]=4*VOLUME+7*(BNDRY/4)+nu-1;
   }
   else
   {
      iy=iup[ix][mu];

      if (iy<VOLUME)
         ip[1]=offset(iy,nu);
      else
      {
         if (iy<(VOLUME+(BNDRY/2)))
            ic=iy-VOLUME-nfc[mu];
         else
            ic=iy-VOLUME-(BNDRY/2);

         ip[1]=4*VOLUME+(BNDRY/4)+3*ic+nu-(nu>mu);
      }
   }

   ip[2]=offset(ix,nu);
   iy=iup[ix][nu];

   if (iy<VOLUME)
      ip[3]=offset(iy,mu);
   else
   {
      if (iy<(VOLUME+(BNDRY/2)))
         ic=iy-VOLUME-nfc[nu];
      else
         ic=iy-VOLUME-(BNDRY/2);

      ip[3]=4*VOLUME+(BNDRY/4)+3*ic+mu-(mu>nu);
   }
}
