
/*******************************************************************************
*
* File plaq_action.c
*
* Copyright (C) 2015, 2017 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Calculation of the quadratic part of the Wilson plaquette action.
*
*   complex_dble plaq_action(sl3_alg_dble *ad1,sl3_alg_dble *ad2)
*     Returns the value of the bilinear form defined by the quadratic part
*     of the plaquette gauge action (see the notes).
*
* The layout of the arrays ad1 and ad2, representing the gauge fields A1
* and A2, must be as described in README.global and README.aflds. It is up
* to the user to set the desired boundary values by calling set_adbc(), for
* example.
*
* The calculated plaquette values are
*
*  s(x,mu,nu)=-(1/4)*tr{F1_{mu,nu}(x)*F2_{mu,nu}(x)},
*
* where
*
*  F1_{mu,nu}(x)=A1_{nu}(x+mu)-A1_{nu}(x)-A1_{mu}(x+nu)+A1_{mu}(x),
*
* is the field tensor of A1 (the one of A2 is defined in the same way). The
* total action is then
*
*  sum_{x0=0}^{T-1}sum_{x1,x2,x3}sum_{mu>nu=0}^3 w(x0,mu,nu)*s(x,mu,nu},
*
* where the weight factor w(x0,mu,nu) is equal to 1/2 for the space-like
* plaquettes at x0=0 and 1 otherwise.
*
*******************************************************************************/

#define PLAQ_ACTION_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "utils.h"
#include "lattice.h"
#include "su3fcts.h"
#include "aflds.h"
#include "global.h"

static int idsm,init=0;
static sl3_alg_dble *ad1b,*ad2b;
static sl3_alg_dble Xw ALIGNED16;
static sl3_alg_dble Yw ALIGNED16;


static void set_ids(void)
{
   idsm=init_hsum(2);
   init=1;
}


static complex_dble plaq(int n,int ix)
{
   int ip[4];

   plaq_aidx(n,ix,ip);

   Xw=ad1b[ip[1]];
   sl3_sub(ad1b+ip[2],&Xw);
   sl3_sub(ad1b+ip[3],&Xw);
   sl3_add(ad1b+ip[0],&Xw);

   Yw=ad2b[ip[1]];
   sl3_sub(ad2b+ip[2],&Yw);
   sl3_sub(ad2b+ip[3],&Yw);
   sl3_add(ad2b+ip[0],&Yw);

   return sl3_biform(&Xw,&Yw);
}


static void local_plaq_action(void)
{
   int ix,t,n;
   complex_dble z,w;

   if (init==0)
      set_ids();

   reset_hsum(idsm);

   for (ix=0;ix<VOLUME;ix++)
   {
      t=global_time(ix);
      z.re=0.0;
      z.im=0.0;

      for (n=0;n<6;n++)
      {
         w=plaq(n,ix);

         if ((n<3)||(t>0))
         {
            z.re+=w.re;
            z.im+=w.im;
         }
         else
         {
            z.re+=0.5*w.re;
            z.im+=0.5*w.im;
         }
      }

      add_to_hsum(idsm,(double*)(&z));
   }
}


complex_dble plaq_action(sl3_alg_dble *ad1,sl3_alg_dble *ad2)
{
   complex_dble z;

   copy_bnd_ad(0,ad1);
   if (ad1!=ad2)
      copy_bnd_ad(0,ad2);

   ad1b=ad1;
   ad2b=ad2;

   local_plaq_action();
   global_hsum(idsm,(double*)(&z));

   z.re*=0.5;
   z.im*=0.5;

   return z;
}
