
/*******************************************************************************
*
* File archive.c
*
* Copyright (C) 2005-2016 Martin Luescher, Mattia Dalla Brida
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Programs to read and write gauge potentials.
*
*   void write_afld(char *out,int n,sl3_alg_dble **ad)
*     Writes the lattice sizes, the process grid sizes, the coordinates of
*     the calling process, the local part of the Wilson action and the local
*     gauge potential ad to the file "out" (see the notes).
*
*   void read_afld(char *in,int n,sl3_alg_dble **ad)
*     Reads the local gauge potential ad from the file "in", assuming it was
*     written to the file by the program write_afld(). The program then checks
*     that the field is compatible with the boundary conditions specified in
*     the parameter data base (see the notes).
*
*   void export_afld(char *out,int n,sl3_alg_dble **ad)
*     Writes the lattice sizes and the global gauge potential ad to the file
*     "out" from process 0 in the universal format specified below (see the
*     notes).
*
*   void import_afld(char *in,int n,sl3_alg_dble **ad)
*     Reads the global gauge potential ad from the file "in" on process 0,
*     assuming the field was written to the file in the universal format.
*     The program then checks that the field is compatible with the boundary
*     conditions specified in the parameter data base (see the notes).
*
* In all programs in this module, the array "ad" is expected to contain the
* n+1 coefficient fields in the expansion
*
*   A=A[0]+g0*A[1]+..+g0^n*A[n]
*
* of the gauge potential in powers of the coupling g0. Each of the fields
* ad[0],..,ad[n] is assumed to be of the type returned by reserve_wad(1)
* (see utils/wspace.c).
*
* The program export_afld() first writes the lattice sizes, the number n+1 of
* coefficient fields and the Wilson plaquette action [uflds/Eplaq.c]. Then
* follow the coefficient fields ad[0],..,ad[n] one after another. In each case
* the field variables on the 8 links in the directions +0,-0,...,+3,-3 at the
* first odd point come first, then those at the second odd point, and so on.
*
* The order of the point (x0,x1,x2,x3) with Cartesian coordinates in the range
* 0<=x0<N0,...,0<=x3<N3 is determined by the index
*
*   ix=x3+N3*x2+N2*N3*x1+N1*N2*N3*x0,
*
* where N0,N1,N2,N3 are the global lattice sizes (N0=NPROC0*L0, etc.).
*
* Independently of the machine, the export function writes the data to the
* output file in little-endian byte order. Integers and double-precision
* numbers on the output file occupy 4 and 8 bytes, respectively, the latter
* being formatted according to the IEEE-754 standard. The import function
* assumes the data on the input file to be little endian and converts them
* to big-endian order if the machine is big endian. Exported fields can
* thus safely be exchanged between different machines.
*
* The programs read_afld() and import_afld() check the compatibility of the
* gauge potential read from file with the chosen boundary conditions by calling
* check_adbc() [lattice/bcnds.c]. They furthermore check that the Wilson action
* of the field coincides with the value read from the file.
*
* All programs in this module require a workspace of n+1 su3_dble link fields.
* They involve global communications and must be called simultaneously on all
* MPI processes.
*
*******************************************************************************/

#define ARCHIVE_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "sl3.h"
#include "utils.h"
#include "lattice.h"
#include "aflds.h"
#include "uflds.h"
#include "linalg.h"
#include "archive.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)

static int endian,ms=0,nbf=0;
static complex_dble *act0,*act1;
static sl3_alg_dble *abuf,*pbuf;


static void alloc_arrays(int n)
{
   int m;

   if (ms>0)
      free(act0);

   m=n+1;
   act0=malloc(2*m*sizeof(*act0));
   act1=act0+m;
   error(act0==NULL,1,"alloc_arrays [archive.c]",
         "Unable to allocate auxiliary arrays");
   ms=m;
}


void write_afld(char *out,int n,sl3_alg_dble **ad)
{
   int ldat[17],iw;
   int ie,m,i;
   su3_dble **wud,*ud;
   FILE *fout;

   error(iup[0][0]==0,1,"write_afld [archive.c]",
         "Geometry arrays are not set");

   m=n+1;
   ie=1;

   for (i=0;i<m;i++)
   {
      if (cpr[0]==(NPROC0-1))
         set_calg2zero(3,ad[i]+4*VOLUME+7*(BNDRY/4));
      ie&=check_adbc(ad[i]);
   }

   error_root(ie!=1,1,"write_afld [archive.c]",
              "Incorrect boundary values");

   if (m>ms)
      alloc_arrays(n);

   fout=fopen(out,"wb");
   error_loc(fout==NULL,1,"write_afld [archive.c]",
             "Unable to open output file");

   ldat[0]=NPROC0;
   ldat[1]=NPROC1;
   ldat[2]=NPROC2;
   ldat[3]=NPROC3;

   ldat[4]=L0;
   ldat[5]=L1;
   ldat[6]=L2;
   ldat[7]=L3;

   ldat[8]=NPROC0_BLK;
   ldat[9]=NPROC1_BLK;
   ldat[10]=NPROC2_BLK;
   ldat[11]=NPROC3_BLK;

   ldat[12]=cpr[0];
   ldat[13]=cpr[1];
   ldat[14]=cpr[2];
   ldat[15]=cpr[3];

   ldat[16]=m;

   wud=reserve_wud(m);
   ud=wud[0];
   a2uflds(n,ad,ud);
   Splaq(n,ud,act0);
   release_wud();

   iw=fwrite(ldat,sizeof(int),17,fout);
   iw+=fwrite(act0,sizeof(complex_dble),m,fout);

   for (i=0;i<m;i++)
      iw+=fwrite(ad[i],sizeof(sl3_alg_dble),4*VOLUME,fout);

   error_loc(iw!=(17+m+4*VOLUME*m),1,
             "write_afld [archive.c]","Incorrect write count");
   fclose(fout);
}


static double cmp_action(int n)
{
   int i;
   double m,d,mx,dx;
   complex_dble w,z;

   mx=0.0;
   dx=0.0;

   for (i=0;i<=n;i++)
   {
      w.re=act0[i].re;
      w.im=act0[i].im;
      z.re=w.re-act1[i].re;
      z.im=w.im-act1[i].im;

      m=w.re*w.re+w.im*w.im;
      d=z.re*z.re+z.im*z.im;

      if (m>mx)
         mx=m;

      if (d>dx)
         dx=d;
   }

   error_root((mx==0.0)&&(dx>0.0),1,"cmp_action [archive.c]",
              "Vanishing action and non-zero action difference");

   return sqrt(dx/mx);
}


void read_afld(char *in,int n,sl3_alg_dble **ad)
{
   int ldat[17],ir;
   int ie,m,i;
   double eps,dev;
   su3_dble **wud,*ud;
   FILE *fin;

   error(iup[0][0]==0,1,"read_afld [archive.c]",
         "Geometry arrays are not set");

   m=n+1;

   if (m>ms)
      alloc_arrays(n);

   fin=fopen(in,"rb");
   error_loc(fin==NULL,1,"read_afld [archive.c]",
             "Unable to open input file");

   ir=fread(ldat,sizeof(int),17,fin);

   error_loc(ir!=17,1,"read_afld [archive.c]","Incorrect read count");

   ie=0;
   ie|=((ldat[0]!=NPROC0)||(ldat[1]!=NPROC1)||
        (ldat[2]!=NPROC2)||(ldat[3]!=NPROC3));
   ie|=((ldat[4]!=L0)||(ldat[5]!=L1)||
        (ldat[6]!=L2)||(ldat[7]!=L3));
   ie|=((ldat[8]!=NPROC0_BLK)||(ldat[9]!=NPROC1_BLK)||
        (ldat[10]!=NPROC2_BLK)||(ldat[11]!=NPROC3_BLK));
   ie|=((ldat[12]!=cpr[0])||(ldat[13]!=cpr[1])||
        (ldat[14]!=cpr[2])||(ldat[15]!=cpr[3]));
   error_loc(ie!=0,1,"read_afld [archive.c]","Unexpected lattice data");

   error_loc(ldat[16]!=m,1,"read_afld [archive.c]",
             "Unexpected number of coefficient fields");

   ir+=fread(act0,sizeof(complex_dble),m,fin);

   for (i=0;i<m;i++)
      ir+=fread(ad[i],sizeof(sl3_alg_dble),4*VOLUME,fin);

   error_loc(ir!=(17+m+4*VOLUME*m),2,
             "read_afld [archive.c]","Incorrect read count");
   fclose(fin);

   ie=1;

   for (i=0;i<m;i++)
   {
      if (cpr[0]==(NPROC0-1))
         set_calg2zero(3,ad[i]+4*VOLUME+7*(BNDRY/4));
      ie&=check_adbc(ad[i]);
   }

   error_root(ie!=1,1,"read_afld [archive.c]",
              "Incorrect boundary values");

   wud=reserve_wud(m);
   ud=wud[0];
   a2uflds(n,ad,ud);
   Splaq(n,ud,act1);
   release_wud();

   dev=cmp_action(n);
   eps=16.0*sqrt((double)(4*N0*N1)*(double)(N2*N3))*DBL_EPSILON;

   error_loc(dev>eps,1,"read_afld [archive.c]",
             "Incorrect value of the gauge action");
}


static void check_machine(void)
{
   error_root(sizeof(stdint_t)!=4,1,"check_machine [archive.c]",
              "Size of a stdint_t integer is not 4");
   error_root(sizeof(double)!=8,1,"check_machine [archive.c]",
              "Size of a double is not 8");
   error_root(sizeof(sl3_alg_dble)!=128,1,"check_machine [archive.c]",
              "The sl3_alg_dble structures are not properly packed");

   endian=endianness();
   error_root(endian==UNKNOWN_ENDIAN,1,"check_machine [archive.c]",
              "Unkown endianness");
}


static void alloc_abuf(int my_rank)
{
   if (nbf==0)
      check_machine();

   if (nbf>0)
      afree(abuf);

   if (my_rank==0)
   {
      abuf=amalloc(4*(L3+N3)*sizeof(sl3_alg_dble),ALIGN);
      pbuf=abuf+4*L3;
   }
   else
   {
      abuf=amalloc(4*L3*sizeof(sl3_alg_dble),ALIGN);
      pbuf=NULL;
   }

   error(abuf==NULL,1,"alloc_abuf [archive.c]",
         "Unable to allocate auxiliary array");
   nbf=1;
}


static void get_afld(int iy,sl3_alg_dble *ad)
{
   int y3,ifc;
   sl3_alg_dble *p,*q;

   q=abuf;
   iy*=L3;

   if (ipt[iy]<(VOLUME/2))
      iy+=1;

   for (y3=0;y3<L3;y3+=2)
   {
      p=ad+8*(ipt[iy+y3]-(VOLUME/2));

      for (ifc=0;ifc<8;ifc++)
         q[ifc]=p[ifc];

      q+=8;
   }
}


static void set_afld(int iy,sl3_alg_dble *ad)
{
   int y3,ifc;
   sl3_alg_dble *p,*q;

   q=abuf;
   iy*=L3;

   if (ipt[iy]<(VOLUME/2))
      iy+=1;

   for (y3=0;y3<L3;y3+=2)
   {
      p=ad+8*(ipt[iy+y3]-(VOLUME/2));

      for (ifc=0;ifc<8;ifc++)
         p[ifc]=q[ifc];

      q+=8;
   }
}


void export_afld(char *out,int n,sl3_alg_dble **ad)
{
   int my_rank,np[4],p;
   int ie,m,i;
   int iw,iwa,dmy,tag0,tag1;
   int x0,x1,x2,x3,y0,y1,y2,ix,iy;
   stdint_t istd[5];
   su3_dble **wud,*ud;
   MPI_Status stat;
   FILE *fout=NULL;

   error(iup[0][0]==0,1,"export_afld [archive.c]",
         "Geometry arrays are not set");

   m=n+1;
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (m>ms)
      alloc_arrays(n);

   if (nbf==0)
      alloc_abuf(my_rank);

   dmy=1;
   tag0=mpi_tag();
   tag1=mpi_tag();
   ie=1;

   for (i=0;i<m;i++)
   {
      if (cpr[0]==(NPROC0-1))
         set_calg2zero(3,ad[i]+4*VOLUME+7*(BNDRY/4));
      ie&=check_adbc(ad[i]);
   }

   error_root(ie!=1,1,"export_afld [archive.c]",
              "Incorrect boundary values");

   wud=reserve_wud(m);
   ud=wud[0];
   a2uflds(n,ad,ud);
   Splaq(n,ud,act0);
   release_wud();

   if (my_rank==0)
   {
      fout=fopen(out,"wb");
      error_root(fout==NULL,1,"export_afld [archive.c]",
                 "Unable to open output file");

      istd[0]=(stdint_t)(N0);
      istd[1]=(stdint_t)(N1);
      istd[2]=(stdint_t)(N2);
      istd[3]=(stdint_t)(N3);
      istd[4]=(stdint_t)(m);

      if (endian==BIG_ENDIAN)
      {
         bswap_int(5,istd);
         bswap_double(2*m,act0);
      }

      iw=fwrite(istd,sizeof(stdint_t),5,fout);
      iw+=fwrite(act0,sizeof(complex_dble),m,fout);

      error_root(iw!=5+m,1,"export_afld [archive.c]","Incorrect write count");
   }

   iwa=0;

   for (i=0;i<m;i++)
   {
      for (ix=0;ix<(N0*N1*N2);ix++)
      {
         x0=ix/(N1*N2);
         x1=(ix/N2)%N1;
         x2=ix%N2;

         y0=x0%L0;
         y1=x1%L1;
         y2=x2%L2;
         iy=y2+L2*y1+L1*L2*y0;

         np[0]=x0/L0;
         np[1]=x1/L1;
         np[2]=x2/L2;

         for (x3=0;x3<N3;x3+=L3)
         {
            np[3]=x3/L3;
            p=ipr_global(np);
            if (my_rank==p)
               get_afld(iy,ad[i]);

            if (p>0)
            {
               if (my_rank==0)
               {
                  MPI_Send(&dmy,1,MPI_INT,p,tag0,MPI_COMM_WORLD);
                  MPI_Recv(abuf,4*L3*16,MPI_DOUBLE,p,tag1,MPI_COMM_WORLD,
                           &stat);
               }
               else if (my_rank==p)
               {
                  MPI_Recv(&dmy,1,MPI_INT,0,tag0,MPI_COMM_WORLD,&stat);
                  MPI_Send(abuf,4*L3*16,MPI_DOUBLE,0,tag1,MPI_COMM_WORLD);
               }
            }

            if (my_rank==0)
            {
               if (endian==BIG_ENDIAN)
                  bswap_double(4*L3*16,abuf);
               iw=fwrite(abuf,sizeof(sl3_alg_dble),4*L3,fout);
               iwa|=(iw!=(4*L3));
            }
         }
      }
   }

   if (my_rank==0)
   {
      error_root(iwa!=0,1,"export_afld [archive.c]",
                 "Incorrect write count");
      fclose(fout);
   }
}


void import_afld(char *in,int n,sl3_alg_dble **ad)
{
   int my_rank,np[4],p;
   int ie,m,i,l;
   int ir,ira,dmy,tag0,tag1;
   int x0,x1,x2,y0,y1,y2,y3,ix,iy;
   stdint_t istd[5];
   double eps,dev;
   su3_dble **wud,*ud;
   MPI_Status stat;
   FILE *fin=NULL;

   error(iup[0][0]==0,1,"import_afld [archive.c]",
         "Geometry arrays are not set");

   m=n+1;
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (m>ms)
      alloc_arrays(n);

   if (nbf==0)
      alloc_abuf(my_rank);

   dmy=1;
   tag0=mpi_tag();
   tag1=mpi_tag();

   if (my_rank==0)
   {
      fin=fopen(in,"rb");
      error_root(fin==NULL,1,"import_afld [archive.c]",
                 "Unable to open input file");

      ir=fread(istd,sizeof(stdint_t),5,fin);
      ir+=fread(act0,sizeof(complex_dble),m,fin);
      error_root(ir!=5+m,1,"import_afld [archive.c]","Incorrect read count");

      if (endian==BIG_ENDIAN)
      {
         bswap_int(5,istd);
         bswap_double(2*m,act0);
      }

      np[0]=(int)(istd[0]);
      np[1]=(int)(istd[1]);
      np[2]=(int)(istd[2]);
      np[3]=(int)(istd[3]);

      error_root((np[0]!=N0)||(np[1]!=N1)||(np[2]!=N2)||(np[3]!=N3),1,
                 "import_afld [archive.c]","Unexpected lattice sizes");

      error_root(m!=(int)(istd[4]),1,"import_afld [archive.c]",
                 "Incorrect number of coefficient fields");
   }

   MPI_Bcast(np,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(act0,2*m,MPI_DOUBLE,0,MPI_COMM_WORLD);
   ira=0;

   for (i=0;i<m;i++)
   {
      for (ix=0;ix<(N0*N1*N2);ix++)
      {
         x0=ix/(N1*N2);
         x1=(ix/N2)%N1;
         x2=ix%N2;

         if (my_rank==0)
         {
            ir=fread(pbuf,sizeof(sl3_alg_dble),4*N3,fin);
            ira|=(ir!=(4*N3));

            if (endian==BIG_ENDIAN)
               bswap_double(4*N3*16,pbuf);
         }

         y0=x0%L0;
         y1=x1%L1;
         y2=x2%L2;
         iy=y2+L2*y1+L1*L2*y0;

         np[0]=x0/L0;
         np[1]=x1/L1;
         np[2]=x2/L2;

         for (y3=0;y3<N3;y3+=L3)
         {
            np[3]=y3/L3;
            p=ipr_global(np);

            if (p>0)
            {
               if (my_rank==0)
               {
                  MPI_Send(pbuf+4*y3,4*L3*16,MPI_DOUBLE,p,tag1,
                           MPI_COMM_WORLD);
                  MPI_Recv(&dmy,1,MPI_INT,p,tag0,MPI_COMM_WORLD,&stat);
               }
               else if (my_rank==p)
               {
                  MPI_Recv(abuf,4*L3*16,MPI_DOUBLE,0,tag1,
                           MPI_COMM_WORLD,&stat);
                  MPI_Send(&dmy,1,MPI_INT,0,tag0,MPI_COMM_WORLD);
               }
            }
            else if (my_rank==0)
            {
               for (l=0;l<(4*L3);l++)
                  abuf[l]=pbuf[4*y3+l];
            }

            if (my_rank==p)
               set_afld(iy,ad[i]);
         }
      }
   }

   if (my_rank==0)
   {
      error_root(ira!=0,1,"import_afld [archive.c]","Incorrect read count");
      fclose(fin);
   }

   ie=1;

   for (i=0;i<m;i++)
   {
      if (cpr[0]==(NPROC0-1))
         set_calg2zero(3,ad[i]+4*VOLUME+7*(BNDRY/4));
      ie&=check_adbc(ad[i]);
   }

   error_root(ie!=1,1,"import_afld [archive.c]",
              "Incorrect boundary values");

   wud=reserve_wud(m);
   ud=wud[0];
   a2uflds(n,ad,ud);
   Splaq(n,ud,act1);
   release_wud();

   dev=cmp_action(n);
   eps=16.0*sqrt((double)(4*N0*N1)*(double)(N2*N3))*DBL_EPSILON;

   error_loc(dev>eps,1,"import_afld [archive.c]",
             "Incorrect value of the gauge action");
}
