
/*******************************************************************************
*
* File dfp.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Application of the Faddeev-Popov operator.
*
*   void dfp(sl3_alg_dble *cd,sl3_alg_dble *rd)
*     Applies the Faddeev-Popov operator to the ghost field cd, assuming
*     it satisfies the chosen boundary conditions, and assigns the result
*     to the field rd. The two fields must be different.
*
* The operator is defined the notes
*
*  M. Luescher, "Normalization of the gradient-flow coupling in LQCD"
*
* [file doc/gfnorm.pdf]. It is taken for granted that the ghost field arrays
* have at least NSPIN elements.
*
* The program in this module performs global operations and must be called
* simultaneously on all processes.
*
*******************************************************************************/

#define DFP_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "utils.h"
#include "pmdb.h"
#include "lattice.h"
#include "linalg.h"
#include "cflds.h"
#include "global.h"

#define N0 (NPROC0*L0)

static int root;
static double *rt=NULL;
static sl3_alg_dble ct0 ALIGNED16;


static void set_rt(void)
{
   int t;
   int cp[4];

   rt=malloc(N0*sizeof(*rt));
   error(rt==NULL,1,"set_rt [dfp.c]",
         "Unable to allocate auxiliary array");

   for (t=0;t<N0;t++)
      rt[t]=(1.0-(double)(t)/(double)(N0))/(double)(N0*N0);

   cp[0]=0;
   cp[1]=0;
   cp[2]=0;
   cp[3]=0;

   root=ipr_global(cp);
}


void dfp(sl3_alg_dble *cd,sl3_alg_dble *rd)
{
   int bc,ix,iy,t,mu;

   cpcd_int_bnd(0,cd);
   set_calg2zero(BNDRY/2,rd+VOLUME);
   bc=bc_type();

   if (bc==1)
   {
      if (rt==NULL)
         set_rt();

      ct0=cd[ipt[0]];
      MPI_Bcast(&ct0,16,MPI_DOUBLE,root,MPI_COMM_WORLD);

      for (ix=0;ix<(VOLUME/2);ix++)
      {
         t=global_time(ix);
         _sl3_alg_mulr_assign(rd[ix],rt[t],ct0);

         if (t>0)
         {
            _sl3_alg_mulr_add_assign(rd[ix],8.0,cd[ix]);
         }

         if (t==1)
         {
            _sl3_alg_sub_assign(rd[ix],ct0);
         }
      }

      for (ix=(VOLUME/2);ix<VOLUME;ix++)
      {
         t=global_time(ix);

         _sl3_alg_mulr_assign(rd[ix],rt[t],ct0);

         if (t>0)
         {
            _sl3_alg_mulr_add_assign(rd[ix],8.0,cd[ix]);

            if (t<(N0-1))
            {
               iy=iup[ix][0];
               _sl3_alg_sub_assign(rd[ix],cd[iy]);
               _sl3_alg_sub_assign(rd[iy],cd[ix]);
            }

            iy=idn[ix][0];
            _sl3_alg_sub_assign(rd[ix],cd[iy]);

            if (t>1)
            {
               _sl3_alg_sub_assign(rd[iy],cd[ix]);
            }

            for (mu=1;mu<4;mu++)
            {
               iy=iup[ix][mu];
               _sl3_alg_sub_assign(rd[ix],cd[iy]);
               _sl3_alg_sub_assign(rd[iy],cd[ix]);

               iy=idn[ix][mu];
               _sl3_alg_sub_assign(rd[ix],cd[iy]);
               _sl3_alg_sub_assign(rd[iy],cd[ix]);
            }
         }
      }
   }
   else
   {
      for (ix=0;ix<(VOLUME/2);ix++)
      {
         _sl3_alg_mulr_assign(rd[ix],8.0,cd[ix]);
      }

      for (ix=(VOLUME/2);ix<VOLUME;ix++)
      {
         t=global_time(ix);
         _sl3_alg_mulr_assign(rd[ix],8.0,cd[ix]);

         if (t<(N0-1))
         {
            iy=iup[ix][0];

            if (t>0)
            {
               _sl3_alg_sub_assign(rd[ix],cd[iy]);
            }
            else
            {
               _sl3_alg_mulr_add_assign(rd[ix],-2.0,cd[iy]);
            }

            _sl3_alg_sub_assign(rd[iy],cd[ix]);
         }

         if (t>0)
         {
            iy=idn[ix][0];
            _sl3_alg_sub_assign(rd[ix],cd[iy]);

            if (t>1)
            {
               _sl3_alg_sub_assign(rd[iy],cd[ix]);
            }
            else
            {
               _sl3_alg_mulr_add_assign(rd[iy],-2.0,cd[ix]);
            }
         }

         for (mu=1;mu<4;mu++)
         {
            iy=iup[ix][mu];
            _sl3_alg_sub_assign(rd[ix],cd[iy]);
            _sl3_alg_sub_assign(rd[iy],cd[ix]);

            iy=idn[ix][mu];
            _sl3_alg_sub_assign(rd[ix],cd[iy]);
            _sl3_alg_sub_assign(rd[iy],cd[ix]);
         }
      }
   }

   cpcd_ext_bnd(0,rd);
}
