
/*******************************************************************************
*
* File rfact.c
*
* Copyright (C) 2015, 2016 Martin Luescher
*
* O(a) boundary correction to the gradient-flow running coupling.
*
*   void rfact(char *type,int N,int T,int L,int bc,double c1,
*              double t,double *r)
*     Computes the coefficients r,r' in lattice units and assigns them
*     to r[0],r[1] (see the notes). r[1] is set to zero in the case of
*     SF boundary conditions.
*
* The calculated coefficients r and r' are defined in the notes
*
*  M. Luescher: "O(a) boundary effects on the gradient-flow coupling"
*
* (file doc/bndimp.pdf).

* The parameters of rfact() are
*
*  type              Specifies the type of coupling ("c","p","cs","ct",
*                    "ps","pt", where "c" stands for clover, "p" for
*                    plaquette, "s" for spatial and "t" for time-like).
*
*  N                 Number of colours.
*
*  T,L               Time-like and spatial extent of the lattice.
*
*  bc                Boundary condition (1: SF, 2: open-SF).
*
*  c1                Coefficient of the double-plaquette terms in the
*                    gauge action.
*
*  t                 Flow time.
*
* The physical time, x0, is set to T/2. All dimensionful quantities must be
* given in lattice units. The coefficient c in the notes is related to the
* flow time t through
*
*  c=sqrt(8*t)/L
*
* In an MPI program, the function rfact() should only be called by the MPI
* root process.
*
*******************************************************************************/

#define RFACT_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "utils.h"
#include "gbar.h"

static int Ts=0,Ls=0;
static double c1s,ts,Ds[4][4],Ks[4][4][4][4];
static double fs[9][9];
static double *ph0,*pc0,*ph1,*pc1;
static double *sx0,*sxp,*sxm,*cx0,*cxp,*cxm;


static void alloc_arrays(int T,int L)
{
   if (Ts!=0)
      free(ph0);

   ph0=malloc((8*T+2*L)*sizeof(*ph0));
   error_root(ph0==NULL,1,"alloc_arrays [rfact.c]",
              "Unable to allocate auxiliary arrays");

   pc0=ph0+T;
   ph1=pc0+T;
   pc1=ph1+L;
   sx0=pc1+L;
   sxp=sx0+T;
   sxm=sxp+T;
   cx0=sxm+T;
   cxp=cx0+T;
   cxm=cxp+T;

   Ts=T;
   Ls=L;
}


static void set_arrays(int T,int L,int bc)
{
   int n;
   double pi,p0,p1,x0;

   pi=4.0*atan(1.0);
   x0=0.5*(double)(T);

   for (n=0;n<T;n++)
   {
      if (bc==1)
         p0=(double)(n)*pi/(double)(T);
      else
         p0=((double)(n)+0.5)*pi/(double)(T);

      ph0[n]=2.0*sin(0.5*p0);
      pc0[n]=cos(0.5*p0);

      sx0[n]=sin(p0*x0);
      sxp[n]=sin(p0*x0+0.5*p0);
      sxm[n]=sin(p0*x0-0.5*p0);

      cx0[n]=cos(p0*x0);
      cxp[n]=cos(p0*x0+0.5*p0);
      cxm[n]=cos(p0*x0-0.5*p0);
   }

   for (n=0;n<L;n++)
   {
      p1=(double)(n)*2.0*pi/(double)(L);
      ph1[n]=2.0*sin(0.5*p1);
      pc1[n]=cos(0.5*p1);
   }
}


static double set_Ds(int T,int bc,int *n)
{
   double ph[4],phs[4],rm[4],vm[4];
   double ph2,ph4,sm;

   ph[0]=ph0[n[0]];
   ph[1]=ph1[n[1]];
   ph[2]=ph1[n[2]];
   ph[3]=ph1[n[3]];

   phs[0]=ph[0]*ph[0];
   phs[1]=ph[1]*ph[1];
   phs[2]=ph[2]*ph[2];
   phs[3]=ph[3]*ph[3];

   ph2=phs[0]+phs[1]+phs[2]+phs[3];
   ph4=phs[0]*phs[0]+phs[1]*phs[1]+phs[2]*phs[2]+phs[3]*phs[3];

   rm[0]=ph2-c1s*(ph4+ph2*phs[0]);
   rm[1]=ph2-c1s*(ph4+ph2*phs[1]);
   rm[2]=ph2-c1s*(ph4+ph2*phs[2]);
   rm[3]=ph2-c1s*(ph4+ph2*phs[3]);

   vm[0]=c1s*ph[0]*phs[0]/rm[0];
   vm[1]=c1s*ph[1]*phs[1]/rm[1];
   vm[2]=c1s*ph[2]*phs[2]/rm[2];
   vm[3]=c1s*ph[3]*phs[3]/rm[3];

   sm=vm[0]*vm[0]*rm[0]+vm[1]*vm[1]*rm[1]+
      vm[2]*vm[2]*rm[2]+vm[3]*vm[3]*rm[3];
   sm=1.0/(1.0+sm);

   if ((n[0]!=0)||(bc==2))
   {
      Ds[0][0]=1.0/rm[0]-sm*vm[0]*vm[0];
      Ds[1][1]=1.0/rm[1]-sm*vm[1]*vm[1];
      Ds[2][2]=1.0/rm[2]-sm*vm[2]*vm[2];
      Ds[3][3]=1.0/rm[3]-sm*vm[3]*vm[3];

      Ds[0][1]=-sm*vm[0]*vm[1];
      Ds[1][0]=Ds[0][1];

      Ds[0][2]=-sm*vm[0]*vm[2];
      Ds[2][0]=Ds[0][2];

      Ds[0][3]=-sm*vm[0]*vm[3];
      Ds[3][0]=Ds[0][3];

      Ds[1][2]=-sm*vm[1]*vm[2];
      Ds[2][1]=Ds[1][2];

      Ds[1][3]=-sm*vm[1]*vm[3];
      Ds[3][1]=Ds[1][3];

      Ds[2][3]=-sm*vm[2]*vm[3];
      Ds[3][2]=Ds[2][3];
   }
   else
   {
      if ((n[1]!=0)||(n[2]!=0)||(n[3]!=0))
         Ds[0][0]=1.0/(ph2-c1s*ph4);
      else
         Ds[0][0]=(double)(T*T);

      Ds[0][1]=0.0;
      Ds[0][2]=0.0;
      Ds[0][3]=0.0;

      Ds[1][0]=0.0;
      Ds[1][1]=0.0;
      Ds[1][2]=0.0;
      Ds[1][3]=0.0;

      Ds[2][0]=0.0;
      Ds[2][1]=0.0;
      Ds[2][2]=0.0;
      Ds[2][3]=0.0;

      Ds[3][0]=0.0;
      Ds[3][1]=0.0;
      Ds[3][2]=0.0;
      Ds[3][3]=0.0;
   }

   return ph2;
}


static void set_Ks(int *n)
{
   int mu,nu,rho,sig;
   double ph[4];

   ph[0]=ph0[n[0]];
   ph[1]=ph1[n[1]];
   ph[2]=ph1[n[2]];
   ph[3]=ph1[n[3]];

   for (mu=0;mu<4;mu++)
   {
      for (nu=mu;nu<4;nu++)
      {
         for (rho=0;rho<4;rho++)
         {
            for (sig=rho;sig<4;sig++)
            {
               if ((mu==nu)||(rho==sig))
                  Ks[mu][nu][rho][sig]=0.0;
               else
               {
                  Ks[mu][nu][rho][sig]=
                     ph[mu]*ph[rho]*Ds[nu][sig]-
                     ph[mu]*ph[sig]*Ds[nu][rho]+
                     ph[nu]*ph[sig]*Ds[mu][rho]-
                     ph[nu]*ph[rho]*Ds[mu][sig];
               }

               Ks[mu][nu][sig][rho]=-Ks[mu][nu][rho][sig];
               Ks[nu][mu][sig][rho]= Ks[mu][nu][rho][sig];
               Ks[nu][mu][rho][sig]=-Ks[mu][nu][rho][sig];
            }
         }
      }
   }
}


static void fsq_SF_clov(int T,int bc,int flags,int *n,double *sm)
{
   int p0,k,l,j,idx;
   double s,ph2,r[2];

   for (k=0;k<2;k++)
   {
      for (l=0;l<9;l++)
         fs[k][l]=0.0;
   }

   for (p0=0;p0<T;p0++)
   {
      n[0]=p0;
      ph2=set_Ds(T,bc,n);
      set_Ks(n);
      r[0]=exp(-ts*ph2)*pc0[p0];
      if (p0==0)
         r[0]*=0.5;
      r[1]=r[0]*sx0[p0];
      r[0]=r[0]*cx0[p0]*pc0[p0];

      idx=0;

      for (k=1;k<=3;k++)
      {
         for (j=1;j<=3;j++)
         {
            fs[0][idx]+=r[0]*Ks[0][k][0][j];
            idx+=1;
         }
      }

      idx=0;

      for (k=1;k<=3;k++)
      {
         for (l=(k+1);l<=3;l++)
         {
            for (j=1;j<=3;j++)
            {
               fs[1][idx]+=r[1]*Ks[k][l][0][j];
               idx+=1;
            }
         }
      }
   }

   sm[0]=0.0;

   if (flags&0x2)
   {
      idx=0;

      for (k=1;k<=3;k++)
      {
         for (j=1;j<=3;j++)
         {
            s=pc1[n[k]]*fs[0][idx];
            sm[0]+=2.0*s*s;
            idx+=1;
         }
      }
   }

   if (flags&0x1)
   {
      idx=0;

      for (k=1;k<=3;k++)
      {
         for (l=(k+1);l<=3;l++)
         {
            for (j=1;j<=3;j++)
            {
               s=pc1[n[k]]*pc1[n[l]]*fs[1][idx];
               sm[0]+=2.0*s*s;
               idx+=1;
            }
         }
      }
   }

   sm[0]/=(double)(T*T);
}


static void fsq_SF_plaq(int T,int bc,int flags,int *n,double *sm)
{
   int p0,k,l,j,idx;
   double s,ph2,r[3];

   for (k=0;k<3;k++)
   {
      for (l=0;l<9;l++)
         fs[k][l]=0.0;
   }

   for (p0=0;p0<T;p0++)
   {
      n[0]=p0;
      ph2=set_Ds(T,bc,n);
      set_Ks(n);
      r[0]=exp(-ts*ph2)*pc0[p0];
      if (p0==0)
         r[0]*=0.5;
      r[1]=r[0]*cxm[p0];
      r[2]=r[0]*sx0[p0];
      r[0]=r[0]*cxp[p0];

      idx=0;

      for (k=1;k<=3;k++)
      {
         for (j=1;j<=3;j++)
         {
            fs[0][idx]+=r[0]*Ks[0][k][0][j];
            fs[1][idx]+=r[1]*Ks[0][k][0][j];
            idx+=1;
         }
      }

      idx=0;

      for (k=1;k<=3;k++)
      {
         for (l=(k+1);l<=3;l++)
         {
            for (j=1;j<=3;j++)
            {
               fs[2][idx]+=r[2]*Ks[k][l][0][j];
               idx+=1;
            }
         }
      }
   }

   sm[0]=0.0;

   if (flags&0x2)
   {
      idx=0;

      for (k=1;k<=3;k++)
      {
         for (j=1;j<=3;j++)
         {
            s=fs[0][idx];
            sm[0]+=s*s;
            s=fs[1][idx];
            sm[0]+=s*s;
            idx+=1;
         }
      }
   }

   if (flags&0x1)
   {
      idx=0;

      for (k=1;k<=3;k++)
      {
         for (l=(k+1);l<=3;l++)
         {
            for (j=1;j<=3;j++)
            {
               s=fs[2][idx];
               sm[0]+=2.0*s*s;
               idx+=1;
            }
         }
      }
   }

   sm[0]/=(double)(T*T);
}


static void fsq_oSF_clov(int T,int bc,int flags,int *n,double *sm)
{
   int p0,k,l,h,j,idx;
   double s,ph2,r[6];

   for (k=0;k<6;k++)
   {
      for (l=0;l<9;l++)
         fs[k][l]=0.0;
   }

   for (p0=0;p0<T;p0++)
   {
      n[0]=p0;
      ph2=set_Ds(T,bc,n);
      set_Ks(n);
      r[0]=exp(-ts*ph2);
      r[1]=r[0]*cx0[p0];
      r[2]=r[0]*sx0[p0]*pc0[p0];
      r[3]=r[1];
      r[4]=r[1]*pc0[p0]*pc0[p0];
      r[5]=r[2];
      r[0]=r[2];

      idx=0;

      for (k=1;k<=3;k++)
      {
         for (h=1;h<=3;h++)
         {
            for (j=(h+1);j<=3;j++)
            {
               fs[0][idx]+=r[0]*Ks[0][k][h][j];
               fs[2][idx]+=r[2]*Ks[0][k][h][j];
               idx+=1;
            }
         }
      }

      idx=0;

      for (k=1;k<=3;k++)
      {
         for (l=(k+1);l<=3;l++)
         {
            for (h=1;h<=3;h++)
            {
               for (j=(h+1);j<=3;j++)
               {
                  fs[1][idx]+=r[1]*Ks[k][l][h][j];
                  fs[3][idx]+=r[3]*Ks[k][l][h][j];
                  idx+=1;
               }
            }
         }
      }

      idx=0;

      for (k=1;k<=3;k++)
      {
         for (j=1;j<=3;j++)
         {
            fs[4][idx]+=r[4]*Ks[0][k][0][j];
            idx+=1;
         }
      }

      idx=0;

      for (k=1;k<=3;k++)
      {
         for (l=(k+1);l<=3;l++)
         {
            for (j=1;j<=3;j++)
            {
               fs[5][idx]+=r[5]*Ks[k][l][0][j];
               idx+=1;
            }
         }
      }
   }

   sm[0]=0.0;
   sm[1]=0.0;

   if (flags&0x2)
   {
      idx=0;

      for (k=1;k<=3;k++)
      {
         for (h=1;h<=3;h++)
         {
            for (j=(h+1);j<=3;j++)
            {
               s=pc1[n[k]]*fs[0][idx];
               sm[0]+=4.0*s*s;

               s=pc1[n[k]]*ph1[n[h]]*fs[2][idx];
               sm[0]-=4.0*c1s*s*s;

               s=pc1[n[k]]*ph1[n[j]]*fs[2][idx];
               sm[0]-=4.0*c1s*s*s;

               idx+=1;
            }
         }
      }
   }

   if (flags&0x1)
   {
      idx=0;

      for (k=1;k<=3;k++)
      {
         for (l=(k+1);l<=3;l++)
         {
            for (h=1;h<=3;h++)
            {
               for (j=(h+1);j<=3;j++)
               {
                  s=pc1[n[k]]*pc1[n[l]]*fs[1][idx];
                  sm[0]+=4.0*s*s;

                  s=pc1[n[k]]*pc1[n[l]]*ph1[n[h]]*fs[3][idx];
                  sm[0]-=4.0*c1s*s*s;

                  s=pc1[n[k]]*pc1[n[l]]*ph1[n[j]]*fs[3][idx];
                  sm[0]-=4.0*c1s*s*s;

                  idx+=1;
               }
            }
         }
      }
   }

   if (flags&0x2)
   {
      idx=0;

      for (k=1;k<=3;k++)
      {
         for (j=1;j<=3;j++)
         {
            s=pc1[n[k]]*fs[4][idx];
            sm[1]+=2.0*s*s;

            idx+=1;
         }
      }
   }

   if (flags&0x1)
   {
      idx=0;

      for (k=1;k<=3;k++)
      {
         for (l=(k+1);l<=3;l++)
         {
            for (j=1;j<=3;j++)
            {
               s=pc1[n[k]]*pc1[n[l]]*fs[5][idx];
               sm[1]+=2.0*s*s;

               idx+=1;
            }
         }
      }
   }

   sm[0]/=(double)(T*T);
   sm[1]/=(double)(T*T);
}


static void fsq_oSF_plaq(int T,int bc,int flags,int *n,double *sm)
{
   int p0,k,l,h,j,idx;
   double s,ph2,r[9];

   for (k=0;k<9;k++)
   {
      for (l=0;l<9;l++)
         fs[k][l]=0.0;
   }

   for (p0=0;p0<T;p0++)
   {
      n[0]=p0;
      ph2=set_Ds(T,bc,n);
      set_Ks(n);
      r[0]=exp(-ts*ph2);
      r[1]=r[0]*sxm[p0];
      r[2]=r[0]*cx0[p0];
      r[3]=r[0]*sxp[p0];
      r[4]=r[1];
      r[5]=r[2];
      r[6]=r[0]*cxp[p0]*pc0[p0];
      r[7]=r[0]*cxm[p0]*pc0[p0];
      r[8]=r[0]*sx0[p0]*pc0[p0];
      r[0]=r[3];

      idx=0;

      for (k=1;k<=3;k++)
      {
         for (h=1;h<=3;h++)
         {
            for (j=(h+1);j<=3;j++)
            {
               fs[0][idx]+=r[0]*Ks[0][k][h][j];
               fs[1][idx]+=r[1]*Ks[0][k][h][j];
               fs[3][idx]+=r[3]*Ks[0][k][h][j];
               fs[4][idx]+=r[4]*Ks[0][k][h][j];
               idx+=1;
            }
         }
      }

      idx=0;

      for (k=1;k<=3;k++)
      {
         for (l=(k+1);l<=3;l++)
         {
            for (h=1;h<=3;h++)
            {
               for (j=(h+1);j<=3;j++)
               {
                  fs[2][idx]+=r[2]*Ks[k][l][h][j];
                  fs[5][idx]+=r[5]*Ks[k][l][h][j];
                  idx+=1;
               }
            }
         }
      }

      idx=0;

      for (k=1;k<=3;k++)
      {
         for (j=1;j<=3;j++)
         {
            fs[6][idx]+=r[6]*Ks[0][k][0][j];
            fs[7][idx]+=r[7]*Ks[0][k][0][j];
            idx+=1;
         }
      }

      idx=0;

      for (k=1;k<=3;k++)
      {
         for (l=(k+1);l<=3;l++)
         {
            for (j=1;j<=3;j++)
            {
               fs[8][idx]+=r[8]*Ks[k][l][0][j];
               idx+=1;
            }
         }
      }
   }

   sm[0]=0.0;
   sm[1]=0.0;

   if (flags&0x2)
   {
      idx=0;

      for (k=1;k<=3;k++)
      {
         for (h=1;h<=3;h++)
         {
            for (j=(h+1);j<=3;j++)
            {
               s=fs[0][idx];
               sm[0]+=2.0*s*s;

               s=fs[1][idx];
               sm[0]+=2.0*s*s;

               s=ph1[n[h]]*fs[3][idx];
               sm[0]-=2.0*c1s*s*s;

               s=ph1[n[j]]*fs[3][idx];
               sm[0]-=2.0*c1s*s*s;

               s=ph1[n[h]]*fs[4][idx];
               sm[0]-=2.0*c1s*s*s;

               s=ph1[n[j]]*fs[4][idx];
               sm[0]-=2.0*c1s*s*s;

               idx+=1;
            }
         }
      }
   }

   if (flags&0x1)
   {
      idx=0;

      for (k=1;k<=3;k++)
      {
         for (l=(k+1);l<=3;l++)
         {
            for (h=1;h<=3;h++)
            {
               for (j=(h+1);j<=3;j++)
               {
                  s=fs[2][idx];
                  sm[0]+=4.0*s*s;

                  s=ph1[n[h]]*fs[5][idx];
                  sm[0]-=4.0*c1s*s*s;

                  s=ph1[n[j]]*fs[5][idx];
                  sm[0]-=4.0*c1s*s*s;

                  idx+=1;
               }
            }
         }
      }
   }

   if (flags&0x2)
   {
      idx=0;

      for (k=1;k<=3;k++)
      {
         for (j=1;j<=3;j++)
         {
            s=fs[6][idx];
            sm[1]+=s*s;

            s=fs[7][idx];
            sm[1]+=s*s;

            idx+=1;
         }
      }
   }

   if (flags&0x1)
   {
      idx=0;

      for (k=1;k<=3;k++)
      {
         for (l=(k+1);l<=3;l++)
         {
            for (j=1;j<=3;j++)
            {
               s=fs[8][idx];
               sm[1]+=2.0*s*s;

               idx+=1;
            }
         }
      }
   }

   sm[0]/=(double)(T*T);
   sm[1]/=(double)(T*T);
}


void rfact(char *type,int N,int T,int L,int bc,double c1,double t,double *r)
{
   int nc,flags,n[4];
   double c0,s0,sm[2];

   nc=strlen(type);

   error_root((nc<1)||(nc>2)||((type[0]!='c')&&(type[0]!='p'))||
              ((nc>1)&&(type[1]!='s')&&(type[1]!='t')),1,"rfact [rfact.c]",
              "Unknown type of running coupling");
   error_root((N<2)||(T<4)||(L<4),1,"rfact [rfact.c]",
              "N,T or L is out of range");
   error_root((bc<1)||(bc>2),1,"rfact [rfact.c]",
              "bc must be 1 or 2");
   error_root((8.0*c1)>=1.0,1,"rfact [rfact.c]",
              "c1 is out of range");
   error_root(t<=0.0,1,"rfact [rfact.c]",
              "t must be positive");

   if (nc==1)
      flags=0x3;
   else if (type[1]=='s')
      flags=0x1;
   else
      flags=0x2;

   if ((T>Ts)||(L>Ls))
      alloc_arrays(T,L);
   set_arrays(T,L,bc);
   c1s=c1;
   ts=t;
   r[0]=0.0;
   r[1]=0.0;
   n[0]=0;

   for (n[1]=0;n[1]<L;n[1]++)
   {
      for (n[2]=0;n[2]<L;n[2]++)
      {
         for (n[3]=0;n[3]<L;n[3]++)
         {
            if (bc==1)
            {
               if (type[0]=='c')
                  fsq_SF_clov(T,bc,flags,n,sm);
               else
                  fsq_SF_plaq(T,bc,flags,n,sm);

               r[0]+=sm[0];
            }
            else
            {
               if (type[0]=='c')
                  fsq_oSF_clov(T,bc,flags,n,sm);
               else
                  fsq_oSF_plaq(T,bc,flags,n,sm);

               r[0]+=sm[0];
               r[1]+=sm[1];
            }
         }
      }
   }

   c0=1.0-8.0*c1;
   s0=t*t*(double)(N*N-1)/(double)(L*L*L);

   if (bc==1)
      r[0]*=2.0*c0*s0;
   else
   {
      r[0]*=0.25*s0;
      r[1]*=c0*s0;
   }
}
