
/*******************************************************************************
*
* File binomial.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Binomial coefficients and factorials.
*
*   long binomial(int n,int m)
*     Returns the binomial coefficient n!/(m!*(n-m)!) if n>=m>=0 and
*     if the calculated value is less than or equal to LONG_MAX. An
*     error occurs in all other cases.
*
*   long factorial(int n)
*     Returns n! if n>=0 and the calculated value is less than or equal
*     to LONG_MAX. An error occurs in all other cases.
*
*   long factorial2(int n)
*     Returns n!! if n>=0 and the calculated value is less than or equal
*     to LONG_MAX. An error occurs in all other cases.
*
* On x86-64 machines, LONG_MAX is 2^63-1 and thus about 9.2x10^18.
*
* The programs in this module do not perform any communications and can be
* locally called.
*
*******************************************************************************/

#define BINOMIAL_C

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "utils.h"
#include "permutations.h"

static int nmxb=0,nmxf=0,nmxf2=0;
static long **binom,*fact,*fact2;


static void set_binom(int n)
{
   int i,j;
   long b1,b2;
   long **bb,*b;

   bb=malloc((n+1)*sizeof(*bb));
   b=malloc(((n*(n+1))/2+n+1)*sizeof(*b));

   error_loc((bb==NULL)||(b==NULL),1,"set_binom [binomial.c]",
             "Unable to allocate auxiliary arrays");

   bb[0]=b;

   for (i=1;i<=n;i++)
      bb[i]=bb[i-1]+i;

   if (nmxb>0)
   {
      for (i=0;i<=nmxb;i++)
      {
         for (j=0;j<=i;j++)
            bb[i][j]=binom[i][j];
      }

      free(binom[0]);
      free(binom);
   }
   else
      bb[0][0]=1L;

   binom=bb;

   for (i=nmxb+1;i<=n;i++)
   {
      binom[i][0]=1L;
      binom[i][i]=1L;

      for (j=1;j<i;j++)
      {
         b1=binom[i-1][j-1];
         b2=binom[i-1][j];

         if (b1<=(LONG_MAX-b2))
            binom[i][j]=b1+b2;
         else
            error_loc(1,1,"set_binom [binomial.c]",
                      "Binomial coefficient is out of range");
      }
   }

   nmxb=n;
}


long binomial(int n,int m)
{
   error_loc((n<m)||(m<0),1,"binomial [binomial.c]",
             "Arguments are out of range");

   if (n==0)
      return 1L;

   if (n>nmxb)
      set_binom(n);

   return binom[n][m];
}


static void set_fact(void)
{
   int i;
   long f;

   f=1L;

   for (i=2;(LONG_MAX/(long)(i))>=f;i++)
      f*=(long)(i);

   nmxf=i-1;
   fact=malloc((nmxf+1)*sizeof(*fact));
   error_loc(fact==NULL,1,"set_fact [binomial.c]",
             "Unable to allocate auxiliary array");

   fact[0]=1L;

   for (i=1;i<=nmxf;i++)
      fact[i]=(long)(i)*fact[i-1];
}


long factorial(int n)
{
   if (nmxf==0)
      set_fact();

   error_loc((n<0)||(n>nmxf),1,"factorial [binomial.c]",
             "Argument is out of range");

   return fact[n];
}


static void set_fact2(void)
{
   int i;
   long f[2];

   f[0]=1L;
   f[1]=1L;

   for (i=2;(LONG_MAX/(long)(i))>=f[i&0x1];i++)
      f[i&0x1]*=(long)(i);

   nmxf2=i-1;
   fact2=malloc((nmxf2+1)*sizeof(*fact2));
   error_loc(fact2==NULL,1,"set_fact2 [binomial.c]",
             "Unable to allocate auxiliary array");

   fact2[0]=1L;
   fact2[1]=1L;

   for (i=2;i<=nmxf2;i++)
      fact2[i]=(long)(i)*fact2[i-2];
}


long factorial2(int n)
{
   if (nmxf2==0)
      set_fact2();

   error_loc((n<0)||(n>nmxf2),1,"factorial2 [binomial.c]",
             "Argument is out of range");

   return fact2[n];
}
