
/*******************************************************************************
*
* File mindex.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generation of multi-indices with specified index ranges.
*
*   long init_mindex(int m,int *n,int (*r)[2],int *cid)
*     Initializes a new instance of a multi-index counter and returns the
*     number of index vectors of the type specified by the arguments m,n
*     and r (see the notes). The id of the new counter is assigned to cid.
*
*   long next_mindex(int cid,int *i)
*     Assigns the next index vector generated by the multi-index counter
*     with id cid to the array i and returns the current index count (see
*     the notes). The program wraps around when all possible index vectors
*     have been generated.
*
*   void stop_mindex(int cid)
*     Resets the multi-index generator with id cid to its uninitialized
*     state.
*
* The multi-index generators run through vectors of nt integers that
* are logically divided into m groups of size
*
*   n[0],n[1],..,n[m-1],
*
*   nt=n[0]+n[1]+..+n[m-1].
*
* In the k'th group, all indices range over the interval [r[k][0],r[k][1]].
* Moreover, the indices i[0],..,i[n[k]-1] in this group are ordered such
* that
*
*   i[0]>=i[1]>=..>=i[n[k]-1].
*
* The indices are otherwise unconstrained.
*
* The input arguments of the initialization program init_mindex() are
*
*  m          Number of groups.
*
*  n[k]       Number of indices in the k'th group (k=0,1,..,m-1).
*
*  r[k][0]    Bounds of the interval over which the indices in the
*  r[k][1]    k'th group range.
*
* They must satisfy the constraints
*
*  m>0 and n[k]>0, r[k][0]<=r[k][1] for all k=0,..,m-1,
*
* and be such that the number of index vectors is less than or equal to
* LONG_MAX (on x86-64 machines, this is 2^63-1 and thus about 9.2x10^18).
* The program init_mindex() aborts with an error message if these
* conditions are not satisfied.
*
* The program next_mindex() cycles through the index vectors like m nested
* for-loops, where the innermost loop runs over the indices in the first
* group of n[0] indices, the next-to-innermost loop over the indices in the
* next group of n[1] numbers, and so on.
*
* At any given time, there can be at most 16 running generators. Running
* generators should therefore be stopped by calling stop_mindex() when they
* are no longer used.
*
* The programs in this module do not perform any communications and can be
* locally called.
*
*******************************************************************************/

#define MINDEX_C
#define CNT_MAX 16

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "utils.h"
#include "permutations.h"

typedef struct
{
   int init;
   int ma,nta,m,nt;
   int *n,(*r)[2];
   int **i;
   long ni,nim;
} counter_t;

static int ncnt=0;
static counter_t cnt[CNT_MAX];


static counter_t *set_cid(int *cid)
{
   int ic;

   for (ic=0;ic<ncnt;ic++)
   {
      if (cnt[ic].init==0)
      {
         (*cid)=ic;
         return cnt+ic;
      }
   }

   error_loc(ncnt==CNT_MAX,1,"set_cid [mindex.c]",
             "Requested too many index generators");

   ncnt+=1;
   (*cid)=ic;
   cnt[ic].ma=0;
   cnt[ic].nta=0;

   return cnt+ic;
}


static void alloc_arrays(int m,int *n,int *cid)
{
   int ma,nta,nt,k;
   int **is,*ns,(*rs)[2];
   counter_t *c;

   c=set_cid(cid);
   ma=(*c).ma;
   nta=(*c).nta;
   nt=0;

   for (k=0;k<m;k++)
      nt+=n[k];

   if ((m>ma)||(nt>nta))
   {
      if (ma>0)
      {
         free((*c).i);
         free((*c).n);
         free((*c).r);
      }

      if (m>ma)
         ma=m;
      if (nt>nta)
         nta=nt;

      (*c).ma=ma;
      (*c).nta=nta;

      is=malloc(ma*sizeof(*is));
      ns=malloc((ma+nta)*sizeof(*ns));
      rs=malloc(ma*sizeof(*rs));

      error_loc((is==NULL)||(ns==NULL)||(rs==NULL),1,
                "alloc_arrays [mindex.c]",
                "Unable to allocate auxiliary arrays");

      (*c).i=is;
      (*c).n=ns;
      (*c).r=rs;
   }

   (*c).m=m;
   (*c).nt=nt;
   (*c).i[0]=(*c).n+m;

   for (k=0;k<m;k++)
   {
      (*c).n[k]=n[k];

      if (k>0)
         (*c).i[k]=(*c).i[k-1]+n[k-1];
   }
}


static void set_arrays(int cid,int (*r)[2])
{
   int ms,k,l;
   int **is,*ns;
   counter_t *c;

   c=cnt+cid;
   ms=(*c).m;
   ns=(*c).n;
   is=(*c).i;

   for (k=0;k<ms;k++)
   {
      (*c).r[k][0]=r[k][0];
      (*c).r[k][1]=r[k][1];

      for (l=0;l<ns[k];l++)
         is[k][l]=r[k][0];
   }
}


static void normalize(int *u,int *v)
{
   int x,y,r;

   x=(*u);
   y=(*v);

   while (y!=0)
   {
      r=(x%y);
      x=y;
      y=r;
   }

   (*u)/=x;
   (*v)/=x;
}


static void set_nim(int cid)
{
   int ms,*ns,(*rs)[2];
   int k,l,n,r,u,v;
   long nim;

   ms=cnt[cid].m;
   ns=cnt[cid].n;
   rs=cnt[cid].r;
   nim=1L;

   for (k=0;k<ms;k++)
   {
      n=ns[k];
      r=rs[k][1]-rs[k][0];

      for (l=1;l<=n;l++)
      {
         u=r+l;
         v=l;

         normalize(&u,&v);
         nim/=(long)(v);

         if (nim<=(LONG_MAX/(long)(u)))
            nim*=(long)(u);
         else
            error_loc(1,1,"set_nim [mindex.c]",
                      "Number of index vectors is out of range");
      }
   }

   cnt[cid].nim=nim;
}


static void assign_idx(int cid,int *i)
{
   int ms,*ns,**is;
   int k,l;

   ms=cnt[cid].m;
   ns=cnt[cid].n;
   is=cnt[cid].i;

   for (k=0;k<ms;k++)
   {
      for (l=0;l<ns[k];l++)
      {
         i[0]=is[k][l];
         i+=1;
      }
   }
}


static int cycle(int cid,int k,int l)
{
   int *is;

   is=cnt[cid].i[k];

   if (l>0)
   {
      if (is[l]<is[l-1])
      {
         is[l]+=1;
         return 1;
      }
      else
      {
         is[l]=cnt[cid].r[k][0];
         return cycle(cid,k,l-1);
      }
   }
   else
   {
      if (is[0]<cnt[cid].r[k][1])
      {
         is[0]+=1;
         return 1;
      }
      else
      {
         is[0]=cnt[cid].r[k][0];

         if (k<(cnt[cid].m-1))
            return cycle(cid,k+1,cnt[cid].n[k+1]-1);
         else
            return 0;
      }
   }
}


long init_mindex(int m,int *n,int (*r)[2],int *cid)
{
   int k;

   error_loc(m<1,1,"init_mindex [mindex.c]",
             "There must be at least one group of indices");

   for (k=0;k<m;k++)
   {
      if ((n[k]<1)||(r[k][0]>r[k][1]))
         error_loc(1,1,"init_mindex [mindex.c]",
                   "Arguments are out of range");
   }

   alloc_arrays(m,n,cid);
   set_arrays(*cid,r);
   set_nim(*cid);

   cnt[*cid].ni=0L;
   cnt[*cid].init=1;

   return cnt[*cid].nim;
}


long next_mindex(int cid,int *i)
{
   int ic;

   if ((cid>=ncnt)||(cid<0)||(cnt[cid].init!=1))
      error_loc(1,1,"next_mindex [mindex.c]",
                "Index generator does not exist or is not initialized");

   assign_idx(cid,i);
   ic=cycle(cid,0,cnt[cid].n[0]-1);
   cnt[cid].ni+=1L;

   if (ic==1)
      return cnt[cid].ni;
   else
   {
      cnt[cid].ni=0L;
      return cnt[cid].nim;
   }
}


void stop_mindex(int cid)
{
   if ((cid<ncnt)&&(cid>=0)&&(cnt[cid].init==1))
      cnt[cid].init=0;
   else
      error_loc(1,1,"stop_mindex [mindex.c]",
                "Index generator does not existent or is already stopped");
}
