
/*******************************************************************************
*
* File sl3fcts.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Functions for matrices of type sl3_alg_dble.
*
*   void sl3_zero(sl3_alg_dble *a)
*     Sets the matrix (*a) to zero.
*
*   void sl3_add(sl3_alg_dble *a,sl3_alg_dble *b)
*     Adds (*a) to (*b). The input matrix is unchanged unless a=b.
*
*   void sl3_sub(sl3_alg_dble *a,sl3_alg_dble *b)
*     Subtracts (*a) from (*b). The input matrix is unchanged unless
*     a=b.
*
*   void sl3_mulr(double *r,sl3_alg_dble *a,sl3_alg_dble *b)
*     Assigns (*r)*(*a) to (*b). The input matrix is unchanged
*     unless a=b.
*
*   void sl3_mulr_add(double *r,sl3_alg_dble *a,sl3_alg_dble *b)
*     Adds (*r)*(*a) to (*b). The input matrix is unchanged
*     unless a=b.
*
*   void sl3_to_su3(sl3_alg_dble *a,su3_dble *u)
*     Assigns the matrix (*a) to (*u).
*
*   void su3_to_sl3(su3_dble *u,sl3_alg_dble *a)
*     Assigns the traceless part (*u)-(1/3)*tr(*u) of the matrix
*     (*u) to (*a).
*
*   void su3_mulr_add_sl3(double *r,su3_dble *u,sl3_alg_dble *a)
*     Computes the matrix r*(*u) and adds its traceless part to (*a).
*
*   void pexp_sl3_alg(int n,sl3_alg_dble *a,su3_dble *u)
*     Expands exp(A), A=a[0]*x+a[1]*x^2+..+a[n-1]*x^n, in the real
*     variable x up to order x^n and assigns the coefficient of x^k
*     to u[k-1] (k=1,2,..,n).
*
*   void mexp_sl3_alg(int n,sl3_alg_dble *a,su3_dble *u)
*     Expands exp(-A), A=a[0]*x+a[1]*x^2+..+a[n-1]*x^n, in the real
*     variable x up to order x^n and assigns the coefficient of x^k
*     to u[k-1] (k=1,2,..,n).
*
*   complex_dble sl3_biform(sl3_alg_dble *a,sl3_alg_dble *b)
*     Returns -2*tr{a*b}.
*
* The programs in this module do not perform any communications and can be
* called locally.
*
* If SSE2 instructions are used, it is assumed that the matrices and complex
* coefficients are aligned to a 16 byte boundary.
*
*******************************************************************************/

#define SL3FCTS_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "utils.h"
#include "su3fcts.h"

static int ns=0;
static double *ck1,*ck2;
static su3_dble *Aw;


static void alloc_arrays(int n)
{
   int k;

   if (ns>0)
   {
      free(ck1);
      afree(Aw);
   }

   ck1=malloc(2*(n+1)*sizeof(*ck1));
   ck2=ck1+n+1;
   Aw=amalloc(n*sizeof(*Aw),4);

   error_loc((ck1==NULL)||(Aw==NULL),1,"alloc_arrays [sl3fcts.c]",
             "Unable to allocate auxiliary arrays");

   ck1[0]=1.0;
   ck2[0]=1.0;

   for (k=1;k<=n;k++)
   {
      ck1[k]=ck1[k-1]/(double)(k);

      if (k&0x1)
         ck2[k]=-ck1[k];
      else
         ck2[k]=ck1[k];
   }

   cm3x3_zero(n,Aw);
   ns=n;
}

#if (defined x64)
#include "sse2.h"

static const sse_double c13={1.0/3.0,1.0/3.0};


void sl3_zero(sl3_alg_dble *a)
{
   __asm__ __volatile__ ("xorpd %%xmm0, %%xmm0 \n\t"
                         "xorpd %%xmm1, %%xmm1 \n\t"
                         "movapd %%xmm0, %0 \n\t"
                         "movapd %%xmm1, %1 \n\t"
                         "movapd %%xmm0, %2 \n\t"
                         "movapd %%xmm1, %3 \n\t"
                         "movapd %%xmm0, %4 \n\t"
                         "movapd %%xmm1, %5 \n\t"
                         "movapd %%xmm0, %6 \n\t"
                         "movapd %%xmm1, %7"
                         :
                         "=m" ((*a).c1),
                         "=m" ((*a).c2),
                         "=m" ((*a).c3),
                         "=m" ((*a).c4),
                         "=m" ((*a).c5),
                         "=m" ((*a).c6),
                         "=m" ((*a).c7),
                         "=m" ((*a).c8)
                         :
                         :
                         "xmm0", "xmm1");
}


void sl3_add(sl3_alg_dble *a,sl3_alg_dble *b)
{
   __asm__ __volatile__ ("movapd %0, %%xmm0 \n\t"
                         "movapd %1, %%xmm1 \n\t"
                         "movapd %2, %%xmm2 \n\t"
                         "movapd %3, %%xmm3"
                         :
                         :
                         "m" ((*b).c1),
                         "m" ((*b).c2),
                         "m" ((*b).c3),
                         "m" ((*b).c4)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm3");

   __asm__ __volatile__ ("addpd %0, %%xmm0 \n\t"
                         "addpd %1, %%xmm1 \n\t"
                         "addpd %2, %%xmm2 \n\t"
                         "addpd %3, %%xmm3"
                         :
                         :
                         "m" ((*a).c1),
                         "m" ((*a).c2),
                         "m" ((*a).c3),
                         "m" ((*a).c4)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm3");

   __asm__ __volatile__ ("movapd %4, %%xmm4 \n\t"
                         "movapd %5, %%xmm5 \n\t"
                         "movapd %6, %%xmm6 \n\t"
                         "movapd %7, %%xmm7 \n\t"
                         "movapd %%xmm0, %0\n\t"
                         "movapd %%xmm1, %1\n\t"
                         "movapd %%xmm2, %2\n\t"
                         "movapd %%xmm3, %3"
                         :
                         "=m" ((*b).c1),
                         "=m" ((*b).c2),
                         "=m" ((*b).c3),
                         "=m" ((*b).c4)
                         :
                         "m" ((*b).c5),
                         "m" ((*b).c6),
                         "m" ((*b).c7),
                         "m" ((*b).c8)
                         :
                         "xmm4", "xmm5", "xmm6", "xmm7");

   __asm__ __volatile__ ("addpd %0, %%xmm4 \n\t"
                         "addpd %1, %%xmm5 \n\t"
                         "addpd %2, %%xmm6 \n\t"
                         "addpd %3, %%xmm7"
                         :
                         :
                         "m" ((*a).c5),
                         "m" ((*a).c6),
                         "m" ((*a).c7),
                         "m" ((*a).c8)
                         :
                         "xmm4", "xmm5", "xmm6", "xmm7");

   __asm__ __volatile__ ("movapd %%xmm4, %0\n\t"
                         "movapd %%xmm5, %1\n\t"
                         "movapd %%xmm6, %2\n\t"
                         "movapd %%xmm7, %3"
                         :
                         "=m" ((*b).c5),
                         "=m" ((*b).c6),
                         "=m" ((*b).c7),
                         "=m" ((*b).c8));
}


void sl3_sub(sl3_alg_dble *a,sl3_alg_dble *b)
{
   __asm__ __volatile__ ("movapd %0, %%xmm0 \n\t"
                         "movapd %1, %%xmm1 \n\t"
                         "movapd %2, %%xmm2 \n\t"
                         "movapd %3, %%xmm3"
                         :
                         :
                         "m" ((*b).c1),
                         "m" ((*b).c2),
                         "m" ((*b).c3),
                         "m" ((*b).c4)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm3");

   __asm__ __volatile__ ("subpd %0, %%xmm0 \n\t"
                         "subpd %1, %%xmm1 \n\t"
                         "subpd %2, %%xmm2 \n\t"
                         "subpd %3, %%xmm3"
                         :
                         :
                         "m" ((*a).c1),
                         "m" ((*a).c2),
                         "m" ((*a).c3),
                         "m" ((*a).c4)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm3");

   __asm__ __volatile__ ("movapd %4, %%xmm4 \n\t"
                         "movapd %5, %%xmm5 \n\t"
                         "movapd %6, %%xmm6 \n\t"
                         "movapd %7, %%xmm7 \n\t"
                         "movapd %%xmm0, %0\n\t"
                         "movapd %%xmm1, %1\n\t"
                         "movapd %%xmm2, %2\n\t"
                         "movapd %%xmm3, %3"
                         :
                         "=m" ((*b).c1),
                         "=m" ((*b).c2),
                         "=m" ((*b).c3),
                         "=m" ((*b).c4)
                         :
                         "m" ((*b).c5),
                         "m" ((*b).c6),
                         "m" ((*b).c7),
                         "m" ((*b).c8)
                         :
                         "xmm4", "xmm5", "xmm6", "xmm7");

   __asm__ __volatile__ ("subpd %0, %%xmm4 \n\t"
                         "subpd %1, %%xmm5 \n\t"
                         "subpd %2, %%xmm6 \n\t"
                         "subpd %3, %%xmm7"
                         :
                         :
                         "m" ((*a).c5),
                         "m" ((*a).c6),
                         "m" ((*a).c7),
                         "m" ((*a).c8)
                         :
                         "xmm4", "xmm5", "xmm6", "xmm7");

   __asm__ __volatile__ ("movapd %%xmm4, %0\n\t"
                         "movapd %%xmm5, %1\n\t"
                         "movapd %%xmm6, %2\n\t"
                         "movapd %%xmm7, %3"
                         :
                         "=m" ((*b).c5),
                         "=m" ((*b).c6),
                         "=m" ((*b).c7),
                         "=m" ((*b).c8));
}


void sl3_mulr(double *r,sl3_alg_dble *a,sl3_alg_dble *b)
{
   __asm__ __volatile__ ("movddup %0, %%xmm0 \n\t"
                         "movddup %0, %%xmm1"
                         :
                         :
                         "m" (*r)
                         :
                         "xmm0", "xmm1");

   __asm__ __volatile__ ("movapd %0, %%xmm2 \n\t"
                         "movapd %1, %%xmm3 \n\t"
                         "movapd %2, %%xmm4 \n\t"
                         "movapd %3, %%xmm5 \n\t"
                         "mulpd %%xmm0, %%xmm2 \n\t"
                         "mulpd %%xmm1, %%xmm3 \n\t"
                         "mulpd %%xmm0, %%xmm4 \n\t"
                         "mulpd %%xmm1, %%xmm5"
                         :
                         :
                         "m" ((*a).c1),
                         "m" ((*a).c2),
                         "m" ((*a).c3),
                         "m" ((*a).c4)
                         :
                         "xmm2", "xmm3", "xmm4", "xmm5");

   __asm__ __volatile__ ("movapd %4, %%xmm6 \n\t"
                         "movapd %5, %%xmm7 \n\t"
                         "movapd %6, %%xmm8 \n\t"
                         "movapd %7, %%xmm9 \n\t"
                         "movapd %%xmm2, %0\n\t"
                         "movapd %%xmm3, %1\n\t"
                         "movapd %%xmm4, %2\n\t"
                         "movapd %%xmm5, %3\n\t"
                         "mulpd %%xmm0, %%xmm6 \n\t"
                         "mulpd %%xmm1, %%xmm7 \n\t"
                         "mulpd %%xmm0, %%xmm8 \n\t"
                         "mulpd %%xmm1, %%xmm9"
                         :
                         "=m" ((*b).c1),
                         "=m" ((*b).c2),
                         "=m" ((*b).c3),
                         "=m" ((*b).c4)
                         :
                         "m" ((*a).c5),
                         "m" ((*a).c6),
                         "m" ((*a).c7),
                         "m" ((*a).c8)
                         :
                         "xmm6", "xmm7", "xmm8", "xmm9");

   __asm__ __volatile__ ("movapd %%xmm6, %0\n\t"
                         "movapd %%xmm7, %1\n\t"
                         "movapd %%xmm8, %2\n\t"
                         "movapd %%xmm9, %3"
                         :
                         "=m" ((*b).c5),
                         "=m" ((*b).c6),
                         "=m" ((*b).c7),
                         "=m" ((*b).c8));
}


void sl3_mulr_add(double *r,sl3_alg_dble *a,sl3_alg_dble *b)
{
   __asm__ __volatile__ ("movddup %0, %%xmm0 \n\t"
                         "movddup %0, %%xmm1"
                         :
                         :
                         "m" (*r)
                         :
                         "xmm0", "xmm1");

   __asm__ __volatile__ ("movapd %0, %%xmm2 \n\t"
                         "movapd %1, %%xmm3 \n\t"
                         "movapd %2, %%xmm4 \n\t"
                         "movapd %3, %%xmm5 \n\t"
                         "mulpd %%xmm0, %%xmm2 \n\t"
                         "mulpd %%xmm1, %%xmm3 \n\t"
                         "mulpd %%xmm0, %%xmm4 \n\t"
                         "mulpd %%xmm1, %%xmm5"
                         :
                         :
                         "m" ((*a).c1),
                         "m" ((*a).c2),
                         "m" ((*a).c3),
                         "m" ((*a).c4)
                         :
                         "xmm2", "xmm3", "xmm4", "xmm5");

   __asm__ __volatile__ ("addpd %0, %%xmm2 \n\t"
                         "addpd %1, %%xmm3 \n\t"
                         "addpd %2, %%xmm4 \n\t"
                         "addpd %3, %%xmm5"
                         :
                         :
                         "m" ((*b).c1),
                         "m" ((*b).c2),
                         "m" ((*b).c3),
                         "m" ((*b).c4)
                         :
                         "xmm2", "xmm3", "xmm4", "xmm5");

   __asm__ __volatile__ ("movapd %4, %%xmm6 \n\t"
                         "movapd %5, %%xmm7 \n\t"
                         "movapd %6, %%xmm8 \n\t"
                         "movapd %7, %%xmm9 \n\t"
                         "movapd %%xmm2, %0\n\t"
                         "movapd %%xmm3, %1\n\t"
                         "movapd %%xmm4, %2\n\t"
                         "movapd %%xmm5, %3\n\t"
                         "mulpd %%xmm0, %%xmm6 \n\t"
                         "mulpd %%xmm1, %%xmm7 \n\t"
                         "mulpd %%xmm0, %%xmm8 \n\t"
                         "mulpd %%xmm1, %%xmm9"
                         :
                         "=m" ((*b).c1),
                         "=m" ((*b).c2),
                         "=m" ((*b).c3),
                         "=m" ((*b).c4)
                         :
                         "m" ((*a).c5),
                         "m" ((*a).c6),
                         "m" ((*a).c7),
                         "m" ((*a).c8)
                         :
                         "xmm6", "xmm7", "xmm8", "xmm9");

   __asm__ __volatile__ ("addpd %0, %%xmm6 \n\t"
                         "addpd %1, %%xmm7 \n\t"
                         "addpd %2, %%xmm8 \n\t"
                         "addpd %3, %%xmm9"
                         :
                         :
                         "m" ((*b).c5),
                         "m" ((*b).c6),
                         "m" ((*b).c7),
                         "m" ((*b).c8)
                         :
                         "xmm6", "xmm7", "xmm8", "xmm9");

   __asm__ __volatile__ ("movapd %%xmm6, %0\n\t"
                         "movapd %%xmm7, %1\n\t"
                         "movapd %%xmm8, %2\n\t"
                         "movapd %%xmm9, %3"
                         :
                         "=m" ((*b).c5),
                         "=m" ((*b).c6),
                         "=m" ((*b).c7),
                         "=m" ((*b).c8));
}


void sl3_to_su3(sl3_alg_dble *a,su3_dble *u)
{
   __asm__ __volatile__ ("xorpd %%xmm8, %%xmm8 \n\t"
                         "movapd %0, %%xmm0 \n\t"
                         "movapd %1, %%xmm1 \n\t"
                         "movapd %2, %%xmm2 \n\t"
                         "movapd %3, %%xmm3 \n\t"
                         "movapd %4, %%xmm4 \n\t"
                         "movapd %5, %%xmm5 \n\t"
                         "movapd %6, %%xmm6 \n\t"
                         "movapd %7, %%xmm7 \n\t"
                         "subpd %%xmm0, %%xmm8 \n\t"
                         "subpd %%xmm1, %%xmm8"
                         :
                         :
                         "m" ((*a).c1),
                         "m" ((*a).c2),
                         "m" ((*a).c3),
                         "m" ((*a).c4),
                         "m" ((*a).c5),
                         "m" ((*a).c6),
                         "m" ((*a).c7),
                         "m" ((*a).c8)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm3",
                         "xmm4", "xmm5", "xmm6", "xmm7",
                         "xmm8");

   __asm__ __volatile__ ("movapd %%xmm0, %0 \n\t"
                         "movapd %%xmm1, %1 \n\t"
                         "movapd %%xmm2, %2 \n\t"
                         "movapd %%xmm3, %3"
                         :
                         "=m" ((*u).c11),
                         "=m" ((*u).c22),
                         "=m" ((*u).c12),
                         "=m" ((*u).c21));

   __asm__ __volatile__ ("movapd %%xmm8, %0 \n\t"
                         "movapd %%xmm4, %1 \n\t"
                         "movapd %%xmm5, %2 \n\t"
                         "movapd %%xmm6, %3 \n\t"
                         "movapd %%xmm7, %4"
                         :
                         "=m" ((*u).c33),
                         "=m" ((*u).c13),
                         "=m" ((*u).c31),
                         "=m" ((*u).c23),
                         "=m" ((*u).c32));
}


void su3_to_sl3(su3_dble *u,sl3_alg_dble *a)
{
   __asm__ __volatile__ ("movapd %0, %%xmm0 \n\t"
                         "movapd %1, %%xmm1 \n\t"
                         "movapd %2, %%xmm2 \n\t"
                         "movapd %3, %%xmm3"
                         :
                         :
                         "m" ((*u).c33),
                         "m" ((*u).c11),
                         "m" ((*u).c12),
                         "m" ((*u).c13)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm3");

   __asm__ __volatile__ ("addpd %%xmm1, %%xmm0 \n\t"
                         "movapd %0, %%xmm4 \n\t"
                         "movapd %1, %%xmm5 \n\t"
                         "movapd %2, %%xmm6 \n\t"
                         "movapd %3, %%xmm7 \n\t"
                         "addpd %%xmm5, %%xmm0 \n\t"
                         "movapd %4, %%xmm8"
                         :
                         :
                         "m" ((*u).c21),
                         "m" ((*u).c22),
                         "m" ((*u).c23),
                         "m" ((*u).c31),
                         "m" ((*u).c32)
                         :
                         "xmm0", "xmm4", "xmm5", "xmm6",
                         "xmm7", "xmm8");

   __asm__ __volatile__ ("movapd %%xmm2, %0 \n\t"
                         "movapd %%xmm3, %1 \n\t"
                         "movapd %%xmm4, %2 \n\t"
                         "mulpd %3, %%xmm0"
                         :
                         "=m" ((*a).c3),
                         "=m" ((*a).c5),
                         "=m" ((*a).c4)
                         :
                         "m" (c13)
                         :
                         "xmm0");

   __asm__ __volatile__ ("movapd %%xmm6, %0 \n\t"
                         "movapd %%xmm7, %1 \n\t"
                         "movapd %%xmm8, %2 \n\t"
                         "subpd %%xmm0, %%xmm1 \n\t"
                         "subpd %%xmm0, %%xmm5 \n\t"
                         "movapd %%xmm1, %3 \n\t"
                         "movapd %%xmm5, %4"
                         :
                         "=m" ((*a).c7),
                         "=m" ((*a).c6),
                         "=m" ((*a).c8),
                         "=m" ((*a).c1),
                         "=m" ((*a).c2)
                         :
                         :
                         "xmm1", "xmm5");
}


void su3_mulr_add_sl3(double *r,su3_dble *u,sl3_alg_dble *a)
{
   __asm__ __volatile__ ("movddup %0, %%xmm0 \n\t"
                         "movddup %0, %%xmm1 \n\t"
                         "movddup %0, %%xmm2"
                         :
                         :
                         "m" (*r)
                         :
                         "xmm0", "xmm1", "xmm2");

   __asm__ __volatile__ ("movapd %0, %%xmm3 \n\t"
                         "movapd %1, %%xmm4 \n\t"
                         "movapd %2, %%xmm5 \n\t"
                         "mulpd %%xmm0, %%xmm3 \n\t"
                         "mulpd %%xmm1, %%xmm4 \n\t"
                         "mulpd %%xmm2, %%xmm5"
                         :
                         :
                         "m" ((*u).c11),
                         "m" ((*u).c12),
                         "m" ((*u).c13)
                         :
                         "xmm3", "xmm4", "xmm5");

   __asm__ __volatile__ ("movapd %0, %%xmm6 \n\t"
                         "movapd %1, %%xmm7 \n\t"
                         "movapd %2, %%xmm8 \n\t"
                         "mulpd %%xmm0, %%xmm6 \n\t"
                         "mulpd %%xmm1, %%xmm7 \n\t"
                         "mulpd %%xmm2, %%xmm8"
                         :
                         :
                         "m" ((*u).c21),
                         "m" ((*u).c22),
                         "m" ((*u).c23)
                         :
                         "xmm6", "xmm7", "xmm8");

   __asm__ __volatile__ ("movapd %0, %%xmm9 \n\t"
                         "movapd %1, %%xmm10 \n\t"
                         "movapd %2, %%xmm11 \n\t"
                         "mulpd %%xmm0, %%xmm9 \n\t"
                         "mulpd %%xmm1, %%xmm10 \n\t"
                         "mulpd %%xmm2, %%xmm11"
                         :
                         :
                         "m" ((*u).c31),
                         "m" ((*u).c32),
                         "m" ((*u).c33)
                         :
                         "xmm9", "xmm10", "xmm11");

   __asm__ __volatile__ ("addpd %0, %%xmm4 \n\t"
                         "addpd %1, %%xmm5 \n\t"
                         "addpd %%xmm3, %%xmm11 \n\t"
                         "addpd %2, %%xmm6 \n\t"
                         "addpd %3, %%xmm8 \n\t"
                         "addpd %%xmm7, %%xmm11 \n\t"
                         "addpd %4, %%xmm9 \n\t"
                         "addpd %5, %%xmm10 \n\t"
                         "mulpd %6, %%xmm11"
                         :
                         :
                         "m" ((*a).c3),
                         "m" ((*a).c5),
                         "m" ((*a).c4),
                         "m" ((*a).c7),
                         "m" ((*a).c6),
                         "m" ((*a).c8),
                         "m" (c13)
                         :
                         "xmm4", "xmm5", "xmm6", "xmm8",
                         "xmm9", "xmm10", "xmm11");

   __asm__ __volatile__ ("addpd %0, %%xmm3 \n\t"
                         "addpd %1, %%xmm7 \n\t"
                         "subpd %%xmm11, %%xmm3 \n\t"
                         "subpd %%xmm11, %%xmm7"
                         :
                         :
                         "m" ((*a).c1),
                         "m" ((*a).c2)
                         :
                         "xmm3", "xmm7");

   __asm__ __volatile__ ("movapd %%xmm4, %0\n\t"
                         "movapd %%xmm5, %1\n\t"
                         "movapd %%xmm6, %2\n\t"
                         "movapd %%xmm8, %3\n\t"
                         "movapd %%xmm9, %4\n\t"
                         "movapd %%xmm10, %5\n\t"
                         "movapd %%xmm3, %6\n\t"
                         "movapd %%xmm7, %7"
                         :
                         "=m" ((*a).c3),
                         "=m" ((*a).c5),
                         "=m" ((*a).c4),
                         "=m" ((*a).c7),
                         "=m" ((*a).c6),
                         "=m" ((*a).c8),
                         "=m" ((*a).c1),
                         "=m" ((*a).c2));
}

#else

static const double c13=1.0/3.0;

void sl3_zero(sl3_alg_dble *a)
{
   (*a).c1.re=0.0;
   (*a).c1.im=0.0;
   (*a).c2.re=0.0;
   (*a).c2.im=0.0;

   (*a).c3.re=0.0;
   (*a).c3.im=0.0;
   (*a).c4.re=0.0;
   (*a).c4.im=0.0;

   (*a).c5.re=0.0;
   (*a).c5.im=0.0;
   (*a).c6.re=0.0;
   (*a).c6.im=0.0;

   (*a).c7.re=0.0;
   (*a).c7.im=0.0;
   (*a).c8.re=0.0;
   (*a).c8.im=0.0;
}


void sl3_add(sl3_alg_dble *a,sl3_alg_dble *b)
{
   (*b).c1.re+=(*a).c1.re;
   (*b).c1.im+=(*a).c1.im;
   (*b).c2.re+=(*a).c2.re;
   (*b).c2.im+=(*a).c2.im;

   (*b).c3.re+=(*a).c3.re;
   (*b).c3.im+=(*a).c3.im;
   (*b).c4.re+=(*a).c4.re;
   (*b).c4.im+=(*a).c4.im;

   (*b).c5.re+=(*a).c5.re;
   (*b).c5.im+=(*a).c5.im;
   (*b).c6.re+=(*a).c6.re;
   (*b).c6.im+=(*a).c6.im;

   (*b).c7.re+=(*a).c7.re;
   (*b).c7.im+=(*a).c7.im;
   (*b).c8.re+=(*a).c8.re;
   (*b).c8.im+=(*a).c8.im;
}


void sl3_sub(sl3_alg_dble *a,sl3_alg_dble *b)
{
   (*b).c1.re-=(*a).c1.re;
   (*b).c1.im-=(*a).c1.im;
   (*b).c2.re-=(*a).c2.re;
   (*b).c2.im-=(*a).c2.im;

   (*b).c3.re-=(*a).c3.re;
   (*b).c3.im-=(*a).c3.im;
   (*b).c4.re-=(*a).c4.re;
   (*b).c4.im-=(*a).c4.im;

   (*b).c5.re-=(*a).c5.re;
   (*b).c5.im-=(*a).c5.im;
   (*b).c6.re-=(*a).c6.re;
   (*b).c6.im-=(*a).c6.im;

   (*b).c7.re-=(*a).c7.re;
   (*b).c7.im-=(*a).c7.im;
   (*b).c8.re-=(*a).c8.re;
   (*b).c8.im-=(*a).c8.im;
}


void sl3_mulr(double *r,sl3_alg_dble *a,sl3_alg_dble *b)
{
   (*b).c1.re=(*r)*(*a).c1.re;
   (*b).c1.im=(*r)*(*a).c1.im;
   (*b).c2.re=(*r)*(*a).c2.re;
   (*b).c2.im=(*r)*(*a).c2.im;

   (*b).c3.re=(*r)*(*a).c3.re;
   (*b).c3.im=(*r)*(*a).c3.im;
   (*b).c4.re=(*r)*(*a).c4.re;
   (*b).c4.im=(*r)*(*a).c4.im;

   (*b).c5.re=(*r)*(*a).c5.re;
   (*b).c5.im=(*r)*(*a).c5.im;
   (*b).c6.re=(*r)*(*a).c6.re;
   (*b).c6.im=(*r)*(*a).c6.im;

   (*b).c7.re=(*r)*(*a).c7.re;
   (*b).c7.im=(*r)*(*a).c7.im;
   (*b).c8.re=(*r)*(*a).c8.re;
   (*b).c8.im=(*r)*(*a).c8.im;
}


void sl3_mulr_add(double *r,sl3_alg_dble *a,sl3_alg_dble *b)
{
   (*b).c1.re+=(*r)*(*a).c1.re;
   (*b).c1.im+=(*r)*(*a).c1.im;
   (*b).c2.re+=(*r)*(*a).c2.re;
   (*b).c2.im+=(*r)*(*a).c2.im;

   (*b).c3.re+=(*r)*(*a).c3.re;
   (*b).c3.im+=(*r)*(*a).c3.im;
   (*b).c4.re+=(*r)*(*a).c4.re;
   (*b).c4.im+=(*r)*(*a).c4.im;

   (*b).c5.re+=(*r)*(*a).c5.re;
   (*b).c5.im+=(*r)*(*a).c5.im;
   (*b).c6.re+=(*r)*(*a).c6.re;
   (*b).c6.im+=(*r)*(*a).c6.im;

   (*b).c7.re+=(*r)*(*a).c7.re;
   (*b).c7.im+=(*r)*(*a).c7.im;
   (*b).c8.re+=(*r)*(*a).c8.re;
   (*b).c8.im+=(*r)*(*a).c8.im;
}


void sl3_to_su3(sl3_alg_dble *a,su3_dble *u)
{
   (*u).c11.re=(*a).c1.re;
   (*u).c11.im=(*a).c1.im;
   (*u).c22.re=(*a).c2.re;
   (*u).c22.im=(*a).c2.im;
   (*u).c33.re=-(*a).c1.re-(*a).c2.re;
   (*u).c33.im=-(*a).c1.im-(*a).c2.im;

   (*u).c12.re=(*a).c3.re;
   (*u).c12.im=(*a).c3.im;
   (*u).c21.re=(*a).c4.re;
   (*u).c21.im=(*a).c4.im;

   (*u).c13.re=(*a).c5.re;
   (*u).c13.im=(*a).c5.im;
   (*u).c31.re=(*a).c6.re;
   (*u).c31.im=(*a).c6.im;

   (*u).c23.re=(*a).c7.re;
   (*u).c23.im=(*a).c7.im;
   (*u).c32.re=(*a).c8.re;
   (*u).c32.im=(*a).c8.im;
}


void su3_to_sl3(su3_dble *u,sl3_alg_dble *a)
{
   complex_dble z;

   z.re=c13*((*u).c11.re+(*u).c22.re+(*u).c33.re);
   z.im=c13*((*u).c11.im+(*u).c22.im+(*u).c33.im);

   (*a).c1.re=(*u).c11.re-z.re;
   (*a).c1.im=(*u).c11.im-z.im;
   (*a).c2.re=(*u).c22.re-z.re;
   (*a).c2.im=(*u).c22.im-z.im;

   (*a).c3.re=(*u).c12.re;
   (*a).c3.im=(*u).c12.im;
   (*a).c4.re=(*u).c21.re;
   (*a).c4.im=(*u).c21.im;

   (*a).c5.re=(*u).c13.re;
   (*a).c5.im=(*u).c13.im;
   (*a).c6.re=(*u).c31.re;
   (*a).c6.im=(*u).c31.im;

   (*a).c7.re=(*u).c23.re;
   (*a).c7.im=(*u).c23.im;
   (*a).c8.re=(*u).c32.re;
   (*a).c8.im=(*u).c32.im;
}


void su3_mulr_add_sl3(double *r,su3_dble *u,sl3_alg_dble *a)
{
   complex_dble z;

   z.re=c13*((*u).c11.re+(*u).c22.re+(*u).c33.re);
   z.im=c13*((*u).c11.im+(*u).c22.im+(*u).c33.im);

   (*a).c1.re+=(*r)*((*u).c11.re-z.re);
   (*a).c1.im+=(*r)*((*u).c11.im-z.im);
   (*a).c2.re+=(*r)*((*u).c22.re-z.re);
   (*a).c2.im+=(*r)*((*u).c22.im-z.im);

   (*a).c3.re+=(*r)*((*u).c12.re);
   (*a).c3.im+=(*r)*((*u).c12.im);
   (*a).c4.re+=(*r)*((*u).c21.re);
   (*a).c4.im+=(*r)*((*u).c21.im);

   (*a).c5.re+=(*r)*((*u).c13.re);
   (*a).c5.im+=(*r)*((*u).c13.im);
   (*a).c6.re+=(*r)*((*u).c31.re);
   (*a).c6.im+=(*r)*((*u).c31.im);

   (*a).c7.re+=(*r)*((*u).c23.re);
   (*a).c7.im+=(*r)*((*u).c23.im);
   (*a).c8.re+=(*r)*((*u).c32.re);
   (*a).c8.im+=(*r)*((*u).c32.im);
}

#endif

void pexp_sl3_alg(int n,sl3_alg_dble *a,su3_dble *u)
{
   int k,l,j;

   if (n>0)
   {
      if (n>ns)
         alloc_arrays(n);

      for (k=0;k<n;k++)
         sl3_to_su3(a+k,Aw+k);

      for (k=n;k>0;k--)
      {
         for (l=(n-k);l>=0;l--)
         {
            cm3x3_mulr(ck1+k,Aw+l,u+l);

            for (j=0;j<l;j++)
               cm3x3_mul_add(Aw+j,u+l-1-j,u+l);
         }
      }
   }
}


void mexp_sl3_alg(int n,sl3_alg_dble *a,su3_dble *u)
{
   int k,l,j;

   if (n>0)
   {
      if (n>ns)
         alloc_arrays(n);

      for (k=0;k<n;k++)
         sl3_to_su3(a+k,Aw+k);

      for (k=n;k>0;k--)
      {
         for (l=(n-k);l>=0;l--)
         {
            cm3x3_mulr(ck2+k,Aw+l,u+l);

            for (j=0;j<l;j++)
               cm3x3_mul_add(Aw+j,u+l-1-j,u+l);
         }
      }
   }
}


complex_dble sl3_biform(sl3_alg_dble *a,sl3_alg_dble *b)
{
   complex_dble zx,zy,sm;

   zx.re=(*a).c1.re+(*a).c2.re;
   zx.im=(*a).c1.im+(*a).c2.im;
   zy.re=(*b).c1.re+(*b).c2.re;
   zy.im=(*b).c1.im+(*b).c2.im;
   sm.re=zx.re*zy.re-zx.im*zy.im;
   sm.im=zx.re*zy.im+zx.im*zy.re;

   sm.re+=(*a).c1.re*(*b).c1.re-(*a).c1.im*(*b).c1.im;
   sm.im+=(*a).c1.re*(*b).c1.im+(*a).c1.im*(*b).c1.re;
   sm.re+=(*a).c2.re*(*b).c2.re-(*a).c2.im*(*b).c2.im;
   sm.im+=(*a).c2.re*(*b).c2.im+(*a).c2.im*(*b).c2.re;
   sm.re+=(*a).c3.re*(*b).c4.re-(*a).c3.im*(*b).c4.im;
   sm.im+=(*a).c3.re*(*b).c4.im+(*a).c3.im*(*b).c4.re;
   sm.re+=(*a).c4.re*(*b).c3.re-(*a).c4.im*(*b).c3.im;
   sm.im+=(*a).c4.re*(*b).c3.im+(*a).c4.im*(*b).c3.re;
   sm.re+=(*a).c5.re*(*b).c6.re-(*a).c5.im*(*b).c6.im;
   sm.im+=(*a).c5.re*(*b).c6.im+(*a).c5.im*(*b).c6.re;
   sm.re+=(*a).c6.re*(*b).c5.re-(*a).c6.im*(*b).c5.im;
   sm.im+=(*a).c6.re*(*b).c5.im+(*a).c6.im*(*b).c5.re;
   sm.re+=(*a).c7.re*(*b).c8.re-(*a).c7.im*(*b).c8.im;
   sm.im+=(*a).c7.re*(*b).c8.im+(*a).c7.im*(*b).c8.re;
   sm.re+=(*a).c8.re*(*b).c7.re-(*a).c8.im*(*b).c7.im;
   sm.im+=(*a).c8.re*(*b).c7.im+(*a).c8.im*(*b).c7.re;

   sm.re*=(-2.0);
   sm.im*=(-2.0);

   return sm;
}
