
/*******************************************************************************
*
* File th_parms.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Specification of the theory and the boundary conditions.
*
*   th_parms_t set_th_parms(theory_t th,int bc,int Nf)
*     The parameters are
*
*       th             The supported theories currently are PHI4,YM,LYM,
*                      QCD and LQCD.
*
*       bc             Boundary condition (1: SF, 2: open-SF, 3: periodic,
*                      4: antiperiodic in time, periodic in space). In the
*                      gauge theories, bc must be 1 or 2.
*
*       Nf             Number of quark flavours. The program sets Nf=0 in
*                      the theories other than QCD and LQCD, independently
*                      of the value of the argument Nf.
*
*     The return value is a structure that contains these parameters.
*
*   th_parms_t th_parms(void)
*     Returns the current theory parameters in a structure that contains
*     the above parameters.
*
*   void print_th_parms(void)
*     Prints the theory parameters to stdout on MPI process 0.
*
*   void write_th_parms(FILE *fdat)
*     Writes the global lattices sizes and the theory parameters to the
*     file fdat on MPI process 0.
*
*   void check_th_parms(FILE *fdat)
*     Compares the global lattice sizes and the theory parameters with
*     the values stored on the file fdat on MPI process 0, assuming the
*     latter were written to the file by the program write_th_parms().
*
*   int bc_type(void)
*     Returns the type of the chosen boundary conditions (i.e. the value
*     of the parameter bc).
*
* To ensure the consistency of the data base, the parameters must be set
* simultaneously on all processes. The program set_th_parms() may be called
* at most once. Moreover, it must be called before the geometry arrays are
* set up.
*
* See the notes doc/gfnorm.pdf for a description of the gauge action and the
* boundary conditions.
*
*******************************************************************************/

#define TH_PARMS_C

#include <stdlib.h>
#include <stdio.h>
#include "mpi.h"
#include "utils.h"
#include "pmdb.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)

static int th_flag=0;
static th_parms_t ths={THEORIES,0,0};


th_parms_t set_th_parms(theory_t th,int bc,int Nf)
{
   int iprms[3];

   error(th_flag!=0,1,"set_th_parms [th_parms.c]",
         "Attempt to reset the theory parameters");

   error(iup[0][0]!=0,1,"set_th_parms [th_parms.c]",
         "Geometry arrays are already set");

   if (NPROC>1)
   {
      iprms[0]=(int)(th);
      iprms[1]=bc;
      iprms[2]=Nf;

      MPI_Bcast(iprms,3,MPI_INT,0,MPI_COMM_WORLD);

      error((iprms[0]!=(int)(th))||(iprms[1]!=bc)||(iprms[2]!=Nf),1,
            "set_th_parms [th_parms.c]","Parameters are not global");
   }

   error_root(((int)(th)<0)||((int)(th)>=(int)(THEORIES)),1,
              "set_th_parms [th_parms.c]","Parameter th is out of range");

   error_root((bc<1)||(bc>4)||((th!=PHI4)&&(bc>2)),1,
              "set_th_parms [th_parms.c]","Parameter bc is out of range");

   error_root(((th==QCD)||(th==LQCD))&&(Nf<1),1,
              "set_th_parms [th_parms.c]","Parameter Nf is out of range");

   ths.th=th;
   ths.bc=bc;
   if ((th==QCD)||(th==LQCD))
      ths.Nf=Nf;
   else
      ths.Nf=0;
   th_flag=1;

   return ths;
}


th_parms_t th_parms(void)
{
   return ths;
}


void print_th_parms(void)
{
   int my_rank;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      error_root(th_flag!=1,1,"print_th_parms [th_parms.c]",
                 "Theory parameters are not set");

      printf("Theory parameters:\n");

      if (ths.th==PHI4)
         printf("Phi^4 theory\n");
      else if (ths.th==YM)
         printf("SU(3) gauge theory\n");
      else if (ths.th==LYM)
         printf("SU(3) lattice gauge theory\n");
      else if (ths.th==QCD)
         printf("QCD with %d quarks\n",ths.Nf);
      else if (ths.th==LQCD)
         printf("Lattice QCD with %d quarks\n",ths.Nf);

      if (ths.bc==1)
         printf("SF boundary conditions\n\n");
      else if (ths.bc==2)
         printf("open-SF boundary conditions\n\n");
      else if (ths.bc==3)
         printf("Periodic boundary conditions\n\n");
      else if (ths.bc==4)
         printf("Antiperiodic boundary conditions in time, "
                "periodic in space\n\n");
   }
}


void write_th_parms(FILE *fdat)
{
   int my_rank,endian,iw;
   stdint_t istd[7];

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   endian=endianness();

   if (my_rank==0)
   {
      error_root(th_flag!=1,1,"write_th_parms [th_parms.c]",
                 "Theory parameters are not set");

      istd[0]=(stdint_t)(N0);
      istd[1]=(stdint_t)(N1);
      istd[2]=(stdint_t)(N2);
      istd[3]=(stdint_t)(N3);
      istd[4]=(stdint_t)(ths.th);
      istd[5]=(stdint_t)(ths.bc);
      istd[6]=(stdint_t)(ths.Nf);

      if (endian==BIG_ENDIAN)
         bswap_int(7,istd);

      iw=fwrite(istd,sizeof(stdint_t),7,fdat);

      error_root(iw!=7,1,"write_th_parms [th_parms.c]",
                 "Incorrect write count");
   }
}


void check_th_parms(FILE *fdat)
{
   int my_rank,endian,ir,ie;
   stdint_t istd[7];

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   endian=endianness();

   if (my_rank==0)
   {
      error_root(th_flag!=1,1,"check_th_parms [th_parms.c]",
                 "Theory parameters are not set");

      ir=fread(istd,sizeof(stdint_t),7,fdat);

      if (endian==BIG_ENDIAN)
         bswap_int(7,istd);

      ie=0;
      ie|=(istd[0]!=(stdint_t)(N0));
      ie|=(istd[1]!=(stdint_t)(N1));
      ie|=(istd[2]!=(stdint_t)(N2));
      ie|=(istd[3]!=(stdint_t)(N3));
      ie|=(istd[4]!=(stdint_t)(ths.th));
      ie|=(istd[5]!=(stdint_t)(ths.bc));
      ie|=(istd[6]!=(stdint_t)(ths.Nf));

      error_root(ir!=7,1,"check_th_parms [th_parms.c]",
                 "Incorrect read count");

      error_root(ie!=0,1,"check_th_parms [th_parms.c]",
                 "Parameters do not match");
   }
}


int bc_type(void)
{
   return ths.bc;
}
