
/*******************************************************************************
*
* File Eplaq.c
*
* Copyright (C) 2015, 2016 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Perturbation expansion of the plaquette Yang-Mills action density.
*
*   void Eplaq(int n,su3_dble *ud,complex_dble *E_sp,complex_dble *E_tm)
*     Computes the Yang-Mills action density up to order n in the gauge
*     coupling, assuming ud contains the link fields obtained in stochastic
*     perturbation theory up to order n (see the notes). The average of
*     the spatial and time-like components of the density are calculated at
*     global time x0=T/2 (where T=NPROC0*L0) and their coefficients at order
*     k=0,..,n are assigned to E_sp[k] and E_tm[k], respectively.
*
*   void Splaq(int n,su3_dble *ud,complex_dble *S)
*     Computes the Wilson gauge action up to order n in the gauge coupling,
*     assuming ud contains the link fields obtained in stochastic perturbation
*     theory up to order n (see the notes).
*
* The Yang-Mills action density E(x) computed in this module is defined by
*
*  E(x)=1/(4*g0^2)*sum_{mu,nu}[F_{mu,nu}^a(x)]^2,
*
*  [F_{mu,nu}^a(x)]^2=(1/2)*sum_{x in p}tr{2-U(p)-U(p)^(-1)},
*
* where the sum in the second formula runs over the 4 plaquette loops in the
* (mu,nu)-plane which pass through x. Eplaq() calculates the contributions to
* E(x) from the spatial planes (mu,nu>0) and the time-like planes (mu=0,nu>0
* or nu=0, mu>0) separately, and their average over the points x at global
* time x0=T/2 is returned.
*
* The Wilson action is defined by
*
*   S=(1/g0^2)*sum_p w0(p)*(1/2)*tr{2-U(p)-U(p)^(-1)},
*
* where the sum runs over all oriented plaquettes p and
*
*   w0(p)=1/2 if p is space-like at time 0,
*
*        =1   otherwise.
*
* The values S[k], k=0,..,n, computed by Splaq() are the coefficients of g0^k
* in the perturbation expansion of S.
*
* At order n in the gauge coupling, stochastic perturbation theory delivers
* sets of n+1 link fields representing the coefficients in the expansion
*
*   U=1+g0*U_0+g0^2*U_1+..+g0^(n+1)*U_n
*
* of the gauge field in powers of the coupling (this is equivalent to knowing
* the gauge potential up to order g0^n). The n+1 link fields U_k, k=0,..,n,
* are assumed to be contained in the array ud passed to Eplaq() in such a way
* that, on any given link, the field variables at order 0,..,n come one after
* the other (cf. a2uflds() [uflds/uflds.c]).
*
* The coefficients of the plaquette values P can then be obtained to order
* g0^(n+2), since the factors in
*
*   2-U-U^(-1)=(1-U)*(1-U^(-1))
*
* are both of order g0 and are known up to order g0^(n+1).
*
* The programs in this module perform global communications and must be
* called simultaneously on all MPI processes.
*
*******************************************************************************/

#define EPLAQ_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "utils.h"
#include "aflds.h"
#include "su3fcts.h"
#include "lattice.h"
#include "uflds.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)
#define NMAX 16

static int ms=0;
static int nsm[NMAX],ism[NMAX][2],isp,itm;
static complex_dble tr ALIGNED16;
static complex_dble *plaq;
static su3_dble *Xw,*Yw,*Uw[4];


static void alloc_arrays(int n)
{
   int m,i;

   if (ms==0)
   {
      for (i=0;i<NMAX;i++)
      {
         nsm[i]=0;
         ism[i][0]=0;
         ism[i][1]=0;
      }
   }

   if (ms>0)
   {
      free(plaq);
      afree(Xw);
   }

   m=n+1;
   plaq=malloc(m*sizeof(*plaq));
   Xw=amalloc(6*m*sizeof(*Xw),ALIGN);

   error((plaq==NULL)||(Xw==NULL),1,
         "alloc_arrays [Eplaq.c]","Unable to allocate auxiliary arrays");

   for (i=0;i<m;i++)
   {
      plaq[i].re=0.0;
      plaq[i].im=0.0;
   }

   Yw=Xw+m;
   Uw[0]=Yw+m;

   for (i=1;i<4;i++)
      Uw[i]=Uw[i-1]+m;

   cm3x3_zero(6*m,Xw);
   ms=m;
}


static void set_ism(int n)
{
   int m,i;

   m=2*(n+1);

   for (i=0;i<NMAX;i++)
   {
      if (nsm[i]==0)
      {
         nsm[i]=m;
         ism[i][0]=init_hsum(m);
         ism[i][1]=init_hsum(m);
      }

      if (nsm[i]==m)
      {
         isp=ism[i][0];
         itm=ism[i][1];
         return;
      }
   }

   error(1,1,"set_ism [Eplaq.c]","Requested too many hsum instances\n");
}


static complex_dble top_order(int n,su3_dble *X,su3_dble *Y)
{
   int j;
   complex_dble z;

   z.re=0.0;
   z.im=0.0;

   for (j=0;j<=n;j++)
   {
      cm3x3_tr(X+j,Y+n-j,&tr);
      z.re-=tr.re;
      z.im-=tr.im;
   }

   return z;
}


static void local_Eplaq(int n,su3_dble *ud)
{
   int m,i;
   int ix,t,pln,ip[4];

   m=n+1;

   for (ix=0;ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (pln=0;pln<6;pln++)
      {
         if ((t==(N0/2))||((t==((N0/2)-1))&&(pln<3)))
         {
            plaq_aidx(pln,ix,ip);

            cm3x3_assign(m,ud+m*ip[0],Uw[0]);
            cm3x3_assign(m,ud+m*ip[1],Uw[1]);
            cm3x3_inv_poly(m,ud+m*ip[2],Uw[2]);
            cm3x3_inv_poly(m,ud+m*ip[3],Uw[3]);

            cm3x3_varmul_poly(m,Uw[3],Uw[2],Xw);
            cm3x3_varmul_poly(m,Uw[1],Xw,Yw);
            cm3x3_varmul_poly(m,Uw[0],Yw,Xw);
            cm3x3_inv_poly(m,Xw,Yw);

            for (i=1;i<m;i++)
            {
               plaq[i-1].re=
                  Xw[i].c11.re+Xw[i].c22.re+Xw[i].c33.re+
                  Yw[i].c11.re+Yw[i].c22.re+Yw[i].c33.re;

               plaq[i-1].im=
                  Xw[i].c11.im+Xw[i].c22.im+Xw[i].c33.im+
                  Yw[i].c11.im+Yw[i].c22.im+Yw[i].c33.im;
            }

            plaq[n]=top_order(n,Xw,Yw);

            if (pln<3)
               add_to_hsum(itm,(double*)(plaq));
            else
               add_to_hsum(isp,(double*)(plaq));
         }
      }
   }
}


void Eplaq(int n,su3_dble *ud,complex_dble *E_sp,complex_dble *E_tm)
{
   int i,iprms[1];
   double r;

   error_root(n<0,1,"Eplaq [Eplaq.c]","Parameter n is out of range");

   if (NPROC>1)
   {
      iprms[0]=n;
      MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);
      error(iprms[0]!=n,1,"Eplaq [Eplaq.c]","Parameters n is not global");
   }

   if ((n+1)>ms)
      alloc_arrays(n);
   copy_bnd_ud(n,ud);

   set_ism(n);
   reset_hsum(isp);
   reset_hsum(itm);
   local_Eplaq(n,ud);

   global_hsum(isp,(double*)(E_sp));
   global_hsum(itm,(double*)(E_tm));
   r=-1.0/(double)(N1*N2*N3);

   for (i=0;i<=n;i++)
   {
      E_sp[i].re*=r;
      E_sp[i].im*=r;
   }

   r*=0.5;

   for (i=0;i<=n;i++)
   {
      E_tm[i].re*=r;
      E_tm[i].im*=r;
   }
}


static void local_Splaq(int n,su3_dble *ud)
{
   int m,i;
   int ix,t,pln,ip[4];

   m=n+1;

   for (ix=0;ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (pln=0;pln<6;pln++)
      {
         plaq_aidx(pln,ix,ip);

         cm3x3_assign(m,ud+m*ip[0],Uw[0]);
         cm3x3_assign(m,ud+m*ip[1],Uw[1]);
         cm3x3_inv_poly(m,ud+m*ip[2],Uw[2]);
         cm3x3_inv_poly(m,ud+m*ip[3],Uw[3]);

         cm3x3_varmul_poly(m,Uw[3],Uw[2],Xw);
         cm3x3_varmul_poly(m,Uw[1],Xw,Yw);
         cm3x3_varmul_poly(m,Uw[0],Yw,Xw);
         cm3x3_inv_poly(m,Xw,Yw);

         for (i=1;i<m;i++)
         {
            plaq[i-1].re=
               Xw[i].c11.re+Xw[i].c22.re+Xw[i].c33.re+
               Yw[i].c11.re+Yw[i].c22.re+Yw[i].c33.re;

            plaq[i-1].im=
               Xw[i].c11.im+Xw[i].c22.im+Xw[i].c33.im+
               Yw[i].c11.im+Yw[i].c22.im+Yw[i].c33.im;
         }

         plaq[n]=top_order(n,Xw,Yw);

         if ((t==0)&&(pln>=3))
         {
            for (i=0;i<m;i++)
            {
               plaq[i].re*=0.5;
               plaq[i].im*=0.5;
            }
         }

         add_to_hsum(isp,(double*)(plaq));
      }
   }
}


void Splaq(int n,su3_dble *ud,complex_dble *S)
{
   int m,i,iprms[1];

   error_root(n<0,1,"Splaq [Eplaq.c]","Parameter n is out of range");

   if (NPROC>1)
   {
      iprms[0]=n;
      MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);
      error(iprms[0]!=n,1,"Splaq [Eplaq.c]","Parameter n is not global");
   }

   m=n+1;

   if (m>ms)
      alloc_arrays(n);
   copy_bnd_ud(n,ud);

   if (cpr[0]==(NPROC0-1))
      cm3x3_zero(3*m,ud+(4*VOLUME+7*(BNDRY/4))*m);

   set_ism(n);
   reset_hsum(isp);
   local_Splaq(n,ud);
   global_hsum(isp,(double*)(S));

   for (i=0;i<m;i++)
   {
      S[i].re=-S[i].re;
      S[i].im=-S[i].im;
   }
}
