
/*******************************************************************************
*
* File cnum.c
*
* Copyright (C) 2014, 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Rational coefficient algebra.
*
*   cnum_t *alloc_cnum(int nc,int mx)
*     Allocates a new array of nc coefficients of type cnum_t with
*     maximal polynomial degree equal to mx. The coefficients and
*     the degree of the polynomials are set to zero.
*
*   void free_cnum(int nc,cnum_t *cn)
*     Frees the contents of the cnum_t structures cn[0],..,cn[nc-1]
*     and then the array cn itself, assuming the array was allocated
*     by alloc_cnum().
*
*   void set_cnum(int n,long (*r)[2],cnum_t *cn)
*     Assigns the rational numbers r[0][0]/r[0][1],..,r[n][0]/r[n][1] to
*     the array elements (*cn).c[0],..(*cn).c[n] and sets the remaining
*     array elements to zero.
*
*   void copy_cnum(cnum_t *cn1,cnum_t *cn2)
*     Copies cn1 to cn2 so that cn2 represents the same coefficient as
*     cn1. The size of the array (*cn2).c is adapted only if needed and
*     all unused elements of the array are set to zero.
*
*   void add_cnum(cnum_t *cn1,cnum_t *cn2)
*     Adds cn1 to cn2.
*
*   void sub_cnum(cnum_t *cn1,cnum_t *cn2)
*     Subtracts cn1 from cn2.
*
*   void mul_cnum(cnum_t *cn1,cnum_t *cn2)
*     Multiplies cn2 by cn1.
*
*   int is_zero_cnum(cnum_t *cn)
*     Returns 1 if cn vanishes (for all values of the unknown) and 0
*     otherwise.
*
*   double eval_cnum(double x,cnum_t *cn)
*     Returns the value of the coefficient cn, assuming the unknown
*     variable has the value x. An error occurs if one of the rational
*     coefficients (*cn).c[i] is too large or too small to fit into
*     the range of double-precision floating-point numbers.
*
*   int print_cnum(FILE *out,char *x,cnum_t *cn)
*     Prints the coefficient cn to the stream out, using the symbol x
*     for the unknown variable. The program returns the number of bytes
*     written to the stream.
*
* The coefficients of type cnum_t are polynomials in one unknown with
* arbitrary-precision rational coefficients, the latter being handled
* by the GNU Multiple-Precision (gmp) library. Such coefficients can be
* initialized, added and multiplied using the functions in this module.
*
* Structures of type cnum_t have the following elements:
*
*  mx        Maximal degree of the polynomial.
*
*  n         Actual degree of the polynomial.
*
*  c[i]      Coefficient number i (i=0,1,..,mx), where c[n+1],..,c[mx]
*            are equal to zero.
*
* The programs adjust mx and reallocate the coefficient array c when needed.
* Structures passed as arguments are assumed to be allocated and properly
* initialized.
*
* The programs in this module depend on the gmp library and the associated
* gmp.h include file (see info gmp).
*
*******************************************************************************/

#define CNUM_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "utils.h"

static int mxs=-1,pmx=0,pmn,nrs=0;
static char *rs=NULL;
static mpq_t *cs=NULL;


static void free_mpq(int mx,mpq_t *c)
{
   int i;

   for (i=0;i<=mx;i++)
      mpq_clear(c[i]);

   free(c);
}


static mpq_t *new_mpq(int mx)
{
   int i;
   mpq_t *c;

   c=malloc((mx+1)*sizeof(*c));

   if (c!=NULL)
   {
      for (i=0;i<=mx;i++)
         mpq_init(c[i]);
   }
   else
      error_loc(1,1,"new mpq [cnum.c]","Unable to allocate mpq array");

   return c;
}


static void check_degree(cnum_t *cn)
{
   int n,i;
   mpq_t *c;

   n=(*cn).n;
   c=(*cn).c;

   for (i=n;i>0;i--)
   {
      if (mpq_sgn(c[i])!=0)
         break;
   }

   (*cn).n=i;
}


static void set_cs(int mx)
{
   if (mx>mxs)
   {
      if (cs!=NULL)
         free_mpq(mxs,cs);

      mxs=mx;
      cs=new_mpq(mxs);
   }
}


static void check_size(mpq_t *c)
{
   int p;

   if (pmx==0)
   {
      (void)(frexp(DBL_MAX,&pmx));
      (void)(frexp(DBL_MIN,&pmn));
      pmx-=1;
   }

   p=mpz_sizeinbase(mpq_numref(*c),2)-mpz_sizeinbase(mpq_denref(*c),2);

   if ((p<pmx)&&(p>pmn))
      return;
   else
      error_loc(1,1,"check_size [cnum.c]",
            "Conversion to double either over- or underflows");
}


static void set_rs(int n)
{
   if (n>nrs)
   {
      if (rs!=NULL)
         free(rs);

      nrs=n;
      rs=malloc(nrs*sizeof(*rs));
      error_loc(rs==NULL,1,"set_rs [cnum.c]",
            "Unable to allocate rational-number string");
   }
}


cnum_t *alloc_cnum(int nc,int mx)
{
   int i;
   cnum_t *cn;

   error_loc((nc<1)||(mx<0),1,"alloc_cnum [cnum.c]","Improper arguments");

   cn=malloc(nc*sizeof(*cn));
   error_loc(cn==NULL,1,"alloc_cnum [cnum.c]",
         "Unable to allocate cnum_t structures");

   for (i=0;i<nc;i++)
   {
      cn[i].mx=mx;
      cn[i].n=0;
      cn[i].c=new_mpq(mx);
   }

   return cn;
}


void free_cnum(int nc,cnum_t *cn)
{
   int i;

   for (i=0;i<nc;i++)
      free_mpq(cn[i].mx,cn[i].c);

   free(cn);
}


void set_cnum(int n,long (*r)[2],cnum_t *cn)
{
   int i;
   mpq_t *c;

   if (n>(*cn).mx)
   {
      free_mpq((*cn).mx,(*cn).c);
      (*cn).mx=n;
      (*cn).c=new_mpq(n);
   }

   (*cn).n=n;
   c=(*cn).c;

   for (i=0;i<=n;i++)
   {
      if (r[i][1]>=0L)
         mpq_set_si(c[i],r[i][0],(unsigned long)(r[i][1]));
      else
         mpq_set_si(c[i],-r[i][0],(unsigned long)(-r[i][1]));

      mpq_canonicalize(c[i]);
   }

   for (i=(n+1);i<=(*cn).mx;i++)
      mpq_set_si(c[i],0L,1U);

   check_degree(cn);
}


void copy_cnum(cnum_t *cn1,cnum_t *cn2)
{
   int n,i;

   n=(*cn1).n;

   if (n>(*cn2).mx)
   {
      free_mpq((*cn2).mx,(*cn2).c);
      (*cn2).mx=n;
      (*cn2).c=new_mpq(n);
   }

   (*cn2).n=n;

   for (i=0;i<=n;i++)
      mpq_set((*cn2).c[i],(*cn1).c[i]);

   for (i=(n+1);i<=(*cn2).mx;i++)
      mpq_set_si((*cn2).c[i],0L,1U);
}


void add_cnum(cnum_t *cn1,cnum_t *cn2)
{
   int n1,n2,i;

   n1=(*cn1).n;
   n2=(*cn2).n;

   if (n1<=n2)
   {
      set_cs(0);

      for (i=0;i<=n1;i++)
      {
         mpq_set(cs[0],(*cn2).c[i]);
         mpq_add((*cn2).c[i],cs[0],(*cn1).c[i]);
      }
   }
   else
   {
      set_cs(n2);

      for (i=0;i<=n2;i++)
         mpq_set(cs[i],(*cn2).c[i]);

      if (n1>(*cn2).mx)
      {
         free_mpq((*cn2).mx,(*cn2).c);
         (*cn2).mx=n1;
         (*cn2).c=new_mpq(n1);
      }

      for (i=0;i<=n1;i++)
      {
         if (i<=n2)
            mpq_add((*cn2).c[i],cs[i],(*cn1).c[i]);
         else
            mpq_set((*cn2).c[i],(*cn1).c[i]);
      }

      (*cn2).n=n1;
   }

   check_degree(cn2);
}


void sub_cnum(cnum_t *cn1,cnum_t *cn2)
{
   int n1,n2,i;

   n1=(*cn1).n;
   n2=(*cn2).n;

   if (n1<=n2)
   {
      set_cs(0);

      for (i=0;i<=n1;i++)
      {
         mpq_set(cs[0],(*cn2).c[i]);
         mpq_sub((*cn2).c[i],cs[0],(*cn1).c[i]);
      }
   }
   else
   {
      set_cs(n2);

      for (i=0;i<=n2;i++)
         mpq_set(cs[i],(*cn2).c[i]);

      if (n1>(*cn2).mx)
      {
         free_mpq((*cn2).mx,(*cn2).c);
         (*cn2).mx=n1;
         (*cn2).c=new_mpq(n1);
      }

      for (i=0;i<=n2;i++)
         mpq_sub((*cn2).c[i],cs[i],(*cn1).c[i]);

      mpq_set_si(cs[0],0L,1U);

      for (i=(n2+1);i<=n1;i++)
         mpq_sub((*cn2).c[i],cs[0],(*cn1).c[i]);

      (*cn2).n=n1;
   }

   check_degree(cn2);
}


void mul_cnum(cnum_t *cn1,cnum_t *cn2)
{
   int n1,n2,n,i,j;

   n1=(*cn1).n;
   n2=(*cn2).n;
   n=n1+n2;
   set_cs(n2+2);

   for (i=0;i<=n2;i++)
      mpq_set(cs[i],(*cn2).c[i]);

   if (n>(*cn2).mx)
   {
      free_mpq((*cn2).mx,(*cn2).c);
      (*cn2).mx=n;
      (*cn2).c=new_mpq(n);
   }

   for (i=0;i<=n;i++)
      mpq_set_si((*cn2).c[i],0L,1U);

   for (i=0;i<=n1;i++)
   {
      for (j=0;j<=n2;j++)
      {
         mpq_mul(cs[n2+1],(*cn1).c[i],cs[j]);
         mpq_set(cs[n2+2],(*cn2).c[i+j]);
         mpq_add((*cn2).c[i+j],cs[n2+1],cs[n2+2]);
      }
   }

   (*cn2).n=n;
   check_degree(cn2);
}


int is_zero_cnum(cnum_t *cn)
{
   int n,i;
   mpq_t *c;

   n=(*cn).n;
   c=(*cn).c;

   for (i=0;i<=n;i++)
   {
      if (mpq_sgn(c[i])!=0)
         return 0;
   }

   return 1;
}


double eval_cnum(double x,cnum_t *cn)
{
   int n,i;
   double r;
   mpq_t *c;

   n=(*cn).n;
   c=(*cn).c;

   check_size(c+n);
   r=mpq_get_d(c[n]);

   for (i=(n-1);i>=0;i--)
   {
      check_size(c+i);
      r=x*r+mpq_get_d(c[i]);
   }

   return r;
}


int print_cnum(FILE *out,char *x,cnum_t *cn)
{
   int n,i,j;
   int nw,s,sm;
   mpq_t *c;

   if (is_zero_cnum(cn))
   {
      nw=fprintf(out,"0");
      return nw;
   }

   n=(*cn).n;
   c=(*cn).c;
   sm=0;

   for (i=0;i<=n;i++)
   {
      s=mpz_sizeinbase(mpq_numref(c[i]),10)+mpz_sizeinbase(mpq_denref(c[i]),10);

      if (s>sm)
         sm=s;
   }

   set_rs(sm+3);
   j=0;
   nw=0;

   for (i=0;i<=n;i++)
   {
      if (mpq_sgn(c[i])!=0)
      {
         (void)(mpq_get_str(rs,10,c[i]));

         if ((j==0)||(mpq_sgn(c[i])<0))
            nw+=fprintf(out,"%s",rs);
         else
            nw+=fprintf(out,"+%s",rs);

         if (i>0)
            nw+=fprintf(out,"*%s",x);
         if (i>1)
            nw+=fprintf(out,"^%d",i);

         j=1;
      }
   }

   return nw;
}
