
/*******************************************************************************
*
* File Amn.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Computation of the vertex functions A_m^n.
*
*   void Amn(int n,sl3_alg_dble **asrc,sl3_alg_dble *vxfld)
*     Assigns the vertex function A_m^n to the field vxfld for given
*     source fields asrc[0],..,asrc[n-2]. The latter are assumed to
*     satisfy the boundary conditions specified in the parameter data
*     base.
*
* See the notes
*
*  M. Luescher: "Lattice QCD vertices in ISPT", file doc/vxfcts.pdf,
*
* for the definition of the vertex functions A_m^n and a description of
* the computational strategy employed.
*
* Gauge potentials are locally stored as described in main/README.global
* and aflds/README.aflds. Both input and output fields must be of this
* type. The output field may not overlap with any of the input fields.
*
* The parameter n must be even and in the range 2<=n<=NMAX, where NMAX is
* defined in vxfcts.h.
*
* The program in this module acts globally and must be called on all MPI
* processes simultaneously.
*
*******************************************************************************/

#define AMN_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "utils.h"
#include "lattice.h"
#include "su3fcts.h"
#include "linalg.h"
#include "aflds.h"
#include "vxfcts.h"
#include "global.h"

static int ns=0;
static su3_dble *Ak,*eA,*As;


static void alloc_arrays(int n)
{
   int nd;

   nd=1<<(n-1);

   if (n>ns)
   {
      if (ns>0)
         afree(Ak);

      Ak=amalloc((n+nd)*sizeof(*Ak),ALIGN);
      error(Ak==NULL,1,"alloc_arrays [Amn.c]",
            "Unable to allocate auxiliary arrays");

      cm3x3_zero(n+nd,Ak);
      ns=n;
   }

   eA=Ak+n-1;
   As=eA+nd;
}


static void set_eA(int n,int ip,sl3_alg_dble **asrc)
{
   int i;

   for (i=0;i<(n-1);i++)
      sl3_to_su3(asrc[i]+ip,Ak+i);

   sym_prod(0,n-1,Ak,eA);
}


static void set_As(int n)
{
   int ir,is,ipmx;
   complex_dble z;
   su3_dble *X;

   ipmx=1<<(n-1);
   cm3x3_zero(1,As);

   for (ir=0;ir<ipmx;ir++)
   {
      is=(ipmx-1)^ir;
      X=eA+is;

      if (bcount(n-1,is)&0x1)
      {
         z.re=-(*X).c11.re-(*X).c22.re-(*X).c33.re;
         z.im=-(*X).c11.im-(*X).c22.im-(*X).c33.im;
      }
      else
      {
         z.re=(*X).c11.re+(*X).c22.re+(*X).c33.re;
         z.im=(*X).c11.im+(*X).c22.im+(*X).c33.im;
      }

      cm3x3_mulc_add(&z,eA+ir,As);
   }
}


void Amn(int n,sl3_alg_dble **asrc,sl3_alg_dble *vxfld)
{
   int iprms[1];
   int ix,ifc,ip;
   double bn;

   error_root((n<2)||(n>NMAX)||((n%2)!=0),1,"Amn [Amn.c]",
              "Parameter n is out of range");

   if (NPROC>1)
   {
      iprms[0]=n;
      MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);
      error(iprms[0]!=n,1,"Amn [Amn.c]","Parameter n is not global");
   }

   if (n>2)
   {
      alloc_arrays(n);
      bn=-bernoulli(n)/(double)(n);
   }
   else
      bn=-0.25;

   ip=0;

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      for (ifc=0;ifc<8;ifc++)
      {
         if (n==2)
            sl3_mulr(&bn,asrc[0]+ip,vxfld+ip);
         else
         {
            set_eA(n,ip,asrc);
            set_As(n);
            cm3x3_mulr(&bn,As,As);
            su3_to_sl3(As,vxfld+ip);
         }

         ip+=1;
      }
   }

   if (cpr[0]==(NPROC0-1))
      set_calg2zero(3,vxfld+4*VOLUME+7*(BNDRY/4));
}
