
/*******************************************************************************
*
* File CFPn.c
*
* Copyright (C) 2015, 2017 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Computation of the vertex functions C_FP^n.
*
*   void CFPn(int n,sl3_alg_dble **asrc,sl3_alg_dble *csrc,
*             sl3_alg_dble *vxfld)
*     Assigns the vertex function C_FP^n to the field vxfld for given
*     source fields asrc[0],..,asrc[n-1] and csrc (representing the gauge
*     and ghost fields, respectively). All source fields are assumed to
*     satisfy the boundary conditions specified in the parameter data base.
*
* See the notes
*
*  M. Luescher: "Lattice QCD vertices in ISPT", file doc/vxfcts.pdf,
*
* for the definition of the vertex functions C_FP^n and a description of
* the computational strategy employed.
*
* Gauge potentials are locally stored as described in main/README.global
* and aflds/README.aflds. Both input and output fields must be of this
* type. The output field may not overlap with any of the input fields.
* The ghost fields are locally stored as described in main/README.global
* and cflds/README.cflds.
*
* The parameter n must be either equal to 1 or an even number in the range
* 2<=n<=NMAX, where NMAX is defined in vxfcts.h.
*
* The program in this module acts globally and must be called on all MPI
* processes simultaneously. It requires a workspace of 1 gauge potential.
*
*******************************************************************************/

#define CFPN_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "utils.h"
#include "lattice.h"
#include "su3fcts.h"
#include "linalg.h"
#include "aflds.h"
#include "cflds.h"
#include "vxfcts.h"
#include "global.h"

#define N0 (NPROC0*L0)

static int ns=0;
static const su3_dble ud0={{0.0,0.0},{0.0,0.0},{0.0,0.0},
                           {0.0,0.0},{0.0,0.0},{0.0,0.0},
                           {0.0,0.0},{0.0,0.0},{0.0,0.0}};
static su3_dble C ALIGNED16;
static su3_dble dC ALIGNED16;
static su3_dble Xw ALIGNED16;
static su3_dble Vmu ALIGNED16;
static su3_dble *Ak,*eA;


static void alloc_arrays(int n)
{
   int nd;

   if (n>ns)
   {
      if (ns>0)
         afree(Ak);

      nd=1<<n;
      Ak=amalloc((n+nd)*sizeof(*Ak),ALIGN);
      error(Ak==NULL,1,"alloc_arrays [CFPn.c]",
            "Unable to allocate auxiliary arrays");

      cm3x3_zero(n+nd,Ak);
      ns=n;
   }

   eA=Ak+n;
}


static void set_eA(int n,int ip,sl3_alg_dble **asrc)
{
   int i;

   for (i=0;i<n;i++)
      sl3_to_su3(asrc[i]+ip,Ak+i);

   sym_prod(0,n,Ak,eA);
}


static void last_prod(int n)
{
   int ir,is,irmx;

   irmx=1<<n;
   Vmu=ud0;

   for (ir=0;ir<irmx;ir++)
   {
      is=(irmx-1)^ir;
      su3xsu3(eA+ir,&dC,&Xw);

      if (bcount(n,is)&0x1)
         cm3x3_mul_sub(&Xw,eA+is,&Vmu);
      else
         cm3x3_mul_add(&Xw,eA+is,&Vmu);
   }
}


void CFPn(int n,sl3_alg_dble **asrc,sl3_alg_dble *csrc,
          sl3_alg_dble *vxfld)
{
   int iprms[1];
   int ix,iy,t,ifc,mu,ip;
   double wc,half,bn;
   sl3_alg_dble **wad,*vmu;

   error_root((n<1)||(n>NMAX)||((n>1)&&((n%2)==1)),1,"CFPn [CFPn.c]",
              "Parameter n is out of range");

   if (NPROC>1)
   {
      iprms[0]=n;
      MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);
      error(iprms[0]!=n,1,"CFPn [CFPn.c]","Parameter n is not global");
   }

   alloc_arrays(n);
   wad=reserve_wad(1);
   vmu=wad[0];
   cpcd_int_bnd(0,csrc);
   half=0.5;
   bn=bernoulli(n);
   ip=0;

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         mu=ifc/2;

         if (ifc&0x1)
         {
            if ((t>0)||(mu>0))
               sl3_to_su3(csrc+ix,&dC);
            else
               dC=ud0;

            iy=idn[ix][mu];
            sl3_to_su3(csrc+iy,&C);
            cm3x3_sub(&C,&dC);
         }
         else
         {
            if ((t<(N0-1))||(mu>0))
            {
               iy=iup[ix][mu];
               sl3_to_su3(csrc+iy,&dC);
            }
            else
               dC=ud0;

            sl3_to_su3(csrc+ix,&C);
            cm3x3_sub(&C,&dC);
         }

         if (n==1)
         {
            cm3x3_mulr(&half,&dC,&dC);
            cm3x3_add(&C,&dC);

            sl3_to_su3(asrc[0]+ip,&Xw);
            su3xsu3(&Xw,&dC,&Vmu);
            cm3x3_mul_sub(&dC,&Xw,&Vmu);
         }
         else
         {
            set_eA(n,ip,asrc);
            last_prod(n);
         }

         su3_to_sl3(&Vmu,vmu);
         ip+=1;
         vmu+=1;
      }
   }

   divad(wad[0],vxfld);

   for (ix=0;ix<VOLUME;ix++)
   {
      t=global_time(ix);

      if (n==1)
         wc=1.0;
      else
         wc=bn;

      if (t==0)
         wc*=0.5;

      sl3_mulr(&wc,vxfld,vxfld);
      vxfld+=1;
   }

   release_wad();
}
